package Sloop::Connection;
our $VERSION = '0.1.1';
# Version 0.1.1
# - Changed IOerror usage to catch EBADFD.
# - Added onCancelHandlers.
use strict;
use warnings;

use parent 'Sloop::Base';

use Sloop::Logger;
use Sloop::Constants (':Debug');
use Sloop::Lib qw(dataLength);
use Errno qw(EAGAIN EBADFD);

require XSLoader;
XSLoader::load('Sloop::Connection', $VERSION);

=encoding utf8

=head1 NAME

Sloop::Connection

=head1 DESCRIPTION

Sloop::Base derived class representing a connection to the server; all socket descriptors
except for the server listen socket are encapsulated by something descended from
Sloop::Connection.  It adds input/output routines to Sloop::Base although the
low level details are in the Sloop::Socket classes (in other words, Sloop::Connection
is still abstract, no pure Sloop::Connection objects are instantiated).

=head1 SYNOPSIS

 my $connection = Sloop::Connection->new (
     debug => (optional) see Sloop::Base
     fault_max => (optional) integer
     fd => (optional) see Sloop::Base
     id => (optional) see Sloop::Base
     logger => (REQUIRED) see Sloop::Base
     max_recv => (optional) integer
     max_send => (optional) integer
     onCancelHandlers => (optional) Array of subroutine references
     timeout => (REQUIRED) integer
 );

Fields not documented in Sloop::Base are described below.

=over

=item B<fault_max>

If C<$connection-E<62>{errcount}> exceeds this value, the connection will be
terminated with cancel(). 'Errcount' is a count of low level read()/write() and
other errors that are consecutive (i.e., it is cleared when such an I/O call
succeeds). The default is 3.  This almost never comes into play;
don't increase it.

=item B<max_recv>

Maximum number of bytes to try reading at once, the default is 262144.  This
doesn't limit the total size of a request, just how much of it is read in
at once.  This is to help prevent fast connections involving a lot of data
from choking off everything else.  You probably should not set this lower
than the size of the system socket receive buffer, see 'man 7 socket'.

=item B<max_send>

Maximum number of bytes to try sending at once, the default is 1048576.  This
doesn't limit the total size of a response, see max_recv above.

=item B<onCancelHandlers>

If set, these will be called in order on the first call to C<$connection->cancel>,
with the C<$connection> object passed in (but do not use it to send data, 
it will be defunct).

This attribute may be set (including undef) any time post construction as well.

=item B<timeout>

Duration in seconds after which an idle $connection will be closed.  "Idle"
here means there is no pending request or response.  This value is set by
the Sloop::Server object, see 'timeout' in the SYNOPSIS there.  Connections
with a timeout of 0 never time out.

=back

=head1 ATTRIBUTES

In addition to the attributes listed for Sloop::Base, the following fields from the
constructor are accessible by the same name, with the same meaning:

=over 2

=item fault_max

=item max_recv

=item max_send

=item timeout

=back

The following attibutes may be accessed/set only I<after> the constructor call:

=over

=item C<$connection-E<62>{buffer_data}>

This is a subroutine reference that can used if you want to send a reply that is
longer than the actual data passed to Sloop::Client::reply() -- in which case,
you should be sure to pass a 'length' argument to that method that matches the
total amount that will be sent, because it is still all part of the same HTTP
response even if the data in the body is delivered in chunks.  If you cannot provide
an exact length of the total data, I<don't do this>.

Sloop::Static::directory() uses this to avoid loading large files into memory
all at once.  The subroutine will be called when the previous data has been sent,
and should return a reference to a scalar containing additional data.  As long as
this field is defined, it will continue to be called for the same reply, so when
the data is complete, the subroutine should leave C<$connection-E<62>{buffer_data}>
undefined.  It receives one argument, the Sloop::Connection derived object to
which it is attached.

=item C<$connection-E<62>{readerr}>

Stores an error code from the last failed sysread().

=item C<$connection-E<62>{readErrStr}>

String representation of the last sysread() error.

=back

=cut

## NON-API FIELDS ##

# $connection->{data_pending}
# This is a reference to a string buffer in which data waiting to be sent is kept;
# it is replenished by e.g., $connection->{buffer_data}.

# $connection->{errcount}
# Used in IOerror() method to track the number of errors recieved.

# $connection->{epfd}
# Server's epollfd -- used in setEvents().

# $connection->{events}
# The last event value set via Sloop::Server::Epoll().

sub new {
	my $class = shift;
	my %args = (
		debug => 0,
		fault_max => 3,
		max_recv => 262144,
		max_send => 1048576,
		events => 0,
		@_
	);

	my $self = $class->SUPER::new ( %args );

	$self->{buffer_data} = undef;
	$self->{data_pending} = undef;
	$self->{fault_max} = $args{fault_max};
	$self->{readerr} = 0;
	$self->{readErrStr} = "";
	$self->{errcount} = 0;

	bless $self, $class;
}

=head1 METHODS

=head2 C<$connection-E<62>cancel>

Calls $connection->shut(), which must be implemented by subclasses (in the
sloop core this is actually done polymorphically via the Sloop::Socket
implementations; see also comments about C<$base-E<62>{SHUT}> in Sloop::Base).

Derived classes which override this and do not call it should possibly
deal with C<$connection->{onCancelHandlers}>.

=cut
sub cancel {
	my $self = shift;
	my $check = $self->{SHUT};
	$self->shut();
	if (defined $self->{onCancelHandlers} && !$check) {
		$_->($self) foreach (@{$self->{onCancelHandlers}});
	}
}


# Dummy to be overridden in derived classes.  Included here  because this is
# called for all Sloop::Server connections. This method is not required to do
# anything, but it  receives one parameter (a time in seconds since the epoch) and
# should return a value: if non-zero, the server will consider the connection
# defunct and remove it from the file descriptor list. See Sloop::Client for an
# example. 
sub checktime {
	my $self = shift;
	$self->dummy('checktime', caller);
}


# Dummy to be overridden in derived classes, called at the end of sendData().
# This method is not required to do anything, but must return
# either 1 to indicate the connection is still valid and should
# continue to be polled, or -1 to indicate the connection is
# defunct.
sub doneWrite {
	my $self = shift;
	$self->dummy('doneWrite', caller);
}


# Tracks errors reported for this connection.
# Returns -1 if the connection was closed, or 0 otherwise.
sub IOerror {
	my $self = shift;
	if ($_[0] == EBADFD) {
		$self->cancel();
		return -1;
	}
	$self->{errcount}++;
	if ($self->{errcount} >= $self->{fault_max}) {
		$self->err_log("'errcount' exceeded 'fault_max' -- disconnecting.");
		$self->cancel();
		return -1;
	}
	return 0;
}


# $connection->readData($bytes)
# Called to read data in from the socket; this is essentially an error checked
# wrapper for sysread().
# 	$bytes is a scalar reference to store the number of bytes read if successful.
# 	       It is not required.
# Returns a reference to the data read, or undef if the connection is not ready or
# an error occurred.  In the later case, $connection->cancel() may be called.
sub readData {
	my ($self, $lenref) = (shift, shift);

	my $buf = "";
	$self->{readerr} = 0;
	my $bytes = $self->sysread($buf, $self->{max_recv});

	if (!defined $bytes) {
	# Non fatal error.
		if ($self->{readerr} == EAGAIN) { return undef }
		$self->err_log("Read error: ".$self->{readErrStr});
		$self->IOerror($self->{readerr});
		return undef;
	} elsif (!$bytes) {
	# Closed / disconnected by the other end
		$self->cancel();
		if ($self->{debug} & DBG_CONNECTION) {
			$self->debug_log("Disconnected.");
		}
		return undef;
	} elsif ($bytes < 0) {
	# Fatal error.
		$self->err_log("Cancelled: $self->{readErrStr}");
		$self->cancel();
		return undef;
	}

	# Reset error count.
	$self->{errcount} = 0;

	if ($self->{debug} & DBG_INPUT) {
		$self->debug_log("Read $bytes bytes.");
	}

	$$lenref = $bytes if ($lenref);
	return \$buf;
}


# Dummy to be overridden in derived classes.  Included here because this is called
# for all Sloop::Server connections when poll() reports data waiting to be read. 
# It should return one of the 'Client Ready' constants in Sloop::Constants. 
sub ready {
	my $self = shift;
	$self->dummy('ready', caller);
}


# This is called by Sloop::Server when Epoll() indicates the
# connection is ready for writing.
# Returns:
#   2 Pending data has been written and there is a pipelined request.
#   1 all pending data has been written
#   0 more data remains to be written
#  -1 the connection is defunct
sub sendData {
	my $self = shift;
	my $len = dataLength($self->{data_pending});
	if (!$len) {
		$self->doneWrite;
		return 1;
	}

	my $bytes = $self->syswrite(${$self->{data_pending}}, $len > $self->{max_send} ? $self->{max_send} : $len);

	if (!defined $bytes) {
	# Only Sloop::Socket::Regular does this.
		if ($! != EAGAIN) {
			$self->err_log("syswrite error: $!");
			return $self->IOerror($!);
		}
		return 0;
	}

	if ($bytes < 0) {
	# Only Sloop::Socket::TLS does this.
		return 0 if ($bytes == -1); # EAGAIN
		return -1 if ($bytes == -3); # Fatal.
		return $self->IOerror(0);
	}

	# Reset error count.
	$self->{errcount} = 0;

	if ($self->{debug} & DBG_OUTPUT) {
		$self->debug_log("syswrite: $bytes");
	}

	if ($bytes < $len) {
		${$self->{data_pending}} = substr(${$self->{data_pending}}, $bytes);
		return 0;
	} else { $self->{data_pending} = undef }

	if (defined $self->{buffer_data}) {
		$self->{data_pending} = $self->{buffer_data}($self);
		return 0 if (defined $self->{data_pending});
	} else { $self->{data_pending} = undef }

	return $self->doneWrite();
}


# Abstract dummy; this should duplictate the functionality of IO::Socket::sysread().
sub sysread {
	my $self = shift;
	$self->dummy('sysread', caller);
}


# Abstract dummy; this should duplictate the functionality of IO::Socket::syswrite().
sub syswrite {
	my $self = shift;
	$self->dummy('syswrite', caller);
}


1;
