package Sloop::Base;
our $VERSION = '0.1.1';
# 0.1.1
# - Added okay() and req_log().
# - Tweaked debug_log to handler "caller" from top level script.
use strict;
use warnings;

use Carp;
our @CARP_NOT = qw(
	Sloop::WebSocket
	Sloop::Client
	Sloop::Client::Request
	Sloop::Multipart
	Sloop::Server
	Sloop::TokenAuth
);

use Sloop::Logger;
use Sloop::Constants;

#use YAML::XS;

=encoding utf8

=head1 NAME

Sloop::Base

=head1 DESCRIPTION

This is the top of the fundamental sloop class hierarchy, which includes
Sloop::Server and Sloop::Client.  It represents some basic properties of an I/O
socket, but does not include any I/O functionality; that is provided by
subclasses, usually via multiple inheritance from one of the Sloop::Socket
classes.  Sloop::Base implements a "dummy" method system whereby if a subclass
does not override one of the dummy methods, calling it is a fatal error (since
perl does not otherwise implement abstract methods).  This same system is used
further down the heirarchy as well.

=head1 SYNOPSIS

 use Sloop::Base;
 my $base = Sloop::Base->new(
	debug => (optional) a Sloop::Constant ":Debug" constants
	fd => (optional) integer file descriptor
	id => (optional) string used in logging
	logger => (REQUIRED) a Sloop::Logger object
 );

=over

=item B<debug>

Zero, or one or more of the 'Debug Constants' from Sloop::Constants
added (or OR'd) together.  The default is 0.

=item B<fd> 

The file descriptor for the socket represented.  The default is -1.

=item B<id>

A string identifier used in logging.  The default is 'noid'.

=item B<logger>

An object matching the specifications of Sloop::Logger.

=back

=head1 ATTRIBUTES

All the fields from the constructor are subsequently accessible by
the same name, with the same meaning.

An additional field which is used internally by sloop and is significant
when subclassing is C<$base-E<62>{SHUT}>.  This is checked via the
API with C<$base-E<62>okay()> and should not be modified directly, so there
is no need to make use of it outside of a subclassing context.  It represents some
of the coupling between the Sloop::Server class and objects derived from
`Sloop::Connection` and `Sloop::Socket` (see comments about C<$connection-E<62>cancel>
in the docs for Sloop::Connection).  Since this coupling is integral to how
the Server object handles Connections, C<SHUT> must always be accounted for
properly.

=cut

# $High and $Count are used to track the number of simultaneous open
# connections for diagnostic purposes (see DESTROY).
our $Count = 0;
our $High = 0;

sub new {
	my $class = shift;
	my $self = {
		debug => 0,
		fd => -1,
		id => 'noid',
		@_
	};
	bless $self, $class;

	
	$self->{SHUT} = 0;

	if (++$Count > $High) { $High = $Count }
	if ($self->{debug} & Sloop::Constants::DBG_CONNECTION) {
		$self->debug_log("Now $Count Base derived objects (high $High).");
	}

	return $self;
}


=head1 METHODS

=head2 C<$base-E<62>err_log($message)>

B<$message> is a message to log. A trailing newline is unnecessary.

Prepended to the message on a separate line is:

 !$base->{id}! class::subroutine

Referring to the caller.

=cut
sub err_log {
	my $self = shift;
	my $msg = "!$self->{id}! ".(caller(1))[3]."\n".(shift);
	if (!$self->{logger}) {
		print STDERR "$msg\n";
		return;
	}
	$self->{logger}->out(LOG_ERR, $msg);
}


=head2 C<$base-E<62>debug_log($message)>

B<$message> is a message to log. A trailing newline is unnecessary.

Prepended to the message on a separate line is:

 {$base->{id}} class::subroutine

Referring to the caller -- unless, as per perldoc C<caller()>, "there is no caller",
i.e., this is a line in the main scipt outside any subroutine, in which case nothing
is prepended.

=cut
sub debug_log {
	my $self = shift;
	my $sub = (caller 1)[3];
	if (!defined $sub) { $sub = "" }
	else { $sub .= "\n" }
	my $msg = "{$self->{id}} $sub".(shift);
	if (!$self->{logger}) {
		print STDERR "$msg\n";
		return;
	}
	$self->{logger}->out(LOG_DEBUG, $msg);
}


=head2 C<$base-E<62>req_log($message)>

B<$message> is a message to log using C<Sloop::Logger::LOG_REQUEST>.  A
trailing newline is unnecessary.

Prepended to the message is "[$base->{id}] ".

=cut
sub req_log {
	my $self = shift;
	my $msg = "[$self->{id}] ".(shift);
	if (!$self->{logger}) {
		print STDOUT "$msg\n";
		return;
	}
	$self->{logger}->out(LOG_REQUEST, $msg);
}


=head2 C<$base-E<62>shut>

Abstract dummy; this should close the connection.

=cut
sub shut {
	my $self = shift;
	$self->dummy('shut', caller);
}


# Used to report dummy functions which have not been overridden.
sub dummy {
	my ($self, $sub) = (shift, shift);
	$_[1] =~ /\/([^\/]+)$/;
	my $msg = "\n$sub() not overridden in subclass of Sloop::Base.\n";
	$self->err_log($msg);
	die $msg;
}


=head2 C<$base-E<62>goodbye($msg, $notHere)>

B<$msg> is a text message.

B<$notHere> is optional.  If defined, the calling subroutine
will not be named.

Can be used to indicate fatal errors, as it calls croak().
The name of the calling subroutine is included in the output
by default; this can be suppressed with '$notHere'.

If called as an OO method, a copy of the message,
sans croak additions, is logged via $self->err_log.

=cut
sub goodbye {
# Check for OO invocation.
	my $self = undef;
	$self = shift if ref $_[0];
	my ($msg, $notHere) = (shift, shift);
	my $txt = "";
	if (!defined $notHere) {
		my @where = caller 1;
		$txt .= "$where[3]()"
	}
	$txt .= "!FATAL!\n* $msg *\n";
	$self->err_log($msg) if $self;
	croak($txt);
}


=head2 C<$base-E<62>okay>

As implemented here, will return a true value for core derived types such
as Sloop::Client and Sloop::Other until they are C<shut()> (which should
happen when rendered, or found to be, defunct by the server).  Otherwise
returns undef.

=cut
sub okay {
	return 1 if !$_[0]->{SHUT};
	return undef;
}


sub DESTROY {
	my $self = shift;
	$self->shut if ($self->{fd} > 0);
	$Count--;
	if ($self->{debug} & Sloop::Constants::DBG_DESTROY) {
		$self->debug_log("DESTROY $self->{id}, remaining $Count (high $High)");
	}
}

1;
