package Sloop::Client::Request;
our $VERSION = '0.1.2';
use 5.010000;
use strict;
use warnings;

# Version 0.1.1:
# - Corrected error in XS parse() method that dropped single character paths.
# - Corrected bug in query().
# Version 0.1.2:
# - Minor adaptations to accommodate handler array as short form for
#   Lib::checkURL().

use Sloop::Lib qw(dataLength);
use Sloop::Base;

require XSLoader;
XSLoader::load('Sloop::Client::Request', $VERSION);

=encoding utf8

=head1 NAME

Sloop::Client::Request

=head1 DESCRIPTION

Class representing an HTTP request from a client.
Objects of this sort are attached to Sloop::Client objects (as the 'request' field).

=head1 ATTRIBUTES

=over

=item C<$req-E<62>{body}>

Scalar containing the raw body of the request, if any.  This may or may not be complete;
the correct way to get the body is to use C<Sloop::Client-E<62>readPost()> or
C<Sloop::Client-E<62>simplePost()>.

If you do not read the entire body, do not bother trying to reply to the request.
Just cancel() the client object.

=item C<$req-E<62>{bad}>

If this exists, it is a string message explaining why the request header was rejected.

=item C<$req-E<62>{headers}>

Reference to a hash of the HTTP header. The keys are field names I<in all lower
case> and the values are the original strings. If the client used the same
header multiple times, the values are amalgamated into a comma seperated list in
one string. 

=item C<$req-E<62>{method}>

A string containing a lower case version of the http method, eg. "get" or "post".

=item C<$req-E<62>{path}>

The HTTP request path split into an array.  Eg. "GET /one/two/three.html" would
be ('one', 'two', 'three.html').

=item C<$req-E<62>{protocol}>

The raw text of the protocol, eg. "HTTP/1.1", or "???" if one was not found.

=item C<$req-E<62>{qargs}>

Does not exist by default; created via C<$req-E<62>qstr2qargs()> or
C<$req-E<62>post2qargs()>. A hash structure keyed by name; all elements are array
references, even if the field was intended to be singular.  This is to deal with
the fact that a request may, accidentally or on purpose, specify the field twice.
If you intended it to be a singular value, use: 

 if (exists $req->{qargs}->{field_name}) {
 	my $val = $req->{qargs}->{field_name}->[0];
 }

Presuming you have already called qstr2qargs() or post2qargs and checked the
return value (or made sure {qargs} is defined).  This will be safe regardless of
how malformed the request is. B<Note that $req-E<62>query() will do all that for
you.>

=item C<$req-E<62>{qstr}>

Reference to the string containing the raw text beyond '?' from the request path
(ie, the query string). Does not exist if there was no '?'.  This string is
destroyed by a $req->qstr2qargs() call.

=item C<$req-E<62>{slash_end}>

True/false value indicating whether the path part of the request-line (not including any
query string) ended in a forward slash.

=back

=cut
### NON-API MEMBER FIELDS ###

# $req->{level}
# Set during processing to one beyond the element index of the handler found and
# called in the the $req->{path} array.  This may be one beyond the end of the
# array. If a handler was not found, may be one beyond the first missing link, or
# the length of the array plus one.  See $req->cascade(). 

# $req->{post_handler}
# This is a subrountine reference called in $req->accumulatePost()
# whenever data is added to $req->{body}; it is assigned in $req->addPost().

# $req->{post_len}
# An integer number of bytes taken from the 'Content-Length' header, if
# applicable.  As the post body is read, this number is decremented.


sub new {
	my ($class, $data) = (shift, shift);
	my $self = parse($$data, dataLength($data));

# HTTP pipelining means another request could be right after this one;
# if so pass that back via the third (scalar ref) arg.
	if (defined $self->{body} && !$self->{headers}->{'content-length'}) {
		my $excessBuffer = shift;
		$$excessBuffer = $self->{body};
		$self->{body} = "";
	}

	$self->{level} = 0;
	$self->{qargs} = undef;

	bless $self, $class;
}


=head1 METHODS

=head2 C<$req-E<62>findCookieValue($name)>

Convenience method for getting the value of the cookie named $name.

B<Returns:>  The value or undef if not found (or malformed). If the value 
is undefined (i.e. $name is a cookie with nothing after the =), an empty string
is returned.

=cut

sub findCookieValue {
	my ($self, $name) = @_;
	my $cookieLine = $self->header('cookie');
	return undef if !$cookieLine;
	# See RFC 6265 and 2616 sec 2.2.
	my $rgx = qr/^([^\(\)<>@,;:\\"\/\[\]\?=\{\}\s]+?)="?([^\s\/,;\"]*)"?/;
	foreach (split /\s*,\s*/, $cookieLine) {
		if ($_ =~ /$rgx/ && $1 eq $name) { return $2 }
	}
	return undef;
}


=head2 C<$req-E<62>header($name)>

Retrieves a header value.

B<$name> is the (all lowercase) name of the HTTP header, eg. "user-agent",
"content-length".

B<Returns:> A text value for the header, if any.

=cut
sub header {
	my $self = shift;
	return $self->{headers}->{(shift)};
}


=head2 C<$req-E<62>hostURI()>

B<Returns:>  A string such as 'http://localhost'.  The protocol is taken from $ENV{SLOOP_PROTOCOL}
and the host from the 'Host:' header, or, if it does not exist, from the `hostname` command.  If
that fails, undef is returned.

=cut
sub hostURI {
	my $hoststr = $ENV{SLOOP_PROTOCOL};
	if (defined $_[0]->{headers}->{host}) { $hoststr .= $_[0]->{headers}->{host} }
	else { return undef if (!eval { $hoststr .= `hostname` }) }
	return $hoststr;
}


=head2 C<$req-E<62>pathFromLevel()>

Retrieves any remainder in $req->{path} beyond the level
of the handler called.  Eg, if $req->{path} is ('one', 'two', 'three.html')
and 'two' triggered a handler, pathFromLevel() will return ('three.html').

B<Returns:> An array; if there is no remainder, the array is empty.

=cut
sub pathFromLevel {
	my $self = shift;
	my $end = $#{$self->{path}};
	if ($self->{level} > $end) {
		my @empty = ();
		return @empty;
	}
	return @{$self->{path}}[$self->{level}..$end];
}


=head2 C<$req-E<62>post2qargs($decode)>

Parses the contents of $req->{body} into $req->{qargs} (see ATTRIBUTES).
Destructive effect on {body}.

B<$decode> is a boolean indicatator of whether or not $req->{body} is URL encoded.

B<Returns:> The $req->{qargs} reference, or undef if there was no body to parse.

=cut
sub post2qargs {
	my $self = shift;
	return undef if (!defined $self->{body});

	$self->{qargs} = qargsFromString($self->{body}, shift);
	return $self->{qargs};
}


=head2 C<$req-E<62>qstr2qargs()>

Parses the contents of $req->{qstr} into $req->{qargs} (see ATTRIBUTES).
Destructive effect on {qstr}.

It's assumed the query string is URL encoded, and it will be decoded as
such.

B<Returns:> The $req->{qargs} reference, or undef if there was no qstr to parse.
A malformed query string which did not lead to an error when the HTTP header was
initially parsed may lead to an empty/partial object but is not explicitly
indicated.

=cut
sub qstr2qargs {
	my $self = shift;
	return undef if (!defined $self->{qstr});

	$self->{qargs} = qargsFromString(${$self->{qstr}}, 1);
	return $self->{qargs};
}


=head2 C<$req-E<62>query(@keys)>

Retrieves the value of a query field, or a number of query fields if more are
passed in. More specifically, the value is the first value in
the array of values for that field (see 'qargs' under SYNOPSIS, above, for an
explanation).  If qstr2qargs() or post2qargs() have yet been called, the former is
tried first, then if there is no query string and C<$req-E<62>{method}> is 'post', the
latter.

If you use this method and the query is in a POST body, it will be treated as
URL encoded unless you call post2qargs() appropriately yourself first. 

B<@keys> is one or more query keys to search for.

B<Returns:> If there was only one key in the argument list, returns that value. 
If there were more, returns an array of values.  Fields which were not found are
undef.  If a field has multiple values, the first one is returned. 

=cut
sub query {
	my $self = shift;
	if (!defined $self->{qargs}) {
		if (!$self->qstr2qargs && $self->{method} eq 'post') {
			return undef if (!$self->post2qargs(1));
		}
		return undef if (!defined $self->{qargs});
	}
	my @vals;
	foreach (@_) {
		if (!exists $self->{qargs}->{$_}) { push @vals, undef }
		else { push @vals, $self->{qargs}->{$_}->[0]; }
	}
	return $vals[0] if $#vals == 0;
	return @vals;
}


=head2 C<$req-E<62>requestPath()>

B<Returns:>  A reconstruction of the path from the request's request-line (the original
was broken down during parsing).

=cut
sub requestPath {
	return '/' if ($#{$_[0]->{path}} == -1);
	my $rqp = '/'.join('/', @{$_[0]->{path}});
	$rqp .= '/' if $_[0]->{slash_end};
	return $rqp;
}


### NON-API METHODS ###

# Called from Sloop::Client::continuePost().  The arguments are
# the length of the data to add and a scalar reference to the data.
# Returns 0 if more data is needed to complete the post body, or
# 1 if the body is complete.
sub accumulatePost {
	my $self = shift;
	$self->{body} .= ${$_[0]};
	$self->{post_len} -= pop;
	if ($self->{post_len} <= 0) {
		$self->{post_handler}(1);
		return 1;
	}
	else {
		$self->{post_handler}(0);
		return 0;
	}
}


# Called from Sloop::Client::readPost().
# Returns
#   <0 if the connection should abort
#    0 if the post should continue
#    1 if the post is already complete
sub addPost {
	my $self = shift;
	$self->{post_handler} = shift;
	$self->{post_len} = $self->header('content-length');

	if (!defined $self->{post_len} || !Sloop::Lib::toInteger($self->{post_len})) {
	# Nb. toInteger() sets the field value type to IV for subsequent use in C.
		$self->{post_handler}(-1);
		return -1;
	}

	my $max = shift;
	if (!defined $max) {
		Sloop::Base::goodbye("The max length for a POST body was undefined; "
			."perhaps you forgot the last argument to "
			."Sloop::Client::readPost() or Sloop::Client::simplePost().")
	}
	if ($self->{post_len} > $max) {
		$self->{post_handler}(-1);
		return -2;
	}

	if (!exists $self->{body}) {
		$self->{body} = "";
	} else {
		$self->{post_len} -= dataLength(\$self->{body});
		if ($self->{post_len} <= 0) {
			$self->{post_handler}(1);
			return 1;
		} else { $self->{post_handler}(0) }
	}
	return 0;
}


# Called from Sloop::Server::processRequest(); finds the appropriate
# content handler for the request and returns it.  Argument is a 
# reference to a handler tree in hash form.
sub cascade {
	my ($self, $handlers) = (shift, shift);
	my $max = $#{$self->{path}};

	while ($self->{level} <= $max) {
		my $cur = $self->{path}->[$self->{level}];
		last if (!exists $handlers->{$cur});
		my $ref = ref($handlers->{$cur});
		if ($ref eq 'CODE' || $ref eq 'ARRAY') {
			$self->{level}++;
			return $handlers->{$cur};
		}
		$handlers = $handlers->{$cur};
		$self->{level}++;
	}

	if (exists $handlers->{'/'}) {
		return $handlers->{'/'};
	}

	return undef;
}


=head2 C<getHeaders($dataref)>

This is a static (i.e., not OO) function that wraps part of the internal HTTP
parser. It's included in the API since it might be useful with Sloop::Other and
C<Sloop::Multipart::process()>. 

B<$dataref> is a scalar reference containing raw HTTP headers.  I<It must be
checked for "\r\n\r\n" termination first or bad things may happen.>
It should also be free from anything that's not a "key: value" pair,
such as the preliminary status line in an HTTP response ("HTTP 1.1 OK", etc.),
or this will count as malformed and the return value will be undef.

B<Returns:> A hash containing the headers as key value pairs, with the key in all
lowercase, is returned, or undef if $dataref is not a reference or 
contains malformed headers.

If the call succeeds, $dataref now refers to whatever starts after the
terminating "/r/n/r/n".

=cut

1;
