package Sloop::Client;
our $VERSION = '0.1.0';
use strict;
use warnings;

use base 'Sloop::Connection';

use Sloop::Client::Request;
use Sloop::Constants qw(:PollEvents :Debug :ClientState);
use Sloop::Lib qw(callerTag dataLength lowerCase);
use Sloop::Logger;

# Situation constants
# CLIENT_CANCELLED The connection is shut if this is set, see cancel().
# CLIENT_WAITING   Initial state, and (for persistent clients), the state after
#                  a request cycle is completed.
# CLIENT_REQUEST   Bitwise includes CLIENT_WAITING; set in takeRequest().
# CLIENT_POST      Set while a POST request body is being read.
# CLIENT_REPLY     Set in reply() to differentiate from other states.
# CLIENT_ALT       Intended for use by clients which have upgraded from the
#                  HTTP protocol (eg., to WS); this will supplant takeRequest()
#                  with a {ALT_handler}->(), which should be defined for the client.
# CLIENT_SPECIAL   Not used by Client class, and guaranteed to be the highest
#                  CLIENT_ constant in Sloop::Client. This is intended for use
#                  by subclasses that override ready(), such as Sloop::Client::Secure.
use constant {
	CLIENT_CANCELLED => 0,
	CLIENT_WAITING => 1,	# NOTE: REQUEST includes WAITING for bitwise AND 
	CLIENT_REQUEST => 3,	# (thus there is no value for 2)
	CLIENT_POST => 4,
	CLIENT_REPLY => 8,
	CLIENT_ALT => 16,
	CLIENT_SPECIAL => 32
};


# These are used in reply().
our %HTTP_STATUS = (
	100 => '100 Continue',
	101 => '101 Switching Protocols',
	200 => '200 OK',
	201 => '201 Created',
	202 => '202 Accepted',
	203 => '203 Non-Authoritative Information',
	204 => '204 No Content',
	205 => '205 Reset Content',
	206 => '206 Partial Content',
	300 => '300 Multiple Choices',
	301 => '301 Moved Permanently',
	302 => '302 Found',
	303 => '303 Use Other',
	304 => '304 Not Modified',
	305 => '305 Use Proxy',
	307 => '307 Temporary Redirect',
	400 => '400 Bad Request',
	401 => '401 Unauthorized',
	402 => '402 Payment Required',
	403 => '403 Forbidden',
	404 => '404 Not Found',
	405 => '405 Method Not Allowed',
	406 => '406 Not Acceptable',
	408 => '408 Request Timeout',
	409 => '409 Conflict',
	410 => '410 Gone',
	411 => '411 Length Required',
	412 => '412 Precondition Failed',
	413 => '413 Request Entity Too Large',
	414 => '414 Request-URI Too Long',
	415 => '415 Unsupported Media Type',
	416 => '416 Request Range Not Satisfiable',
	417 => '417 Expectation Failed',
	500 => '500 Internal Server Error',
	501 => '501 Not Implemeneted',
	503 => '503 Service Unavailable',
	505 => '505 HTTP Version Not Supported'
);


=encoding utf8

=head1 NAME

Sloop::Client

=head1 DESCRIPTION

Sloop::Client represents a connection to a remote HTTP client; such a client
may have multiple such connections simultaneously, each via a separate
Sloop::Client object.

The class adds HTTP related functionality to base class Sloop::Connection. 
It still does not include the low level I/O routines necessary for a functioning
socket, and so a Sloop::Client object by itself is not useful.  Instead,
Sloop::Client is a mutual superclass, together with  one of the Sloop::Socket::*
classes, of the Sloop::Client::Regular and ::Secure subclasses.
However, all of the API methods are here or in one of its superclasses.

=head1 SYNOPSIS

Note that Sloop::Client objects should not be created by the user, they are
received in a C<$sloop-E<62>{handler}> callback.  However, the defaults used may
be overridden via  C<$sloop-E<62>{client_constructor}> (see Sloop::Server
SYNOPSIS).  Note that although some "REQUIRED" parameters must be supplied, they
are supplied by the server, so there are no requirements for client_constructor
as long as you pass the argument stack through. 

 my $client = Sloop::Client->new (
    debug => (optional) see Sloop::Base
    fault_max => (optional) see Sloop::Connection
    fd => (optional) see Sloop::Base
    generic => (optional) hash ref conforming to $Sloop::Generic::Pages
    id => (optional) see Sloop::Base
    ip => (REQUIRED) string
    logger => (REQUIRED) see Sloop::Base
    logRequest => (optional) sub ref
    maxage => (optional) integer
    max_recv => (optional) see Sloop::Connection
    max_send => (optional) see Sloop::Connection
    timeout => (optional) see Sloop::Connection
    postRequestHandler => (optional) sub ref
 );

Parameters not documented in a base class are described below.

=over

=item B<generic>

A hash ref matching the specifications of C<$Sloop::Generic::Pages>.  If not
explicit, the default will be used and Sloop::Generic included via C<require>. 
This defines the html accompaniment for generic status based responses
(e.g., 404 "Not Found").

=item B<logRequest>

A subrountine reference called after a request header has been successfully read
and parsed.  This function does not have to do anything; the default
C<$client-E<62>logReq()> is documented under METHODS below.  Note that if you override
C<$client-E<62>{logger}>, you may want to override this as well, since logReq() calls
the logger.

=item B<ip>

The IP address of the remote party.

=item B<maxage>

Value in seconds to use with 'Cache-Control: max-age=' in HTTP response headers.  If
zero, an 'Expires' header in the past is added to deal with old non-compliant browsers.
See also C<$client-E<62>reply()> under METHODS.

=item B<postRequestHandler>

If defined, this is called by the server after a request header is parsed.  It
receives one argument, a Sloop::Client object, to which a Sloop::Client::Request
will be attached. 

If the handler returns a true value, the request handling continues normally. 
In this case, you should NOT have sent any response or closed the connection. 

If the handler returns false, the handling simply stops.  In this case, you
SHOULD have either replied -- via e.g., reply() or redirect() -- or else closed
the connection with cancel(). If not, the connection will be left in a limbo
state which does not time out or allow for further communication (i.e.,
essentially leaked until the other side closes). 

If you wish to read POST data in the request, you can do that too -- return
false and send a reply in the post handler. 

You can only set one primary such handler, but you can chain them together, e.g.:

 my $existingHandler = $client->{postRequestHandler};
 $client->{postRequestHandler} = sub {
     my $client = shift;
     # Do whatever.
     if ($existingHandler) { return $existingHandler($client) }
 }

=back

=head1 ATTRIBUTES

In addition to attributes inherited from Sloop::Base or Sloop::Connection,
the following fields from the constructor are accessible with the same name
and meaning:

=over

=item generic

=item logRequest

=item ip

=item maxage

=item postRequestHandler

=back

There are a few potential post construction attributes as well:

=over

=item C<$client-E<62>{request}>

If defined, a Sloop::Client::Request object.  If not, there is no request pending.

=item C<$client-E<62>{afterReplyHandler}>

If defined, this was set via C<$client-E<62>setAfterReplyHandler()> (see below).
This is useful if you wish to chain such handlers.

=back

=cut

### NON-API OBJECT MEMBER FIELDS ###

# $client->{situation} 
# One of the "situation constants" detailed above.

# $client->{raw_head}
# Used in $client->takeRequest().

# $connection->{persist}
# A boolean value for HTTP clients indicating whether the connection should be
# kept alive after a response is sent.  Set in takeRequest() and tested in
# doneWrite().

# $client->{startTime}
# Used in the checktime() method.

# $client->{ALT_handler}
# Used for non-HTTP transactions by (eg) Sloop::WebSocket. Qv. ready().
# Receives $client as the first arg.

# $client->{pipeline}
# Buffers pipelined requests.

sub new {
	my $class = shift;
	my %args = (
		maxage => 0,
		timeout => 32,
		logRequest => \&logReq,
		@_,
		situation => CLIENT_WAITING,
	);
	if (!defined $args{generic}) {
		require Sloop::Generic;
		$args{generic} = $Sloop::Generic::Pages;
	}

	my $self = $class->SUPER::new ( %args );

	$self->{raw_head} = "";
	$self->{pipeline} = "";
	$self->{startTime} = time;

	bless $self, $class;
}


=head1 METHODS

=head2 C<$client-E<62>cancel()>

You should call this if for whatever reason you don't want to reply to a client.
Do NOT do this if you have sent a reply, at least until you are sure it has been
completely transmitted (see C<$client-E<62>setAfterReplyHandler()>, below).

This immediately closes the connection and flags the object such that all
internal references to $client will be removed at the next opportunity.

=cut
### Below is a description of an old problem worth remembering. ###
###################################################################
# The circularity problem occurs in, eg., beginPost() handlers, which contain a
# reference to $client:
# $Sloop->{handlers} = {
#	testpost => sub {
#		my $client = shift;
#		$Sloop->beginPost($client, sub {
#			my $complete = shift;
#			return if ($complete <= 0);
#			$client->reply("whatever");
#		}, 1024);
#	},
# }
#
# The inner anonymous subrountine is passed through to
# $client->{request}->addPost() via $Sloop->beginPost(). Normally, $client->reply()
# deletes the request.  However, if the post fails, this is never called, meaning
# there remains a (circular) reference to $client in $client->{request}->{post_handler}.
sub cancel {
#my @who = caller 1;
#print "$who[0]::$who[3].$who[2]\n";
	my $self = shift;
	$self->SUPER::cancel();
	$self->{situation} = CLIENT_CANCELLED;
	delete($self->{request}) if (exists $self->{request});
}


=head2 C<$client-E<62>generic($response_code)>

B<$response_code> is an HTTP response code, eg, 500 or 404.

Error checked wrapper for $client->{generic}->{$response_code}->($client),
sending an HTTP response with the corresponding response code. The default
Sloop::Generic module includes a simple html page for most common simple error codes.
If a generic page corresponding to $response_code does not exist, an even simpler page
displaying just the number is sent.

=cut
sub generic {
	my ($self, $code) = (shift, shift);
	if (!exists $self->{generic}->{$code}) {
		$self->err_log("No generic page defined for '$code'.");
		$self->reply (
			\"<html><body><h1>$code</h1></body></html>", 
			status => $code
		);
		return;
	}
	$self->{generic}->{$code}($self);
}


=head2 C<$client-E<62>logReq()>

This is the default assigned to the subroutine reference $client->{logRequest} in the 
constructor (see SYNOPSIS).  It prints a message via $client->{logger} as a LOG_REQUEST
(see Sloop::Constants) with the following format:

 [$client->{id}] path ip user-agent
 
Where "path" is from C<$client-E<62>{request}-E<62>requestPath>, "ip" from
C<$client-E<62>{ip}>, and "user-agent" from
C<$client-E<62>{request}-E<62>header('user-agent')>; the last field is truncated
to 32 characters.  This function needn't be called manually, and is documented
only because of its role as a default for $client->{logRequest}. 

=cut
sub logReq {
	my $self = shift;
	my $msg = $self->{request}->requestPath;
	$msg .= " $self->{ip}";
	if (my $ua = $self->{request}->header('user-agent')) {
		$msg .= ' '.substr($ua, 0, 32);
	}
	$self->{logger}->out(LOG_REQUEST, "[$self->{id}] ".$msg)
}


=head2 C<$client-E<62>readPost($handler, $max)>

An alternative to $client->simplePost() which fires a callback
for each chunk of post data read in.  This is necessary when 
using Sloop::Multipart.

B<$handler> is a callback that will recieve an integer arg, 1 if
the post is complete, 0 if it is not, and -1 if it and $client have been
cancelled due to a logged error.  The accumulated post is always
available in $client->{request}->{body}.  You may manipulate that
data at any point, as long as {body} is not left undefined (or
redefined as something other than a scalar).

B<$max> is the same as described for simplePost().

B<Returns> -1 on error; this will be logged and C<$client-E<62>cancel> called.
Otherwise 0.

=cut
sub readPost {
	my $self = shift;
	my $state = $self->{request}->addPost(@_);
# Done.
	return 0 if $state == 1;
# Continue.
	if (!$state) {
		if ($self->setEvents(SPFD_READ)) {
			$self->err_log("setEvents() failed: $!");
			$self->cancel;
			return -1;
		}
		$self->setSituation(CLIENT_POST);
		return 0;
	}
# Error.
	$self->cancel;
	my $msg = "";
	if ($state == -1) { $msg = "Improper Content-Length header." }
	else { $msg = "Content-Length exceeded specified maximum." }
	$self->err_log($msg);
	return -1;
}


=head2 C<$client-E<62>redirect($url, $note)>

Sends the client an HTTP 302 Redirect message.  This substitutes for a
$client->reply() and either one or the other (not both) should be called per
request.  As per the HTTP 1.1 protocol, it includes a body with a brief note for
clients that do not redirect automatically.  By default this is "You have been
redirected, click here.", unless $note is defined, in which case that is used.

B<$url> is an I<absolute> URL (eg, "http://wherever") for use in the HTTP 'Location:'
header and the "click here" href.

B<$note> is the optional HTML body.

You can send a 301, 303, etc. redirect by using the appropriate %args with
$client->reply().   As with all the Sloop::Client methods, this is a non-blocking
call.  It queues data to be sent, it does not wait for its transmission.

=cut

sub redirect {
	my ($self, $url, $body) = (shift, shift, shift);
	if (!defined $body) {
		$body = "<p>You have been redirected, <a href=\"$url\">click here</a>.";
	}

	$self->reply (\$body,
		extra => {Location => $url},
		status => 302
	);
}


=head2 C<$client-E<62>reply($body, %args)>

Sends an HTTP response to the client, including a user supplied
message body.  As with all the Sloop::Client methods, this is a
non-blocking call.  It queues data to be sent, it does not wait
for its transmission.

An example:

 $client->reply (
     \$data,
     type => 'application/octet-stream',
     maxage => 3600
 );

B<$body> is scalar reference to the response body.  It may be undefined, in
which case no body is sent. B<WARNING:> If this argument is defined but not a
I<reference>, the call is FATAL.  If it is not a I<scalar> reference, the
behavior is undefined. The content is copied out during the call without
modification, so you can do what you like with it afterward.

B<%args> are optional; the defaults are given below:

=over

=item C<status =E<62> 200>

An HTTP status-code; it may be any of the codes from RFC 2616 section 10.  This
list is also available via C<%Sloop::Client::HTTP_STATUS>.  You can add to those
if you want to create a custom code, although for a general purpose web server
this is a bad idea as it violates the HTTP 1.1 protocol. 

=item C<maxage =E<62> $self-E<62>{maxage}>

This can be used to  override the value set for 'Cache-control: maxage' and
'Expires:' (see  "Auto-generated Headers", below).  It is in seconds and the
default is 0.  You can override this by defining a custom C<client_constructor>.
The reason the global default is 0 and cannot be changed more simply is that you
should ideally set it in the handler reply using this parameter.  If you have a
collection of static pages, this is easily done using
C<Sloop::Static::directory()>; if the content is dynamic, you should probably
stick with zero. 

=item C<type =E<62> 'text/html; charset=UTF-8'>

This is only included if $body is defined. It sets the 'Content-type' header
(see "Auto-generated Headers", below).

=item C<length =E<62>> I<bytes>

This will override the 'Content-length' header (see "Auto-generated Headers",
below). You should not normally use this as an appropriate value will be
determined.  It will only be included if C<$body> is defined. 

=item C<cookies =E<62> undef>

An array ref containing strings to use with a 'Set-Cookie' header; each string gets
its own header.  C<Sloop::Lib::makeCookie()> can be used to create such a string.

=item C<extra =E<62> undef>

A hashref containing HTTP headers to add as key (header name)
value pairs.  Eg:

 extra => {
	'Accept-Language' => 'en_CA',
	'Max-Forward' => 2
 }

These will B<not> replace any header of the same name added by sloop; their
values will be amalgamated into a comma seperated list. One way to examine the
header as sent is to use DBG_REPLY from Sloop::Constants.  

=back

B<Auto-generated Headers>

Sloop adds appropriate values for the following headers to all responses: 

 Cache-Control
 Server

"Cache-Control" uses a "max-age" as described in SYNOPSIS; if this is zero an
"Expires" header is also used.  "Server" is "Perl Sloop".

Responses with a body also have a "Content-Length", and "Content-Type" (defaults
to "text/html; charset=UTF-8") header.  If you want to override "Content-Type",
use C<$args-E<62>{type}>, above.   Non-persistent connections (e.g., for HTTP 1.0
clients) contain a "Connection: close" header.    

=cut
sub reply {
	my $self = shift;
# Validation.
	$self->goodbye("Uneven hash args.") if ($#_ % 2 != 0);
	my $body = shift;
	if (defined $body) {
		$self->goodbye("'\$body' arg was not a scalar reference!")
			if (ref($body) ne 'SCALAR');
		$body = undef if !$$body;
	}

	if (!exists $self->{request}) {
		my $tag = callerTag(__PACKAGE__);
		$self->err_log("Reply without a request, cancelled.\n$tag");
		return;
	}

# Defaults.
	my %args = (
		extra => undef,
		maxage => $self->{maxage},
		status => 200,
		type => 'text/html; charset=UTF-8',
		@_
	);

# Create the response string.
	if (!defined $Sloop::Client::HTTP_STATUS{$args{status}}) {
		$self->goodbye("Invalid status code '$args{status}'.");
	}
	my $reply = <<END;
HTTP/1.1 $Sloop::Client::HTTP_STATUS{$args{status}}\r
Server: Perl Sloop\r
Cache-Control: max-age=$args{maxage}\r
END

	# Compensate for IE 7/8 WRT Cache-Control.
	$reply .= "Expires: Sat, 1 Jan 2000 4:20:20 GMT\r\n" if ($args{maxage} == 0);

	# Content-Length/Type
	if (defined $body) {
		my $len;
		if (defined $args{length}) {
			$len = $args{length};
			if ($len < 0) {
				$self->err_log (
					'Negative \'length\' argument, sending "500 Server Error".'
				);
				return;
			}
		} else { $len = dataLength($body) }
		if ($len < 0) {
			$self->err_log (
				"dataLength() returned $len..."
				.'bad reply body? Sending "500 Server Error".'
			);
			$self->generic(500);
			return;
		}
		$reply .= "Content-Length: $len\r\nContent-Type: $args{type}\r\n";
	}

	# Non-persistent connections
	$reply .= "Connection: close\r\n" if (!$self->{persist});

	# Add user defined header values.
	if (defined $args{extra}) {
		while (my ($k, $v) = each (%{$args{extra}})) {
			$reply .= "$k: $v\r\n"
		}
	}

	# Add cookies.
	if (defined $args{cookies}) {
		$reply .= "Set-Cookie: $_\r\n" foreach (@{$args{cookies}})
	}

	# Terminate response header.
	$reply .= "\r\n";

# Report header if required.
	if ($self->{debug} & DBG_REPLY) {
		$self->debug_log("Response header:\n$reply");
	}

# Add to data_pending.
	if (defined $self->{request}->{method}
	  && $self->{request}->{method} ne 'head' && defined $body) {
		$reply .= $$body;
	}
	$self->{data_pending} = \$reply;

# Set events and situation and remove request object.
	if (my $err = $self->setEvents(SPFD_WRITE)) {
		$self->err_log("setEvents() failed, errno=$err");
		$self->cancel;
		return READY_CANCEL;
	}
	$self->setSituation(CLIENT_REPLY);
	delete($self->{request});
}


=head2 C<$client-E<62>setAfterReplyHandler($coderef)>

Sets a callback that will be fired once a message sent via reply() has
successfully completed transmission.   For non-persistent connections, this is
I<immediately before> the connection will be closed.  For persistent connections,
it will be before any subsequent request is read. If sending the reply failed,
the callback will not be fired and the connection will be closed.  

B<$coderef> is a subroutine reference.  It receives one argument, the client
object on which it was set. 

If the subroutine returns 0 or undef, it will only fire once.  If you wish to
set a different handler, return a reference to it.  If the return value is
defined but not a reference, the same handler will be left attached.  Do not try
and call C<setAfterReplyHandler()> itself inside the handler. 

There can only be one such handler set at a time, but as described in the
SYNOPSIS with regard to C<postRequestHandler>, they may be chained together. 
Hence, it never hurts to check; the relevant member field is
C<$client-E<62>{afterReplyHandler}>. 

=cut
sub setAfterReplyHandler { $_[0]->{afterReplyHandler} = $_[1] }


=head2 C<$client-E<62>simplePost($handler, $max)>

Wrapper for readPost() which skips reporting each chunk as it is recieved.
If reading the POST fails, an error is logged but no callback is fired;
$client will be left defunct.  If C<$client-E<62>{request}> is not a POST, $client
is sent a generic "405 Method Not Allowed" response and simplePost returns
-1.

B<$handler> is a subroutine reference.  It will fire when reading the POST
data has successfully completed.

B<$max> is the maximum value to accept in the Content-Length header; if
this is exceeded, $client will be cancelled.  Note that if a client sends
data in excess of the Content-Length, that data will be left in the request
pipeline and cause the client to be subsequently deleted.
Clients which send less than the Content-Length will never have a completed
post, and the data will be lost on a subsequent request in the same pipeline.

B<Returns> The same thing as readPost().

=cut
sub simplePost {
	my ($self, $handler) = (shift, shift);

	if ($self->{request}->{method} ne 'post') {
		Sloop::Generic::page($self, 405,
			uc($self->{request}->{method}).' not allowed.',
			{ Allow => 'POST' }
		);
		$self->err_log("POST handler recieved a '".$self->{request}->{method}."' request.");
		return -1
	}
	
	return $self->readPost(sub {
		return if $_[0] != 1;
		$handler->();
	}, pop)
}


# Called from $client->ready() or $client->continueRead().
# Returns:
#   READY_DONE if the post is complete
#   READY_CANCEL on error
#   READY_CONTINUE otherwise
sub continuePost {
	my $self = shift;
	my $bytes;
	my $dataref = $self->readData(\$bytes);
	if (!defined $dataref) {
		return READY_CANCEL # fatal to $client
			if ($self->{situation} == CLIENT_CANCELLED);
		return READY_CONTINUE;
	}

	if ($self->{request}->accumulatePost($dataref, $bytes)) {
		return READY_DONE;
	}

	return READY_CONTINUE;
}


# Fulfills Sloop::Connection
sub doneWrite {
	my $self = shift;

	if (my $err = $self->setEvents(SPFD_READ)) {
		$self->err_log("setEvents() failed, errno=$err");
		$self->cancel;
		return -1;
	}

	$self->setSituation(CLIENT_WAITING) if $self->{situation} != CLIENT_ALT;

	if ($self->{debug} & DBG_REPLY) {
		$self->debug_log("Reply completed.");
	}

	if (defined $self->{afterReplyHandler}) {
		my $check = $self->{afterReplyHandler}->($self);
		if (!$check) {
			$self->{afterReplyHandler} = undef;
		} elsif (ref $check) {
			$self->{afterReplyHandler} = $check;
		}
	}

	if (!$self->{persist}) {
		$self->cancel;
		return -1;
	}

# Reset timeout.
	$self->{startTime} = time;
	return 2 if $self->{pipeline};
	return 1;
}


# Fulfills Sloop::Connection.
sub ready {
	my $self = shift;
	# read event was turned off if request is complete, so using AND here
	# should only apply to incomplete requests (see takeRequest())
	if ($self->{situation} & CLIENT_WAITING) { return $self->takeRequest() }
	elsif ($self->{situation} == CLIENT_POST) { return $self->continuePost() }
	elsif ($self->{situation} == CLIENT_CANCELLED) { return READY_CANCEL }
	elsif ($self->{situation} == CLIENT_ALT) { return $self->{ALT_handler}->($self) }
	else { return READY_DONE }
}


# Takes a time in seconds since the epoch and returns:
#   0 if the connection is still okay
#   1 if the connection is now defunct
sub checktime {
	my $self = shift;
	return 1 if $self->{situation} == CLIENT_CANCELLED;
	return 0 if (
		$self->{situation} != CLIENT_WAITING
		|| !$self->{timeout}
		|| $self->{startTime} >= (shift) - $self->{timeout}
	);

	if ($self->peekInSocket() > 0) {
		$self->err_log("Timed out with data in recv buffer!");
	}

	$self->cancel;

	if ($self->{debug} & DBG_CONNECTION) {
		$self->debug_log("Timed out.");
	}

	return 1;
}


# Argument is one of the CLIENT_ "Situation constants"
# described at the top of this file.
sub setSituation {
	$_[0]->{situation} = $_[1];
}


# Called from $client->ready() to read and parse a new HTTP request.
# Returns one of the 'Client Ready' constants from Sloop::Constants.
sub takeRequest {
## TEST WITH: NO PATH, NO METHOD, GET/?query ##
	my $self = shift;

	$self->{situation} = CLIENT_REQUEST;

# Read in.
	if ($self->{pipeline}) {
	# If there's data in {pipeline}, we weren't called by epoll() and
	# a read call may return zero although the client is still connected,
	# so rather than sorting this out (eg, with peekInSocket), just leave
	# the socket as is for now.
		$self->{raw_head} .= $self->{pipeline};
		$self->{pipeline} = "";
	} else {
		my $raw = $self->readData();
		if (!defined $raw) {
		# disconnected or error
			if ($self->{situation} == CLIENT_CANCELLED) {
				return READY_CANCEL;
			}
		# EAGAIN
			return READY_CONTINUE;
		}
		$self->{raw_head} .= $$raw;
	}

# Determine if request header is complete.
	# Could optimize this by starting from the back w/ C.
	my $end = index($self->{raw_head}, "\r\n\r\n");

	if ($end < 0) {
		if (dataLength(\$self->{raw_head}) >= $self->{max_recv}) {
			$self->err_log("Header exceeded $self->{max_recv} bytes, disconnecting.");
			$self->cancel;
			return READY_CANCEL;
		}
		return READY_CONTINUE;
	}

# Parse header.
	if ($self->{debug} & DBG_REQUEST) {
		$self->debug_log("Request header:\n".substr($self->{raw_head}."\n", 0, $end));
	}

	# If there are pipelined requests, those will end up in {pipeline}.
	$self->{request} = Sloop::Client::Request->new (
		\$self->{raw_head},
		\$self->{pipeline}
	);

	if (exists $self->{request}->{bad}) {
		$self->err_log(
			"Header rejected ('$self->{request}->{bad}'), disconnecting"
		);
		$self->cancel;
		return READY_CANCEL;
	}

	$self->{raw_head} = "";

# Set persist.
	if ($self->{request}->{protocol} eq 'HTTP/1.1') {
		$self->{persist} = 1;
	} else { $self->{persist} = 0 }
	# 'Connection:' header can override HTTP/1.1 default persistance.
	my $conhead = $self->{request}->header('connection');
	if ($conhead) {
		lowerCase($conhead);
		if ($conhead eq 'keep-alive') { $self->{persist} = 1 }
		elsif ($conhead eq 'close') { $self->{persist} = 0 }
	}

# Log.
	$self->{logRequest}($self);

# Turn read off until the request is finished.
	if (my $err = $self->setEvents(0)) {
		$self->err_log("setEvents() failed, errno=$err");
		$self->cancel;
		return READY_CANCEL;
	}

	return READY_REQUEST;
}


1;
