package Sloop::Client::Secure;
our $VERSION = '0.1.0';
use strict;
use warnings;

use parent qw(Sloop::Socket::TLS Sloop::Client);

=encoding utf8

=head1 NAME

Sloop::Client::Secure

=head1 DESCRIPTION

Used by Sloop::Server when run in secure HTTPS mode to create objects representing
accepted client connections.  Conjoins the functionality of its parents,
Sloop::Client and Sloop::Socket::TLS.

=head1 SYNOPSIS

Invocation is the same as for the Sloop::Client superclass (although sloop
client objects should not be created except by the server).

Sloop::Client::Secure sets the 'id' field inherited from Sloop::Base
to 'fdN', where N is the file descriptor number.

From a user perspective, there isn't any difference between this and
Sloop::Client::Regular; the entire API is inherited from Sloop::Client.  You can
switch from to secure to unsecure mode without altering any code beyond the
Sloop::Server constructor. 

=cut

sub new {
	my $class = shift;
	my %args = ( @_ );

	my $self = $class->SUPER::new (
		id => "fd$args{fd}",
		%args,
	);
	bless $self, $class;

	if (!$self->serverSession) {
		$self->err_log($self->{tls_error});
		return undef;
	}

	my $check = $self->shakehands;
	if ($check == -1) {
		$self->err_log($self->{tls_error});
		$self->cancel;
		return undef;
	} elsif ($check == 0) { $self->{situation} = Sloop::Client::CLIENT_SPECIAL }

	return $self;
}


sub ready {
	my $self = shift;
	return $self->SUPER::ready() if $self->{situation} != Sloop::Client::CLIENT_SPECIAL;

# TLS handshake isn't complete, try again.
	my $check = $self->shakehands;
	if ($check == -1) {
		$self->err_log($self->{tls_error});
		$self->cancel;
		return Sloop::Constants::READY_CANCEL;
	} elsif ($check == 0) {
		return Sloop::Constants::READY_CONTINUE;
	}

# Handshake is now complete!
	$self->{situation} = Sloop::Client::CLIENT_WAITING;
	return $self->SUPER::ready();
}


1;
