package Sloop::Client::TokenAuth;
our $VERSION = '0.1.0';
use strict;
use warnings;


use Sloop::Lib qw(makeCookie randomToken);
use Sloop::Base;

=encoding utf8

=head1 NAME

Sloop::Client::TokenAuth

=head1 DESCRIPTION

Implements client authentication using cookies.  A TokenAuth object is
associated with a single cookie name. TokenAuth provides a random base 62 token
for each client  to be used as the cookie value; it does not actually set the
cookie as this must be done in an HTTP response.  It can do subsequent validation
of a Sloop::Client object. 

=head1 SYNOPSIS

 use Sloop::Client::TokenAuth;
 my $auth = Sloop::Client::TokenAuth->new (
 	cookieName => [REQUIRED] string
	duration => [REQUIRED] integer
	attributes => [optional] hash ref
	checkIP => [optional] 0/undef or non-zero/defined
	tokenSize => [optional] integer
 );

=over

=item B<cookieName>

The name to use to identify cookies. There should be no whitespace or any of the
following characters: ( ) < > @ , ; : \ / " { } [ ] ? =  

Those will be pruned out; you can check post construction in the C<cookieName>
attribute. 

=item B<duration>

How long a cookie remains valid for, in minutes.

=item B<attributes>

A hash ref with the following (all optional) fields.  For more information, see
RFC 6265. 

=over 2

=item domain

To be used as the value of a cookie's I<Domain> attribute.

=item path

To be used as the value of a cookie's I<Path> attribute.

=item secure

If true, the I<Secure> attribute will be added to cookies.  Make sure you use
this if you are using Sloop::Client::Secure. 

=item httpOnly

If true, the I<HttpOnly> attribute will be added to cookies.

=back

=item B<checkIP>

If defined C<validate()> calls will confirm that the IP address of the client is
the same as when the cookie was set; if not, validation fails.  The default is to
not do this.

This does not mean there can only be one client per IP address, just that a
token associated with a given IP address cannot be used from a different address.

=item B<tokenSize>

The length of the base 62 token; the default is 6, which is ~57 billion possible
combinations.

=back

=head1 ATTRIBUTES

Referring to object and not cookie attributes.

The C<cookieName>, C<tokenSize> and C<checkIP> parameters from the constructor
may be accessed as attributes of the same name.

=over

=item B<cookieAttr>

This only exists if an 'attributes' hash was defined in the constructor call. 
It is an array of strings used when setting cookies. 

=item B<duration>

This is the same as the constructor argument but in seconds.  It is modifiable.

=item B<validTokens>

This is a hash ref keyed by token.  Generally you should not have to make use
of it, but if you want to, e.g., suddenly invalidate everyone, you can
C<$auth-E<62>{validTokens}={}>.

=back

=cut
sub new {
	my $class = shift;
	my $self = {
		tokenSize => 6,
		checkIP => 0,
		@_
	};
	if ($self->{attributes}) {
		$self->{cookieAttr} = [];
		my $attr = $self->{attributes};
		push @{$self->{cookieAttr}}, "Path=$attr->{path}" if $attr->{path};
		push @{$self->{cookieAttr}}, "Domain=$attr->{domain}" if $attr->{domain};
		push @{$self->{cookieAttr}}, "Secure" if $attr->{secure};
		push @{$self->{cookieAttr}}, "HttpOnly" if $attr->{httpOnly};
		delete $self->{attributes};
	}
	Sloop::Base::goodbye("No cookieName provided!") if !defined $self->{cookieName};
	$self->{cookieName} =~ s/[\[\]{}()?=\\\/,;:<>"@\s]//g;
	Sloop::Base::goodbye("No duration provided!") if !defined $self->{duration};
	$self->{duration} *= 60;
	bless $self, $class;
}


=head1 METHODS

=head2 C<$auth-E<62>add($ip, $sessionData)>

Creates a new token and returns it.  The token should be submitted to the
client (e.g., via a "Set-Cookie:" header line) as the value half of the cookie.
For example, if the cookie name (given to the constructor) is "Chocolate" and
the token returned is "ABC123", the basic cookie would be "Chocolate=ABC123".
The cookie() method can be used to construct a complete header cookie line
with attributes.

B<$ip> is a string IP address to associate the cookie with.  If C<$auth-E<62>{checkIP}>
is not set, this doesn't matter -- just don't change that later.

B<$sessionData> is an optional scalar or reference (to anything) to attach to
each token.  This can be retrieved via fetch().

An example:

 my $token = $auth->add($client->{ip});
 my $cookie = $auth->cookie($token);
 $client->reply(\$page, cookies => [ $cookie ]);

=cut
sub add {
	my $self = shift;
# Take this opportunity to purge stale tokens.
	$self->purge;

	my $token;
	do { $token = randomToken($self->{tokenSize}) } while (exists $self->{$token});
	$self->{validTokens}->{$token} = {
		ip => shift,
		ts => time,
		data => shift
	};
	return $token;
}


=head2 C<$auth-E<62>cookie($token, $suffix)>

This is a wrapper around Sloop::Lib::makeCookie(), and returns a line suitable
for setting a cookie via the HTTP "Set-Cookie" header, etc.  This is the cookie
required by and checked for by TokenAuth.  See the 'cookies' parameter to
Sloop::Client::reply() for an easy way to use this.  

B<$token> should be the value returned by add() for the client in question.

B<$suffix> is an optional suffix to append to the cookie name to facilitate the use
of multiple tokens on the same client browser (e.g., for logins as a different user).

Any cookie attributes passed to the TokenAuth constructor are included.

=cut
sub cookie {
	my ($self, $token, $suf) = (shift, shift, shift);
	my $name = $self->{cookieName};
	$name .= $suf if $suf;
	return Sloop::Lib::makeCookie (
		$name,
		$token,
		$self->{duration} / 60,
		@{$self->{cookieAttr}}
	);
}


=head2 C<$auth-E<62>fetch($token)>

Returns whatever was submitted as the $sessionData arg to add(), for
the record associated with B<$token>.

=cut
sub fetch {
	my $self = shift;
	my $record = $self->{validTokens}->{$_[0]};
	return undef if !$record;
	return $record->{data};
}


=head2 C<$auth-E<62>delete($client, $suffix)>

Removes a cookie record; attempts to authenticate with the associated token will
subsequently fail.

B<$client> is a Sloop::Client derived object with an associated Sloop::Client::Request
object object attached.

B<$suffix> should be the same as was used for this client with C<cookie()>.

=cut
sub delete {
	my ($self, $client, $suf) = @_;
	my $token;
	my $name = $self->{cookieName};
	$name .= $suf if $suf;
	$token = $client->{request}->findCookieValue($name);
	delete $self->{validTokens}->{$token} if $token;
}


=head2 C<$auth-E<62>update($token, $suffix)>

This updates the valid duration of B<$token> by setting its timestamp to now;
i.e., increasing it by the value of the C<duration> attribute of $auth.

In order for this to be effective, the new duration must be returned to the
client or the cookie will be removed by the browser.  To that end, this method
returns the same thing as C<auth-E<62>cookie($token)>, and B<$suffix> has the
same significance as there.

Note that the validity of $token is not checked, and if you haven't called
C<$auth-E<62>validate($token)> yet, it may have expired already.  However,
if $token was previously purged, it will return undef.  This is guaranteed
not to happen between calls of validate() and update() in the same handler.

=cut
sub update {
	my ($self, $token, $suffix) = (shift, shift, shift);
	my $record = $self->{validTokens}->{$token};
	return undef if (!$token);
	$record->{ts} = time;
# Take this opportunity to purge stale tokens.
	$self->purge;
	return $self->cookie($token, $suffix);
}


=head2 C<$auth-E<62>validate($client, $suffix)>

Checks a client request to see it if includes our cookie.  If so, and the cookie
is still valid, it returns the associated unique token for this client (the same
one returned by add(), which you can use with fetch()).  If not, it returns 0. 

B<$client> is a Sloop::Client derived object. This should have a Sloop::Request
object attached to it that has not been responded to yet.

B<$suffix> has the purpose and significance described for the C<cookie()> method.

=cut
sub validate {
	my ($self, $client, $suffix) = (shift, shift, shift);
	return 0 if (!$client->{request});
	my $tokens = $self->{validTokens};
	my $name = $self->{cookieName};
	$name .= $suffix if $suffix;

	my $token = $client->{request}->findCookieValue($name);
	return 0 if !$token || !exists $tokens->{$token};
	return 0 if ($self->{checkIP} && $tokens->{$token}->{ip} ne $client->{ip});

	if ($self->{duration}) {
	# Check timestamp.
		my $now = time;
		if ($tokens->{$token}->{ts} < $now - $self->{duration}) {
			delete $self->{validTokens}->{$token};
			return 0;
		}
	}

	return $token;
}


sub purge {
	my $self = shift;
	my $now = time;
	while (my ($k, $v) = each (%{$self->{validTokens}})) {
		delete $self->{validTokens}->{$k} if ($v->{ts} < $now - $self->{duration});
	}
}

1;
