package Sloop::Constants;
our $VERSION = '0.1.0';
use 5.010000;
use strict;
use warnings;

use base 'Exporter';
our @EXPORT_OK = qw(
	READY_CANCEL
	READY_CONTINUE
	READY_REQUEST
	READY_DONE
	DBG_CONNECTION
	DBG_DESTROY
	DBG_INPUT
	DBG_OUTPUT
	DBG_POLL
	DBG_REQUEST
	DBG_REPLY
	DBG_EVERYTHING
	SPFD_READ
	SPFD_WRITE
	SPFD_RW
);
our %EXPORT_TAGS = (
	ClientState => [ qw(
		READY_CANCEL
		READY_CONTINUE
		READY_REQUEST
		READY_DONE
	) ],
	Debug => [ qw(
		DBG_CONNECTION
		DBG_DESTROY
		DBG_INPUT
		DBG_OUTPUT
		DBG_POLL
		DBG_REQUEST
		DBG_REPLY
		DBG_EVERYTHING
	) ],
	PollEvents => [ qw(
		SPFD_READ
		SPFD_WRITE
		SPFD_RW)
	],
	ALL => [ qw(
		:ClientState
		:Debug
		:PollEvents
	) ]
);

=encoding utf8

=head1 NAME

Sloop::Constants

=head1 DESCRIPTION

This module contains various groups of constants used throughout the Sloop core,
but only one of them is useful with the API, C<:Debug>.

 use Sloop::Constants qw(:Debug);

Imports all the constants explained below.  They may also be imported
individually. These can be OR'd together with the 'debug' parameter to any
Sloop::Base derived object, but since this is passed on to created Sloop::Client
objects by the server, the best place to set it is via the Sloop::Server
constructor. This may be overruled if you define a client_constructor (see
Sloop::Server SYNOPSIS). Having different values set on the server and attached
clients will result in different outcomes.  If you wish to track a particular
client you can try setting it on that object, but then not everything indicated
here may apply.

They may also be applied via a command line option (C<-g>) to the server process,
see the 'options' parameter to Sloop::Server.

They cause the Sloop::Logger object associated with the Base derived object to
report various details via the LOG_DEBUG stream.  The constants are:

=over

=item B<DBG_CONNECTION>

When applied to Sloop::Server, this will:

 require Devel::Peek;
 require Test::Memory::Cycle;

If attached to an individual client, or clients as a group via client_constructor,
I<you must use/require these yourself>.

This enables a report of the reference count on Sloop::Connection derived
objects when their file descriptor has been removed from the server queue.
Unless you are aware of a reason otherwise, it should be reported as 1,
indicating the object will be properly destroyed -- otherwise it may be leaked. 
This same check also reports the result of 
C<Test::Memory::Cycle::weakened_memory_cycle_ok> as a "Circular reference check".
This may sometimes be preceded by "Unhandled type: GLOB" which appears to be
an issue in Test::Memory::Cycle and not Sloop.

Elsewhere, it will report:

- When a Sloop::Base derived object is created; this will also report the number
of currently existing such objects, and the highest number of simultaneously
existing objects noted up until now.

- When a file descriptor is added in C<Sloop::Server::addFD()>. At the same time
this will check if the same descriptor is already in the server queue, which
should never, never happen; if it does while this switch is enabled, the server
will croak().  Do NOT use this just for safety, however.

- Pipelined requests.

- When a connection times out (see Sloop::Connection 'timeout').

- When a connection is disconnected from the other end.

=item B<DBG_DESTROY>

This will report when a Sloop::Base derived object is destroyed by indicating
the remaining number of such objects, and the highest number of simultaneously
existing objects yet observed. 

=item B<DBG_INPUT>

The will report the number of bytes read in for each low level system call via
a Sloop::Connection derived object I<created by the server>.  For, e.g.,
Sloop::Other connections, you will need to set this yourself in the constructor.

=item B<DBG_OUTPUT>

The will report the number of bytes written out for each low level system call
via a Sloop::Connection derived object I<created by the server>.  For, e.g.,
Sloop::Other connections, you will need to set this yourself in the constructor.

=item B<DBG_POLL>

At the end of each system level C<epoll()>, this will report whether the call
timed out, and if not, which connections came back as NEW (just connected), IN (input
pending), OUT (output pending), or DONE (disconnected). 

It will also report when the server checks connection timeouts (even if there
aren't any).

=item B<DBG_REQUEST>

This will print HTTP request headers as exactly as they are received, before any
parsing is done.

=item B<DBG_REPLY>

This will print HTTP response headers generated by the server before they are
sent.  It will also report "Reply completed" when the complete reply has
actually been transmitted.

=item B<DBG_EVERYTHING>

This combines all of the above.

=back

=cut

# Poll event constants
use constant {
	SPFD_READ => 1,
	SPFD_WRITE => 4,
	SPFD_RW => 5
};

# Debug constants
use constant {
	DBG_CONNECTION => 1,
	DBG_DESTROY => 2,
	DBG_INPUT => 4,
	DBG_OUTPUT => 8,
	DBG_POLL => 16,
	DBG_REQUEST => 32,
	DBG_REPLY => 64,
	DBG_EVERYTHING => 127
};

# Used w/ Server -g option.
our %Debug = (
	connection => DBG_CONNECTION,
	destroy => DBG_DESTROY,
	input => DBG_INPUT,
	output => DBG_OUTPUT,
	poll => DBG_POLL,
	request => DBG_REQUEST,
	reply => DBG_REPLY,
	everything => DBG_EVERYTHING
);

# Client Ready constants
use constant {
	READY_CANCEL => -1,
	READY_CONTINUE => 0,
	READY_REQUEST => 1,
	READY_DONE => 2
};

1;
