package Sloop::Lib;
our $VERSION = '0.2.0';
# 0.2.0:
# - Added callerTag() and checkURL().
use 5.010000;
use strict;
use warnings;

require Exporter;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(
	callerTag
	checkURL
	dataLength
	errorString
	fastSlurp
	fieldPairs
	httpTimestamp
	lcSuffix
	lowerCase
	makeCookie
	randomToken
	splitstr
	terminateDirPath
	untaintPath
);
our %EXPORT_TAGS = (all => \@EXPORT_OK);

require XSLoader;
XSLoader::load('Sloop::Lib', $VERSION);

=encoding utf8

=head1 NAME

Sloop::Lib

=head1 DESCRIPTION

Sloop::Lib is a hodge podge of simple utility functions used by other modules,
some of which are documented for external use here.  You can import them by name.

=head1 SUBROUTINES

=head2 C<callerTag ($packageName)>

Wrapper on C<caller()> which goes up the call stack until the package is not
$packageName (which can just be C<__PACKAGE__>, to apply to the current one)
-- in the spirit of Carp but seemingly more accurate, and also more detailed.
The output is a single string:

 In: lib/foo/bar.pm (line 123?)
 Package: foo::bar
 Sub: whatever()
 Call: hello()

There is no final newline.

The difference between "Call" and "Sub" is one stack frame, with the former
being up from everything else (i.e., presumably where the original error is).

If the walk up leads to a top level line in the base script outside any
subroutine (which is unlikely in an event driven Sloop context), then an
empty string is returned; likewise for "Call: ".

=cut
sub callerTag {
	my $package = shift;
	my $depth = 1;
	my @info = caller $depth++;
	return "" if !@info;
	while ($info[0] eq $package) {
		@info = caller $depth++;
		return "" if !@info;
	}
	# Go up one more.
	my @up = caller $depth;
	my $call = "";
	$call = $up[3] if @up;
	return "In: $info[1] (line $info[2]?)\nPackage: $info[0]\n\n"
		."Sub: $info[3]()\nCall: $call";
}


=head2 C<checkURL ($client, $min, $max)>

Uses Sloop::Client::Request::pathFromLevel() to determine if the remainder
of a request path (after the handler cascade) is appropriate.

The number of levels should be between C<$min> and C<$max>, inclusive.
However, both can be left undefined, in which case only a 0 remainder is
acceptable.  If only C<$min> is defined, then the remainder levels can be
anything greater than that.

If this test fails, C<$client> is sent a generic 404 message, a debug
message including the original request path is logged, and the function
returns undef.  Otherwise, C<$client> is returned unmodified.

=cut
sub checkURL {
# @_ = client, min levels, max levels.
	my $check = $_[0]->{request}->pathFromLevel;
	if (!$check) {
		return $_[0] if !$_[1];
		$_[0]->debug_log("Not found: ".($_[0]->{request}->requestPath));
		$_[0]->generic(404);
		return undef;
	}
	return $_[0] if (
		defined $_[1] && 
		$check >= $_[1] && 
		(!$_[2] || $check <= $_[2])
	);
	$_[0]->debug_log("Not found: ".($_[0]->{request}->requestPath));
	$_[0]->generic(404);
	return undef;
}


=head2 C<fastSlurp ($path)>

Reads a file into a scalar. This is about ten lines of XS code combining low
level POSIX C I/O with the perl API.  It is faster than any method I've found for
slurping files in perl on linux. 

$! is set on failure.

B<$path> File to fetch.

B<Returns> A scalar containing the file or undef if the file can't be read.

=cut


=head2 C<httpTimestamp($time, $zone)>

This function returns a timestamp string valid for use in HTTP headers as per
RFC 822 section 5.

B<$time> is an array reference to the return value of perl's localtime() or
gmtime() functions.

B<$zone> is the three letter time zone in capitals, e.g., "GMT".

=cut
our @Days = qw(Sun Mon Tue Wed Thu Fri Sat);
our @Months = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
sub httpTimestamp {
	my @data = @{(shift)};
	return sprintf("%s, %02d-%s-%d %02d:%02d:%02d ".(shift),
		$Days[$data[6]], $data[3], $Months[$data[4]], $data[5]+1900,
		$data[2], $data[1], $data[0]);
}


=head2 C<lcSuffix($filename)>

Returns a lower cased version of a file's suffix.

B<$filename> A file name or path.

B<Returns> A lower cased version of the file suffix, or undef if there wasn't one.


=head2 C<makeCookie($name, $value, $duration, @attr)>

This is a convenience function for constructing an HTTP header value,
 to be used with "Set-Cookie:" in a response.  See the
the 'cookie' parameter to Sloop::Client::reply()).
This line can also be used with javascript methods for setting cookies.

B<$name> and B<$value> is the name/value pair identifying the cookie.  You can
fetch a cookie $value set on a Sloop::Client::Request object by calling
C<$req-E<62>findCookieValue($name)>.

The C<$name> should not contain any whitespace or the following characters:
( ) < > @ , ; : \ / " { } [ ] ? = 

Those will be pruned out if present.  The C<$value> should not contain whitespace
or any of the following characters: \ " , ;

I<Niether name nor value should contain non-ASCII characters>, but C<makeCookie()>
does not check for this!

B<$duration> is how long a cookie remains valid for, in minutes, from the time
it is issued to a client.

B<@attr> is an optional list of string cookie attributes.  If this is an https
connection, make sure to include 'Secure'.  If you want an alternate path or domain
associated with the cookie, use 'Path=/path' or 'Domain=wherever'.  If you want
an HTTP only cookie, use 'HttpOnly'.  For more information, see RFC 6265.

=cut
sub makeCookie {
	my $cookie = shift;
	$cookie =~ s/[\[\]{}()?=\\\/,;:<>"@\s]//g;
	my $value = shift;
	$value =~ s/[\s",;\\]//g;
	$cookie .= "=$value; Expires=";
	my $seconds = (shift)*60;
	my @data = gmtime(time + $seconds);
	$cookie .= httpTimestamp(\@data, 'GMT');
	$cookie .= '; max-age='.$seconds;
	$cookie .= '; '.$_ foreach (@_);
	return $cookie;
}


=head2 C<randomToken($length)>

Creates a random string composed of the characters A-Z, a-z,
and 0-9 (i.e., base 62 digits).

B<$length> is the length of the string.

B<Returns:> The random token.

=cut
our @Base62 = ('0'..'9', 'A'..'Z', 'a'..'z');
sub randomToken {
	my $len = pop;
	my @chars;
	push @chars, $Base62[rand 62] foreach (1..$len);
	return join('', @chars);
}


=head2 C<terminateDirPath($path)>

Appends '/' to a directory path if it is not present.

B<$path> is a scalar containing the path.  This should NOT be
a reference -- it will be appended to via XS code.


=head2 C<untaintPath($pathref)>

Checks a path for '..' and escapes (by prefixing '\') \ * | < and >.

B<$pathref> is a I<reference> to scalar containing the path.

B<Returns> 0 if the path contains '..', 1 otherwise.

=cut
sub untaintPath {
	my $path = shift;
	return 0 if index($$path, '..') >= 0;
	$$path =~ s/([\\|\*<>])/\\$1/g;
	return 1;
}


1;
