package Sloop::Logger;
our $VERSION = '0.1.0';
use strict;
use warnings;

use parent 'Exporter';
our @EXPORT = qw(LOG_ERR LOG_DEBUG LOG_REQUEST);

# These values are guaranteed not to change in the future.
use constant {
	LOG_ERR => 0,
	LOG_DEBUG => 1,
	LOG_REQUEST => 2
};

=encoding utf8

=head1 NAME

Sloop::Logger

=head1 DESCRIPTION

Sloop::Logger is a simple class for logging to I<open> filehandles.  It is used by many of the
other Sloop classes.  If you wish to subclass it (e.g., because you prefer the log be opened
and closed for each message), the only method you need to override is out().

=head1 SYNOPSIS

 use Sloop::Logger;
 my $log = Sloop::Logger->new (
	logs => (optional) Reference to array of 3 file handle refs.
 );

There should be three entries in the 'logs' array; these are indexed
corresponding to the constants exported by this module:

 0 = LOG_ERR
 1 = LOG_DEBUG
 2 = LOG_REQUEST

The entries themselves should be file handle references.
If not specified, defaults to C<[ \*STDERR, \*STDERR, \*STDOUT ]>

The constructor then turns off line buffering, a.k.a. turns on autoflush,
for all three handles.

Logger is inflexible in terms of formatting the
output, but can be easily subclassed to override the out() method below.

=head1 ATTRIBUTES

The C<logs> array is accessible and can be modified at any point.

=cut
sub new {
	my $class = shift;
	my $self = {
		logs => [ \*STDERR, \*STDERR, \*STDOUT ],
		@_
	};

# Disable line buffering for all handles.
	my $oldfd = select();
	foreach (@{$self->{logs}}) {
		select($_);
		$| = 1;
	}
	select($oldfd);

	bless $self, $class;
}


=head1 METHODS


=head2 C<$log-E<62>out($fh_index, $message)>

B<$fh_index> is one of the the LOG_ values from Sloop::Constant
(see constructor synopsis above).

B<$message> is the text to print to the log.

This appends a timestamp as described in the timestamp() method below.

If you subclass Logger to override this method, it will handle all
calls to print messages via the logger object, such as the err_log()
method of Sloop::Base inherited by Sloop::Client and Sloop::Server.

=cut
sub out {
	my $self = shift;
	my $fh = $self->{logs}->[$_[0]];
	my @msg = split(/\n+/, $_[1]);
	print $fh $self->timestamp.(shift @msg)."\n";
	print $fh "                $_\n" foreach @msg;
}


=head2 C<$log-E<62>timestamp()>

Returns a timestamp corresponding to the current local time in the form:

 DD [Month] HH:MM:SS

That's the day, the month in three letters, (e.g. "Jan"), and the time.
A space is appended at the end.

=cut
sub timestamp {
	my ($sec,$min,$hour,$day,$m) = ((localtime(time()))[0,1,2,3,4]);
	my @months = qw(Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec);
	foreach ($sec,$min,$hour,$day) { $_ = "0$_" if ($_ < 10) }
	return "$day $months[$m] $hour:$min:$sec ";
}


1;
