package Sloop::Multipart;
our $VERSION = '0.1.0';
use strict;
use warnings;

require XSLoader;
XSLoader::load('Sloop::Multipart', $VERSION);

use Sloop::Multipart::Data;
use Sloop::Base;
use Sloop::Client::Request;
use Sloop::Lib qw(dataLength fieldPairs splitstr fastSlurp);
use URI::Encode;

# Process constants
use constant {
	GET_HEADERS => 0,
	CONTINUE_VALUE => 1,
	MULTIPART_ERROR => 2,
	MULTIPART_COMPLETE => 3
};

=encoding utf8

=head1 NAME

Sloop::Multipart

=head1 DESCRIPTION

Class for parsing HTTP multipart POST request, as defined in RFCs 1867, 2046, 2388, etc.
This is the normal way of uploading files from a browser, via a "multipart/form-data"
HTML form.

=head1 REQUIRES

This refers only to modules not part of Sloop or the perl core.

B<URI::Encode>

=head1 SYNOPSIS

 my $smp = Sloop::Multipart->new (
	$request,
	$max,
	$tmpdir,
	@expected
 );

B<$request> is a Sloop::Client::Request object.  I<The constructor will return undef
if the request header is incorrect for a multipart upload.>  Be sure to check that.

B<$max> is the maximum total size of the upload.  This is checked as it proceeds
primarily to prevent illegitimate oversized payloads; if you want to validate requests before
the upload begins, you should check $request->header('content-length') yourself.

B<$tmpdir> is the path to a directory that the Sloop process has write permissions on.  This is required
by Sloop::Multipart::Data objects to hold the upload data as it accumulates.  If the value is 'C<_MEM_>',
it will be stored in a scalar instead.

B<@expected> is an optional array (i.e., the remainder of the args array); each element will be a
key for the $smp->{expected} hash -- see {parts} and {expected} under ATTRIBUTES.

=head1 ATTRIBUTES

=over

=item C<$smp-E<62>{error}>

Used to store an error message, see $smp->process() under METHODS.

=item C<$smp-E<62>{parts}>

Array reference of Sloop::Multipart::Data objects created by $smp->process().  These represent
the seperate parts of the upload I<except> those that are in $smp->{expected}.

=item C<$smp-E<62>{expected}>

A hash ref of Sloop::Multipart::Data objects constructed by $smp->process() if '@expected' args
were passed to the constructor (see SYNOPSIS).  If the POST body contained more than one item
for a given key, only the first one will be in this hash -- the rest will be in the {parts}
array.

=back

=cut

### NON-API MEMBER FIELDS ###

# $smp->{boundary}
# This is the boundary string, extracted from the request 'Content-Type' header.

# $smp->{blen}
# Length of $smp->{boundary} in bytes.

# $smp->{cur}
# Used in $smp->process (current Multipart::Data object).

# $smp->{data}
# Buffer used in $smp->process().

# $smp->{partial}
# Temporary buffer for partial matches on $smp->{boundary} at the end of
# a data chunk fed to $smp->process().  Qv. method endCheck().

# $smp->{state}
# One of the 'Process constants' (see top), used in $smp->process().

sub new {
	my ($class, $req) = (shift, shift);
	my $content_type = $req->header('content-type');
	return undef if (!$content_type);

	my $check =()= $content_type =~ /.*?oundary\s*=\s*([^\s]+)/;
	return undef if !$check;
# $1 is the boundary marker string.  Add "\r\n" to the beginning,
# since it is on a newline, and add "\r\n" to our collected data
# to simulate this for the leading boundary on the first line of
# the POST body.  
# Also add "--" to the boundary (see RFC 2046 "dash-boundary").
	my $self = {
		boundary => "\r\n--".$1,
		maxBytes => shift,
		tmpdir => shift,
		state => GET_HEADERS,
		data => "\r\n",
		parts => [],
		cur => undef
	};
	$self->{blen} = dataLength(\$self->{boundary});
	$self->{expected} = { };
	foreach (@_) { $self->{expected}->{$_} = undef }

	bless $self, $class;
}


=head1 METHODS

=head2 C<$smp-E<62>process($dataref)>

This is called to add data from the POST body for processing.

B<$dataref> is a scalar I<reference> to the data to be processed.
Generally, it should be to the {body} field of a Sloop::Client::Request
object, in a Sloop::Client::readPost() or simplePost() handler.
$$dataref will be left empty by this call.

B<Returns> -1 on error, 0 to indicate there should still be data to process, and
1 once the last part has been added.  If 0, you may do as you wish with the
Sloop::Multipart::Data objects attached (including remove them); they are
complete even if the rest of the upload is not.  If the upload turns out to be
empty, process() will return -1 and report this as an error; thus success
guarantees there is at least one element in $smp-E<62>{parts}. Calling process()
again after it has returned -1 I<will be fatal to the server>, so be sure to
check this. The nature of the error will be reported as a string in
$smp-E<62>{error}. 

=cut

sub process {
	my ($self, $dataref) = (shift, shift);
	$self->{data} .= $$dataref;
	$$dataref = "";

	while (1) {
		if ($self->{state} == GET_HEADERS) {
			my $end = index($self->{data}, "\r\n\r\n");
			my $start = index($self->{data}, $self->{boundary});
		# Check if the previous boundary ends with "--", then there are no more parts.
			if ($end == -1) {
				return 0 if $start == -1;
				$start += $self->{blen};
			# If there aren't two characters left after the boundary, return.
				return 0 if ($start + 2 > dataLength(\$self->{data}));
				if (substr($self->{data}, $start, 2) eq '--') {
					$self->{state} = MULTIPART_COMPLETE;
					next;
				}
				$start += 2;
			} else {
				if ($start == -1) {
					return $self->mpartError("Malformed multipart POST, no leading boundary.");
				}
				$start += $self->{blen} + 2;
			}
		# Extract the headers.
			my $head = substr($self->{data}, $start, $end - $start + 4);
			my $headers = Sloop::Client::Request::getHeaders(\$head);
			if (!exists $headers->{'content-disposition'}) {
				return $self->mpartError("Malformed multipart POST, no Content-Disposition.");
			}
			$headers->{'content-disposition'} = fieldPairs($headers->{'content-disposition'}, 1);
		# Discard the raw header data.
			Sloop::Lib::tailstr($self->{data}, $end+4);

		# Get name for this part.
			if (!exists $headers->{'content-disposition'}->{name}) {
				return $self->mpartError("Malformed multipart POST, no field name.");
			}
		# Construct a new Multipart::Data object for this part.
			$self->{cur} = Sloop::Multipart::Data->new (
				$headers,
				$self->{tmpdir}
			);

		# Advance state.
			$self->{state} = CONTINUE_VALUE;
		}

		elsif ($self->{state} == CONTINUE_VALUE) {
		# If there was a partial match on the boundary from the last chunk,
		# push that onto the front of the current chunk.
			if (exists $self->{partial}) {
				$self->{data} = join('', ($self->{partial}, $self->{data}));
				delete $self->{partial};
			}
			my $end = index($self->{data}, $self->{boundary});
		# If the boundary isn't found, check for a partial match at the end
		# via endCheck() and write what is out.
			if ($end < 0) {
				$self->endCheck;
				return -1 if $self->addData(\$self->{data}) < 0;
				return 0;
			}
		# If the boundary is there, keep it in {data}, write the rest out,
		# and complete this part.
			my $data = splitstr($self->{data}, $end);
			return -1 if $self->addData(\$data) < 0;
			$self->{cur}->done;
			if (
				exists $self->{expected}->{$self->{cur}->{name}}
				&& !defined $self->{expected}->{$self->{cur}->{name}}
			) { $self->{expected}->{$self->{cur}->{name}} = $self->{cur} }
			else { push @{$self->{parts}}, $self->{cur} }
		# Reset state and continue loop.
			$self->{state} = GET_HEADERS;
		}

		elsif ($self->{state} == MULTIPART_COMPLETE) {
			if ($#{$self->{parts}} == -1) {
				return $self->mpartError("No parts in upload.") 
			}
			return 1;
		}

		elsif ($self->{state} == MULTIPART_ERROR) { Sloop::Base::goodbye("You failed to error check!") }

	}
}

sub mpartError {
	my $self = shift;
	$self->{state} = MULTIPART_ERROR;
	$self->{error} = shift;
	return -1;
}


=head2 C<$smp-E<62>valueFromExpected($partName, $max)>

A convenience wrapper around Sloop::Multipart::Data::getContent(); see
that documentation regarding the side effects of calling this method.

B<$partName> is a name submitted to the constructor in the @expected array.

B<$max> is the maximum number of bytes to permit.  If the length of the
part's content is greater than this, the content (but not the rest of the 
Sloop::Multipart::Data object) is deleted and undef is returned.

B<Returns> the content of the part, or undef if it does not exist or was too
big.

=cut
sub valueFromExpected {
	my $self = shift;
	my $part = $self->{expected}->{(shift)};
	if ($part->{length} > (shift)) {
		$part->cancel;
		return undef;
	}
	return undef if !defined $part;
	return $part->getContent;
}


# Passes data to the current Sloop::Multipart::Data object.
sub addData {
	my ($self, $dataref) = (shift, shift);
	my $add = dataLength($dataref);
	$self->{maxBytes} -= $add;
	if ($self->{maxBytes} < 0) {
		$self->{cur}->cancel();
		$self->{error} = "Max bytes exceeded.";
		$self->{state} = MULTIPART_ERROR;
		return -1;
	}
	if ($self->{cur}->addData($dataref, $add) == -1) {
		$self->{error} = "Temp file not available for \"$self->{cur}->{name}\".";
		$self->{state} = MULTIPART_ERROR;
		return -1;
	}
	return $add;
}

1;
