package Sloop::Multipart::Data;
our $VERSION = '0.1.1';
# 0.1.1
# - Minor perldoc correction.
use strict;
use warnings;

use File::Copy;
use File::Temp;
use Sloop::Lib qw(dataLength fastSlurp);

=head1 NAME

Sloop::Multipart::Data

=head1 DESCRIPTION

Sloop::Multipart::Data objects are constructed by a Sloop::Multipart object 
for each part found in an HTTP multipart post body.

=head1 ATTRIBUTES

=over

=item C<$mpd-E<62>{name}>

The name of the part.  This corresponds to the name attribute of the HTML
input tag responsible for this part.  Since some inputs can provide
multiple parts, it is not necessarily unique in relation to other
Sloop::Multipart::Data objects produced by the same request.

=item C<$mpd-E<62>{length}>

This is the I<accumulated> length of the content.

=item C<$mpd-E<62>{headers}>

The HTTP protocol requires individual headers for each part; {headers}
is a hash reference to them much like the one in a Sloop::Client::Request,
except this is two tiered: B<'content-disposition'> is a sub hash.  Note
that C<$mpd-E<62>{headers}-E<62>{'content-disposition'}-E<62>{name}> is
the same as C<$mpd-E<62>{name}>.

=item C<$mpd-E<62>{tmpfile}>

The full pathname of the tmp file used to store content.
B<This file is automatically deleted by $mpd-E<62>DESTROY> when 
(e.g.) $mpd goes out of scope.

=back

=cut

### NON-API MEMBER FIELDS ###

# $mpd->{fh}
# File::Temp file handle.

sub new {
	my $class = shift;
	my $self = { headers => shift };
	my $path = shift;
	if ($path ne '_MEM_') {
		($self->{fh}, $self->{tmpfile}) = File::Temp::tempfile(DIR => $path);
	} else { $self->{memdata} = "" };
	$self->{name} = $self->{headers}->{'content-disposition'}->{name};
	$self->{length} = 0;
	bless $self;
}

=head1 METHODS

=cut

# Add data to the part's content.
sub addData {
	my ($self, $dataref) = (shift, shift);
	if ($self->{fh}) {
		my $fh = $self->{fh};
		return -1 if (tell($fh) == -1);
		print $fh $$dataref;
	} else { $self->{memdata} .= $$dataref; }
	$self->{length} += shift;
	$$dataref = "";
	return 0;
}


# Used to abort.
sub cancel {
	my $self = shift;
	return if !defined $self->{fh};
	close $self->{fh};
	unlink($self->{tmpfile});
}


# Finalizes the part.
sub done {
	my $self = shift;
	close $self->{fh} if (defined $self->{fh});
}


=head2 C<$mpd-E<62>filename()>

Retrieves the base filename of the part, if this a file upload.

B<Returns> The file name, or undef.

=cut

sub filename {
	my $self = shift;
	my $disp = $self->header('content-disposition');
	return $disp->{'filename'};
}


=head2 C<$mpd-E<62>header($key)>

Convenience method for retrieving a value from $mpd->{headers}.

B<$key> is the name of the header.

=cut

sub header {
	my $self = shift;
	return $self->{headers}->{(shift)};
}


=head2 C<$mpd-E<62>placeFile($filepath)>

This will move (via File::Copy::move()) $mpd->{tmpfile} to a
more permanent location.

B<$filepath> is the full path to move the file to.  The sloop
process must have write permission to this location.

B<Returns> 1 for success, 0 for failure.

=cut
sub placeFile {
	my $self = shift;
	return 0 if !defined $self->{fh};
	close($self->{fh});
	return move($self->{tmpfile}, shift);
}


=head2 C<$mpd-E<62>getContent()>

Convenience method for getting the content, rather than dealing with
$mpd->{tmpfile}.  Do not call this method this after placeFile() or otherwise
moving the {tmpfile}, or call it multiple times; {tmpfile} will be deleted  after
the call. 

B<Returns> The content of $mpd->{tmpfile}.

=cut
sub getContent {
	my $self = shift;
	return $self->{memdata} if defined $self->{memdata};
	close $self->{fh};
	my $content = fastSlurp($self->{tmpfile});
	unlink($self->{tmpfile});
	return $content;
}

=head2 C<$mpd-E<62>size()>

API call to retrieve the size of $mpd->{data} I<before> the content is accessed.
Afterward, this will return undef.

B<Returns> For _MEM_ based objects, undef. Otherwise, the size in bytes of
C<$mpd-E<62>{tmpfile}>.

=cut

sub size {
	my $self = shift;
	return undef if defined $self->{memdata};
	return (stat($self->{tmpfile}))[7];
}

sub DESTROY {
	my $self = shift;
	unlink($self->{tmpfile}) if defined $self->{fh};
}


1;
