package Sloop::Other;
our $VERSION = '0.1.1';
use strict;
use warnings;
# Version 0.1.1
# - Removed 'dest_port' requirement for LOCAL connections.
# - Moved 'okay()' to Sloop::Base.
# - Set default epfd to -1.
# - Added check on 'incoming' return value in ready()

use parent 'Sloop::Connection';

require XSLoader;
XSLoader::load('Sloop::Other', $VERSION);

use Carp;
use Scalar::Util qw(looks_like_number);
use Sloop::Constants qw(:ClientState :PollEvents);

=encoding utf8

=head1 NAME

Sloop::Other

=head1 DESCRIPTION

Sloop::Other allows for connections to other TCP servers as a client.  Like
Sloop::Client, it must be combined with Sloop::Socket::Standard or Sloop::Socket::TLS,
which is the purpose of Sloop::Other::Regular and Sloop::Other::Secure.  However,
both these inherit this API.

You may use a Secure Other with a normal Server and vice versa.

To be useful, the object must first be passed to C<$sloop-E<62>connectOther()>,
where $sloop is a Sloop::Server object.

=head1 SYNOPSIS

 use Sloop::Other
 my $other = Sloop::Other->new (
     debug => (optional) see Sloop::Base
     dest_addr => (REQUIRED) scalar IP address
     dest_port => Integer port number, required for INET connections.
     id => (optional) see Sloop::Base
     incoming => (REQUIRED) sub or array ref
     logger => (REQUIRED) Sloop::Logger object
     max_recv => (optional) see Sloop::Connection
     max_send => (optional) see Sloop::Connection
     proto => (optional) 1 (Unix local) or (INET IPv) 4 or 6; default 4.
 )

Parameters not described in a base class (Sloop::Base and Sloop::Connection),
and/or requiring further explication here are listed below.

=over

=item B<debug>

This has the same significance it does for other client objects attached to the
server (see Sloop::Constants), but because the $other object is not created by
the server you must set this value yourself; normal clients have it set the same
as the 'debug' argument to C<Sloop::Server::new>.

=item B<dest_addr>

An IP address (not a domain name) in string form, e.g. "192.168.1.2" for IPv4
or "ff:80ab::1234:2015::dead:beef" for IPv6.

=item B<dest_port>

The port number to connect to.

=item B<id>

If not defined, the package name ("Sloop::Other") is used.

=item B<incoming>

Either a subroutine reference, or a reference to an array of sub refs.  If the
latter, when required the first sub in the array will be used. If it returns
> 0, it will then be shifted off the front of the array. 

The sub will be fired when a message is received.  It is passed three arguments,
the Sloop::Other object, a reference to a scalar containing the message, and the
length of the message in bytes.  If it returns -1, the connection will be
closed and the object is subsequently not usable for communication.

 $other->{incoming} = sub {
     my ($self, $msgref, $bytes) = (shift, shift, shift);
     # Acknowledge reciept to the other party.
     $self->message(\"Recieved $bytes bytes.");
     return 1;
 }

If the other party is also an HTTP server, you can use
C<Sloop::Client::Request::getHeaders()> to parse the response headers, just
remember to remove the first line (status) first.

=item B<proto>

This is 1 for AF_LOCAL sockets or else the IP version to use; the default is 4.
It does not have to be the same as the Sloop::Server.

=back

=cut
sub new {
	my $class = shift;
	my %args = (
		proto => 4, # 1 = LOCAL, 4 = INET, 6 = INET6
		id => __PACKAGE__,
		@_,
	);
	$args{dest_port} = -1 if ($args{proto} == 1);
	my $self = $class->SUPER::new ( %args );
	croak __PACKAGE__."::new() dest_port not defined." if (!defined $self->{dest_port});
	croak __PACKAGE__."::new() dest_addr not defined." if (!defined $self->{dest_addr});
	croak __PACKAGE__."::new() logger not defined." if (!defined $self->{logger});
	croak __PACKAGE__."::new() incoming not appropriately defined."
		if (!defined $self->{incoming} || !ref $self->{incoming});

	$self->{fd} = $self->connectFD;
	if ($self->{fd} < 0) {
		$self->{SHUT} = 1;
		$self->{logger}->out (0,
			__PACKAGE__."::connectFD() failed: $self->{error}"
		);
	}

	$self->{epfd} = -9;

	bless $self, $class;
}


=head1 METHODS

=head2 C<$other-E<62>message($dataref)>

Queues data to be sent to the remote server.  Sloop does not do any processing of
this.  This is a non-blocking call; it queues data to be sent and does not wait
for its transmission.

B<$dataref> is a reference to a scalar containing the data.  It may be of any sort
(i.e., text or binary).

=cut
sub message {
	my ($self, $dataref) = (shift, shift);
	croak __PACKAGE__.'::message() $sloop->connectOther() not used post initialization!'
		if $self->{epfd} == -9;

	if ($self->{SHUT}) {
		$self->{logger}->out(0, __PACKAGE__."::message() called on closed connection.");
		return 0;
	}
	$self->setEvents(SPFD_RW);
	if ($self->{data_pending}) { ${$self->{data_pending}} .= $$dataref }
	else { $self->{data_pending} = $dataref }
	return 1;
}


# End of API.

# Required by Sloop::Connection.
sub checktime {
	croak __PACKAGE__.'::checktime() $sloop->connectOther() not used post initialization!'
		if $_[0]->{epfd} == -9;
	return 1 if $_[0]->{SHUT};
	return 0;
}


# Required by Sloop::Connection.
sub doneWrite {
	return -1 if $_[0]->{SHUT};
	$_[0]->setEvents(SPFD_READ);
	return 1;
}


# Required by Sloop::Connection.
sub ready {
	my $self = shift;
	croak __PACKAGE__.'::ready() $sloop->connectOther() not used post initialization!'
		if $self->{epfd} == -9;

	my $bytes = 0;
	my $dataref = $self->readData(\$bytes);
	if (!$dataref) {
		return READY_CANCEL if $self->{SHUT};
		return READY_CONTINUE;
	}
	if (ref $self->{incoming} eq 'CODE') {
		my $check = $self->{incoming}->($self, $dataref, $bytes);
		if (!looks_like_number($check)) {
			$self->err_log("Sloop::Other: 'incoming' should return a number.");
			return READY_CONTINUE;
		}
		return READY_CANCEL if $check == -1;
		return READY_CONTINUE;
	}
	my $sub = $self->{incoming}->[0];
	if (!$sub) {
		$self->{logger}->out (0,
			"Sloop::Other '$self->{id}':\n"
			."Message recieved but incoming array empty!"
		);
		return READY_CANCEL;
	}
	my $check = $sub->($self, $dataref, $bytes);
	if (!looks_like_number($check)) {
		$self->err_log("Sloop::Other: 'incoming' should return a number.");
		return READY_CONTINUE;
	}
	return READY_CANCEL if $check == -1;
	shift @{$self->{incoming}} if $check > 0;
	return READY_CONTINUE;
}


=head2 C<$other-E<62>shut>

You can call this if you want to arbitrarily render the connection defunct
but ok() still returns true.  It gets called by the server on certain 
errors, or via the destructor (actually, via the Sloop::Base destructor,
although it is implemented in one of the Sloop::Socket classes).

"Defunct" means you cannot use this object to send or receive messages
anymore.

=cut

1;
