package Sloop::Other::Secure;
our $VERSION = '0.1.0';
use strict;
use warnings;

use parent qw(Sloop::Socket::TLS Sloop::Other);

use Sloop::Constants (":ClientState");

use constant {
	TRUST_ALL => 1,
	TRUST_CERTIFIED =>2
};

=encoding utf8

=head1 NAME

Sloop::Other::Secure

=head1 DESCRIPTION

This class can be used for Sloop::Other connections to ssl/tls servers.  It does
not require that the Sloop::Server instance be run in secure mode, but if it is
not, then you must call C<Sloop::Socket::TLS::init()> before you call
C<$sloop-E<62>connectOther>. 

Currently there are some limitations:

- Only X509 based authentication is available.

- Use of client certificates is not supported.

- Certificate revocation lists (CRLs) are not checked; authenticating the server
is done only on the basis of what is in your trust file. 

=head1 SYNOPSIS

The API is described in Sloop::Other, with a few extra optional arguments to the
constructor -- which returns undef if the initial connection fails; there
should be a GnuTLS error logged in this case.

=over

=item B<trustLevel>

This is one of Sloop::Other::Secure::TRUST_CERTIFIED (the default) or ::TRUST_ALL.
If the former, the remote server is authenticated using the 'trustFile' passed to
C<Sloop::Socket::TLS::init()>.  If the latter, the certificate is not checked..

=item B<otherHostname>

This is required to authenticate the other server's certificate, unless you
use TRUST_ALL (see above).  The default is 'localhost'.

=item B<priority>

This is a string passed as the second argument to C<gnutls_priority_set_direct()>,
defining what ciphers are allowed in order of preference.  The default is:

 NORMAL:%PARTIAL_RENEGOTIATION

The nature of the string is partially described in 'man gnutls_priority_init',
and more thoroughly in the GnuTLS guide:

I<http://gnutls.org/manual/html_node/Priority-Strings.html>

=back

=cut
sub new {
	my $class = shift;
	my $self = $class->SUPER::new (
		trustLevel => TRUST_CERTIFIED,
		otherHostname => 'localhost',
		@_
	);
	$self->{passedShake} = 0;
	bless $self, $class;

	if (!$self->clientSession) {
		$self->err_log($self->{tls_error});
		$self->cancel;
		return undef;
	}

	my $check = $self->shakehands;
	if ($check == -1) {
		$self->err_log($self->{tls_error});
		$self->cancel;
		return undef;
	} elsif ($check == 1) { $self->{passedShake} = 1 }

	return $self;
}


sub ready {
	my $self = shift;
	return $self->SUPER::ready() if $self->{passedShake};

# TLS handshake isn't complete, try again.
	my $check = $self->shakehands;
	if ($check == -1) {
		$self->err_log($self->{tls_error});
		$self->cancel;
		return Sloop::Constants::READY_CANCEL;
	} elsif ($check == 0) {
		return Sloop::Constants::READY_CONTINUE;
	}

# Handshake is now complete!
	$self->{passedShake} = 1;
	return $self->SUPER::ready();
}


1;

