#define PERL_NO_GET_CONTEXT
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#include "ppport.h"

#include <arpa/inet.h>
#include <grp.h>
#include <errno.h>
#include <netinet/in.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/epoll.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

#define MAXPOLLEVENTS 128

// defines for use in Epoll()
#define ACTIVE_IN_SET 1
#define ACTIVE_OUT_SET 2
#define ACTIVE_DONE_SET 4
#define ACTIVE_NEW_SET 8

static int acceptClient (HV *self, SV *ip) {
/* acceptClient (HV *self, SV *ip)
Calls accept() on the server listen socket and sets the new connection non-block.
	'self' is a Sloop::Server object.
	'ip' is an empty string used to pass back the IP address of the client connected,
	     if successful.
Returns the file descriptor for the new connection, or:
 -1 sock() failed
 -2 fcntl get flags failed
 -3 fcntl set non-block failed */
	dTHX;

	SV **field = hv_fetch(self, "fd", 2, 0);
	int srv = SvIV(*field);

	int ipv = AF_INET;
	field = hv_fetch(self, "IPv", 3, 0);
	if (field) {
		int check = SvIV(*field);
		if (check == 6) ipv = AF_INET6;
	}

	struct sockaddr_in in4;
	struct sockaddr_in6 in6;
	struct sockaddr *info = (struct sockaddr*)&in4;
	socklen_t len = sizeof(in4);
	if (ipv == AF_INET6) {
		info = (struct sockaddr*)&in6;
		len = sizeof(in6);
	}

	int sock = accept(srv, info, &len);
	if (sock == -1) return -1;

	// Set keep-alive.
	int yes = 1;
	setsockopt(sock, SOL_SOCKET, SO_KEEPALIVE, &yes, sizeof(yes));

	// Set close-on-exec.
	int flags = fcntl(sock, F_GETFD, 0);
	if (flags == -1) {
		close(sock);
		return -2;
	}
	if (fcntl(sock, F_SETFD, flags | FD_CLOEXEC) == -1) {
		close(sock);
		return -3;
	}

	// Set non-block.
	flags = fcntl(sock, F_GETFL, 0);
	if (flags == -1) {
		close(sock);
		return -2;
	}

	if (fcntl(sock, F_SETFL, flags | O_NONBLOCK) == -1) {
		close(sock);
		return -3;
	}

	// Set address string to pass back.
	char addrstr[128];
	const char *check;
	if (ipv == AF_INET6) check = inet_ntop(ipv, &(in6.sin6_addr), addrstr, 128);
	else check = inet_ntop(ipv, &(in4.sin_addr), addrstr, 128);
	if (!check) strcpy(addrstr, "??????");
	sv_setpv(ip, addrstr);

	return sock;
}


static struct sockaddr *createServerAddress (int proto, int port, socklen_t *len) {
	if (proto == AF_INET6) {
		*len = sizeof(struct sockaddr_in6);
		struct sockaddr_in6 *addr = malloc(*len);
		addr->sin6_family = AF_INET6;
		addr->sin6_port = htons(port);
		addr->sin6_addr = in6addr_any;
		return (struct sockaddr*)addr;
	}
	*len = sizeof(struct sockaddr_in);
	struct sockaddr_in *addr = malloc(*len);
	addr->sin_family = AF_INET;
	addr->sin_port = htons(port);
	addr->sin_addr.s_addr = htonl(INADDR_ANY);
	return (struct sockaddr*)addr;
}


static void getInetPort (int proto, int sock, int *port) {
	struct sockaddr_in in4;
	struct sockaddr *info = (struct sockaddr*)&in4;
	struct sockaddr_in6 in6;
	socklen_t len = sizeof(in4);
	if (proto == AF_INET6) {
		info = (struct sockaddr*)&in6;
		len = sizeof(in6);
	}
	if (getsockname(sock, info, &len) == -1) {
		*port = -1;
		return;
	}
	if (proto == AF_INET6) *port = ntohs(in6.sin6_port);
	else *port = ntohs(in4.sin_port);
}


MODULE = Sloop::Server		PACKAGE = Sloop::Server

PROTOTYPES: ENABLE


int addToEpoll (self, client, state)
HV *self
HV *client
int state
CODE:
// See addFD() in perl code.
	SV **field = hv_fetch(client, "fd", 2, 0);
	int fd = SvIV(*field);

	// Create epoll args
	struct epoll_event ev = { .events = state };
	sprintf((char*)&ev.data.ptr, "%d", fd);

	// Add to epoll set
	field = hv_fetch(self, "epollfd", 7, 0);
	int epfd = SvIV(*field);
	if (epoll_ctl(epfd, EPOLL_CTL_ADD, fd, &ev) == -1) XSRETURN_IV(errno);
	SvREFCNT_inc(client);

	RETVAL = 0;
OUTPUT:
	RETVAL


int bindAndListen (proto, portSV, queue_sz)
int proto
SV *portSV
int queue_sz
CODE:
/* Creates a new tcp/ip socket and calls bind() and listen() on it.
Returns the file descriptor for the socket, or:
 -1 socket() failed.
 -2 bind() failed.
 -3 listen() failed.
An SV is used for the port number so that if 0 is used it can
be set correctly. */
	int port = SvIV(portSV);

	int ipv = AF_INET;
	if (proto == 6) ipv = AF_INET6;

	int sock = socket(ipv, SOCK_STREAM, 0),
		yes = 1;
	if (sock == -1) XSRETURN_IV(-1);

	setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int));

	socklen_t len;
	struct sockaddr *addr = createServerAddress(ipv, port, &len);


	if (bind(sock, addr, len) == -1) {
		close(sock);
		free(addr);
		XSRETURN_IV(-2);
	}

	if (listen(sock, (unsigned int)queue_sz) == -1) {
		close(sock);
		free(addr);
		XSRETURN_IV(-3);
	}

	// Get actual port if 0 was used.
	if (!port) {
		getInetPort(ipv, sock, &port);
		sv_setiv(portSV, port);
    }

	free(addr);
	RETVAL = sock;
OUTPUT:
	RETVAL


void closeSocket (sock)
int sock
CODE:
// for closing the server listen socket at end
	shutdown(sock, 2);
	close(sock);


int Epoll (self)
HV *self
CODE:
/* Epoll() returns:
	<0 Error.
	 0 Nothing to do.
	>0 A combination of one or more of the ACTIVE #define constants.
*/
	struct epoll_event current[MAXPOLLEVENTS];
	// Fetch some stuff.
	SV **field = hv_fetch(self, "pollwait", 8, 0);
	int timeout = SvIV(*field) * 1000;
	field = hv_fetch(self, "epollfd", 7, 0);
	int epfd = SvIV(*field);
	field = hv_fetch(self, "fd", 2, 0);
	int sloop = SvIV(*field);
	field = hv_fetch(self, "fdlist", 6, 0);
	HV *fdlist = (HV*)SvRV(*field);
	// The "active" fd list parts.
	field = hv_fetch(self, "active", 6, 0);
	HV *active = (HV*)SvRV(*field);
	AV *in = NULL,
		*new = NULL,
		*out = NULL,
		*done = NULL;

	/* The poll */
	int sz = epoll_wait(epfd, current, MAXPOLLEVENTS, timeout);
	if (sz < 1) {
		if (sz == 0) XSRETURN_IV(0);
		if (errno < 0) XSRETURN_IV(errno);
		XSRETURN_IV(errno * -1);
	}

	/* Process */
	HV *client;
	int ev, set = 0;
	for (int i = 0; i < sz; i++) {
	// Get client from fdlist.
		const char *fdstr = (const char*)&current[i].data.ptr;
		field = hv_fetch(fdlist, fdstr, strlen(fdstr), 0);
		if (!field) {
			fprintf(stderr, "*!*!*! Epoll() fd %s not in fdlist! !*!*!*\n", fdstr);
			int fd = atoi(fdstr);
			if (fd > 2) {
				int check = epoll_ctl(epfd, EPOLL_CTL_DEL, fd, NULL);
			// Rather than gambling on the potentially different and incompatible
			// versions of strerror_r() here, stick with plain strerror(); the GNU
			// implementation is demonstrably threadsafe anyway.
				if (check) fprintf (
					stderr,
					"epoll_ctl () fail: %s\n",
					strerror(errno)
				);
				check = close(fd);
				if (check) fprintf (
					stderr,
					"close() fail: %s\n",
					strerror(errno)
				);
			}
			continue;
		}
		client = (HV*)SvRV(*field);

		ev = current[i].events;

	// hang-ups
		if (ev & EPOLLHUP) {
			if (!done) {
				field = hv_fetch(active, "done", 4, 0);
				done = (AV*)SvRV(*field);
				set += ACTIVE_DONE_SET;
			}
			av_push(done, newRV_inc((SV*)client));
			continue;
		}

	// errors
		if (ev & EPOLLERR) {
		// Check if the client has already been shut down/cancelled.
			field = hv_fetch(client, "SHUT", 4, 0);
			int checkshut = SvIV(*field);
			if (!checkshut) {
			// Check for previous errors.
				field = hv_fetch(client, "errcount", 8, 0);
				int ec = SvIV(*field) + 1;
				field = hv_fetch(client, "fault_max", 9, 0);
				int max = SvIV(*field);
				hv_store(client, "errcount", 8, newSViv(ec), 0);
				if (ec >= max) checkshut = 1;
			}
			if (checkshut) {
				if (!done) {
					field = hv_fetch(active, "done", 4, 0);
					done = (AV*)SvRV(*field);
					set += ACTIVE_DONE_SET;
				}
				av_push(done, newRV_inc((SV*)client));
				continue;
			}
		}

	// in
		if (ev & EPOLLIN) {
			field = hv_fetch(client, "fd", 2, 0);
			int fd = SvIV(*field);
			if (fd == sloop) { // new
				if (!new) {
					field = hv_fetch(active, "new", 3, 0);
					new = (AV*)SvRV(*field);
					set += ACTIVE_NEW_SET;
				}
			// accept and set details
				SV *addr = newSV(64);
				int sock = acceptClient(self, addr);
				av_push(new, newSViv(sock));
				av_push(new, addr);
			} else {
				if (!in) {
					field = hv_fetch(active, "in", 2, 0);
					in = (AV*)SvRV(*field);
					set += ACTIVE_IN_SET;
				}
				av_push(in, newRV_inc((SV*)client));
			}
		}

	// out
		if (ev & EPOLLOUT) {
			if (!out) {
				field = hv_fetch(active, "out", 3, 0);
				out = (AV*)SvRV(*field);
				set += ACTIVE_OUT_SET;
			}
			av_push(out, newRV_inc((SV*)client));
		}
	}

	RETVAL = set;
OUTPUT:
	RETVAL


int epollCreate ()
CODE:
	RETVAL = epoll_create1(EPOLL_CLOEXEC);
OUTPUT:
	RETVAL


void removeFromEpoll (self, client)
HV *self
HV *client
CODE:
	SV **field = hv_fetch(client, "fd", 2, 0);
	int fd = SvIV(*field);
	field = hv_fetch(self, "epollfd", 7, 0);
	epoll_ctl(SvIV(*field), EPOLL_CTL_DEL, fd, NULL);
	SvREFCNT_dec(client);


int setGroups (gids)
AV* gids
CODE:
	int end = av_top_index(gids);
	if (end == -1) XSRETURN_IV(0);

	gid_t *list = malloc(sizeof(gid_t) * (end + 1));
	for (int i = 0; i <= end; i++) {
		SV *cur =  av_shift(gids);
		list[i] = SvIV(cur);
	}

	int rv = setgroups((size_t)(end + 1), list);
	free(list);
	RETVAL = rv;
OUTPUT:
	RETVAL
