package Sloop::Server;
our $VERSION = '0.2.0';
# 0.1.1:
# - Corrected how close-on-exec is set in XS acceptClient().
# - Corrected occasional use of "`keys` on reference" ("is experimental").
# 0.1.2:
# - Fixed leak when replacing objects via addFD().
# - Added check in XS code for connection "SHUT".
# - Added check in connectOther().
# 0.2.0:
# - Handler may now optionally be an array to use Sloop::Lib::checkURL().
require 5.10.0;
use strict;
use warnings;

use parent 'Sloop::Base';

require XSLoader;
XSLoader::load('Sloop::Server', $VERSION);

use Carp;
our @CARP_NOT = qw(Sloop::Client Sloop::Client::Request);
use Errno qw(EINTR);
use Sloop::Constants qw(:ClientState :Debug :PollEvents);
use Sloop::Lib qw(checkURL errorString);
use Sloop::Logger;

use constant {
# Mirrors those used with Epoll() in XS code.
	ACTIVE_IN_SET => 1,
	ACTIVE_OUT_SET => 2,
	ACTIVE_DONE_SET =>4,
	ACTIVE_NEW_SET => 8
};

=encoding utf8

=head1 NAME

Sloop::Server

=head1 DESCRIPTION

This is the fundamental sloop server class.
It is a subclass of Sloop::Base.

=head1 SYNOPSIS

 my $sloop = Sloop::Server->new (
     client_constructor => (optional) sub ref
     debug => (optional) integer
     groups => (optional) array ref
     id => (optional) see Sloop::Base
     IPv => (optional) 4 or 6
     lastwords => (optional) sub ref
     logger => (optional) Sloop::Logger object
     options => (optional) see below; default 1
     pollwait => (optional) integer seconds
     port => (optional) integer
     queue_sz => (optional) integer
     secure => (optional) hash ref
     sigpipe => (optional) scalar or sub ref
     timeout => (optional) integer seconds
     uid => (optional) integer
     virtualHosts => (optional) hash ref
 );

The default for 'id' is 'MAIN'.  The other options are described below.

=over

=item B<client_constructor>

This must be from a Sloop::Client derived class.  It should NOT be a simple coderef
such as C<\&Sloop::Client::Regular::new>. Instead, use the following form:

 client_constructor => sub {
     Sloop::Client::Regular->new (
         @_,
         debug => DBG_INPUT
     );
 }

In this example, the 'debug' value will override any in @_ passed by the sloop server.

=item B<debug>

Zero, or one or more of the 'Debug Constants' from Sloop::Constants added (or
OR'd) together.  Unless specifically overridden in the client_constructor, this
is also the debug value passed to new client objects.  The default is 0. 

=item B<groups>

If defined, this should be an array of supplementary group names (not GIDs) to add
for the server process.  Doing this may require starting with privileges.

=item B<IPv>

Valid values are either 4 (the default) or 6.  When using IPv6, Sloop will run
in "dual stack" mode; i.e., IPv4 clients will be accepted as IPv6 clients (see
C<man 7 ipv6>).  The default is 4. 

=item B<lastwords>

An optional subroutine reference; if defined it will be called in shut(), e.g.,
when the object is destroyed.

=item B<logger>

An object matching the specifications of Sloop::Logger.  If not defined
Sloop::Logger will be included via 'require' to create the default. Unless
specifically overridden in the client_constructor, this is also the logger value
passed to new client objects. 

=item B<options>

This is enabled by default by setting it to 1, but it can be disabled
by setting it to undef or an empty hash.

If defined and NOT a hash reference, Sloop will require Getopt::Std and process
the following options in the following order.
Subsequently, $self->{options} will be a reference to the hash from getopts(). 
If defined and a hash reference, that will be processed as if returned from
getopts (meaning, all actual command line options will be ignored).

In either case, these take precedence over anything set by other constructor
arguments.

=over 

=item -1 [path]

STDOUT will be redirected to the specified file path.  If this path is just '2',
it will be redirected to STDERR. 

=item -2 [path]

STDERR will be redirected to the specified file path.  If this path is just '1',
it will be redirected to STDOUT. 

=item -d

(Daemonize) When $sloop->run is called, the process will fork into the
background, setsid, and redirect STDIN from /dev/null.  It does NOT change
directory. 

=item -g [debug constants]

The argument is one or more of the keys to DebugConstants or'd together, sans
DBG_ and case insensitive, e.g., "request | reply" or "connection".  This will
replace any value submitted to the constructor's optional 'debug' argument, so no
argument will turn debugging output off.

=item -p [port number]

A port number to use.

=back

=item B<pollwait>

Time in seconds to return from the internal epoll() call if nothing happens (no
requests are received, no replies are ready).  Perhaps more significantly, this
is also the frequency with which sloop checks client timeouts, regardless of
whether epoll ever times out or not (the frequency of the check is not the same
as the timeout length, see 'timeout' below).  The default is 15. 

=item B<port>

The port to use.  The default is 80.  You may use zero, in which case the
system will assign a random open port; this is accessible after the
constructor call via C<$sloop-E<62>{port}>.

=item B<queue_sz>

The maximum wait queue length to pass to listen().  The default is 500.

=item B<secure>

If defined, this will make sloop run in secure (https) mode using gnuTLS. The
argument is a hash ref with same keys described for the arguments to
C<Sloop::Socket::TLS:init()>; Required are at least the 'certificateChain' and
'privKey' fields. 

If you specify this and use a custom client_constructor as well, be sure it is
Sloop::Client::Secure based.

=item B<sigpipe>

By default the constructor sets C<$SIG{PIPE} = "IGNORE">; you can override
that with this (or set it differently post construction).  See "Signals" in
perlipc.

=item B<timeout>

The lifespan of a persistent connection in seconds.  This only applies when
awaiting a new request -- clients in the middle of an upload, or in need of a
reply, will not be timed out. The default is 30.  Note that if there is no
activity, the timeout will be checked when the server epoll() times out; since
this value is set with `pollwait` (above), it makes sense to set `timeout to
some multiple of `pollwait` (but it is not required to be).

=item B<uid>

If defined, the sloop process will setuid and gid to this value I<after> the
server socket has been initialized.  Note that (as per normal), doing this does
not include any of the  other groups the user represented by uid is a member of,
but you can add any you want via the C<groups> option above. 

=item B<virtualHosts>

If you want to use name based virtual hosting, you can define this I<instead of>
{handlers} (see ATTRIBUTES).  The top level keys of the hash should be the
hostnames B<in all lower case>.  Their values should be hash references identical
to {handlers}.

Note that name based virtual hosting won't work with HTTP 1.0 clients that don't
include a 'Host:' header (HTTP 1.1 clients are required to).  If you want to do
something special with these, define C<default> as one of the hosts.  If this is
not defined, the client receives a 404.

=back

=head1 ATTRIBUTES

In addition to the attributes listed for Sloop::Base, the following fields from the
constructor are accessible by the same name, with the same meaning:

=over

=item client_constructor

=item groups*

=item IPv*

=item lastwords

=item options*

=item pollwait

=item port*

=item queue_sz*

=item timeout

=item uid*

=item virtualHosts

=back

An asterisk indicates changing this field won't have any effect.

=over

=item B<$sloop-E<62>{handlers}>

You need to define this before you call C<$sloop-E<62>run>.  It should be a hash
ref reference whose values are either further hash references or subroutine
references.  Sloop cascades through this to resolve paths.  For example: 

 $sloop->{handlers} = {
     '/' => \&handlerA,
     foo => {
         '/' => \&handlerB,
         bar => \&handlerC
     },
     banana => \&handlerD
 };

'/' represents the relative top level, so in this case,

=over

=item - Requests beginning with /foo/bar/ will go to handlerC.

=item - Requests beginning with /foo/ but not /foo/bar/ will go to handlerB.

=item - Requests beginning with /banana/ will go to handlerD.

=item - All other requests will go to handlerA.

=back

Notice this means I<any> request beginning with /foo/bar (of which there is
obviously an infinite number) will go to handlerC, and likewise any request
beginning with /banana (e.g., /banana/apple/cherry) will go to handlerD,
etc. This can be delimited by using a special array format instead, as
described below. The handler itself receives one argument, a Sloop::Client
derived object with a Sloop::Client::Request attached.  Its return value is
ignored. 

Sloop will automatically prune '..' out of any path, so '/foo/../bar' will
be handled as '/foo/bar/'.  No redirect is sent to the client, but note
the URL will be stable this way.

Alternately, a handler may be an array reference of the form:

 '/' => [ \&handler, $min, $max ]

in order to invoke C<Sloop::Lib::checkURL($client, $min, $max)>
before the callback fires (which will only happen if this check succeeds).
This is an easy way of controlling the number of path elements accepted
beyond the end of the cascade.  As per checkURL() parameters, the last
two elements are optional.

For more discussion about this and handlers in general,
see the HTML guide page on "Request Handling".

=item B<$sloop-E<62>{secure}>

If defined, this is an HTTPS server using TLS 1.0 for client connections.  If
not, it is a regular HTTP server.  I<You should not change this field.>  Unlike
the constructor 'secure' parameter, this does not contain any actual information,
it is just defined (or not).

=back

=head1 ENVIRONMENT VARIABLES

The constructor sets the variable $ENV{SLOOP_PROTOCOL} to either "http://" or "https://", as
appropriate, so this will be inherited by any subprocess.  You may also make use of it in your
code, presuming it executes after the constructor call.

=head1 SIGNAL HANDLERS

As noted in SYNOPSIS, by default Sloop sets C<$SIG{PIPE} = "IGNORE">.This may
occur because of violations of or failures in low level protocols -- e.g.,
sudden, inappropriate disconnections.  Such defunct connections will be
discarded.

=cut
sub new {
	my $class = shift;
	my %args = (
		debug => 0,
		pollwait => 15,
		IPv => 4,
		queue_sz => 500,
		id => 'MAIN',
		sigpipe => "IGNORE",
		timeout => 30,
		options => 1,
		@_
	);

	$SIG{PIPE} = $args{sigpipe};
	delete $args{sigpipe};

# Command line option handling, if enabled.

	if (defined $args{options}) {
		if (ref($args{options}) ne 'HASH') {
			require Getopt::Std;
			$args{options} = {};
			Getopt::Std::getopts('d1:2:g:p:', $args{options});
		}
		if (defined $args{options}->{'1'}) {
			if ($args{options}->{'1'} eq '2') {
				open STDOUT, '>&STDERR' ||
					croak "Could not redirect STDOUT to STDERR.";
			} else {
				open STDOUT, '>>', $args{options}->{'1'} ||
					croak "Could not open $args{options}->{'1'} for writing."
			}
		}
		if (defined $args{options}->{'2'}) {
			if ($args{options}->{'2'} eq '1') {
				open STDERR, '>&STDOUT' ||
					croak "Could not redirect STDERR to STDOUT.";
			} else {
				open STDERR, '>>', $args{options}->{'2'} ||
					croak "Could not open $args{options}->{'2'} for writing."
			}
		}
		if (defined $args{options}->{d}) {
			require POSIX;
			my $pid = fork;
			if ($pid < 0) { croak "\n***Fork failed!***\n" }
			elsif ($pid > 0) { exit 0; }
			POSIX::setsid();
			open STDIN, '<', '/dev/null';
		}
		if (defined $args{options}->{g}) {
			my $debug = 0;
			foreach (split /\s*\|\s*/, $args{options}->{g}) {
				my $c = $Sloop::Constants::Debug{(lc $_)};
				croak "Undefined -g debug constant '$_'" if (!$c);
				$debug += $c;
			}
			$args{debug} = $debug;
		}
		$args{port} = $args{options}->{p} if defined $args{options}->{p};
	}

	if (!exists $args{logger}) {
		require Sloop::Logger;
		$args{logger} = Sloop::Logger->new(logs => [\*STDERR, \*STDERR, \*STDOUT]);
	}

	if (!defined $args{port}) {
		if (defined $args{secure}) { $args{port} = 443 }
		else { $args{port} = 80 }
	}

	if (defined $args{virtualHosts}) {
		if (!exists $args{virtualHosts}->{default}) {
			$args{virtualHosts}->{default} = {
				'/' => sub {
					my $client = shift;
					$client->generic(404);
				}
			}
		}
	}

	my $sock = bindAndListen($args{IPv}, $args{port}, $args{queue_sz});
	if ($sock < 0) {
		if ($sock == -1) { croak "Socket error: $!" }
		if ($sock == -2) { croak "Bind error: $!" }
		if ($sock == -3) { croak "Listen error: $!" }
		return undef;
	}

# Stuff necessary for Sloop::Base superclass constructor.
	my $self = $class->SUPER::new (
		debug => $args{debug},
		fd => $sock,
		id => $args{id},
		logger => $args{logger}
	);
	bless $self, $class;
	if ($self->{debug} & DBG_CONNECTION) {
		require Devel::Peek;
		require Test::Memory::Cycle;
	}

# Stuff from user config.
	while (my ($k, $v) = each (%args)) {
		$self->{$k} = $v;
	}

# Non configurable stuff.
	$self->{epollfd} = epollCreate();
	$self->goodbye("epoll_create() failed") if ($self->{epollfd} < 1);
	# The fdlist is keyed by numerical file descriptor; the values are
	# Sloop::Connection derived objects.
	$self->{fdlist} = {};
	$self->{active} = {
		in => [],
		out => [],
		done => [],
		new => []
	};
	$self->addFD($self, SPFD_READ, 0);

# Set client constructor type.
	if (defined $self->{secure}) {
		require Sloop::Client::Secure;
		my %args = (
			priority => 'NORMAL:%PARTIAL_RENEGOTIATION',
			err_log_fd => fileno($self->{logger}->{logs}->[0])
		);
		@args{ keys %{$self->{secure}} } = values %{$self->{secure}};
	# Password, key, etc. won't survive scope of block.
		$self->{secure} = 1;
		return undef if Sloop::Socket::TLS::init(\%args);
		if (!exists $self->{client_constructor}) {
			$self->{client_constructor} = sub { Sloop::Client::Secure->new(@_) };
		}
		$ENV{SLOOP_PROTOCOL} = "https://";
	} else {
		$ENV{SLOOP_PROTOCOL} = "http://";
		if (!exists $self->{client_constructor}) {
			require Sloop::Client::Regular;
			$self->{client_constructor} = sub { Sloop::Client::Regular->new(@_) }
		}
	}

# Set supplementary groups.
	if (defined $args{groups}) {
		croak "Sloop::Server 'groups' arg not an array ref" if ref($args{groups} ne 'ARRAY');
		my @gids;
		foreach (@{$args{groups}}) {
			my $gid = getgrnam($_);
			croak("No such group: $_") if (!$gid);
			push @gids, $gid;
		}
		croak "Could not set supplementary groups: $!" if setGroups(\@gids) == -1;
	}


# Set effective uid/gid if required.
	if ($args{uid}) {
		$) = $args{uid};
		$> = $args{uid};
	}

	return $self;
}


=head1 METHODS

=head2 C<$sloop-E<62>connectOther($other)>

B<$other> is a Sloop::Other derived object.

Used to add a Sloop::Other connection to the server.
This is necessary in order to make use of $other for
sending and receiving messages.  It adds a reference to
the server's FD list which should be purged when C<$other>
is shut/defunct, so holding additional references are not required.

Returns undef if $other is undefined or defunct (as indicated by
$other->okay), and a true value otherwise.

=cut
sub connectOther {
	my ($self, $other) = (shift, shift);
	if (
		!$other ||
		!$other->okay
	) { return undef }
	$other->{epfd} = $self->{epollfd};
	$self->addFD($other, SPFD_READ) if !$other->{SHUT};
	return 1;
}


=head2 C<$sloop-E<62>epollSize()>

This returns the number of connections still being actively polled.
This includes the server listen socket, so it is normally at least 1;
the server will stop running within C<$sloop-E<62>{timeout}> seconds
if there are none.  See C<stopListening()>, below, for how that
might happen.

=cut
sub epollSize {
	return 0 if (!$_[0]->{fdlist});
	return scalar keys %{$_[0]->{fdlist}};
}


=head2 C<$sloop-E<62>run()>

Starts the server main loop.  The call blocks until
the server stops running.

=cut
sub run {
	my $self = shift;
	my $lastcheck = time;

	croak "No handlers defined!" if (!$self->{handlers} && !$self->{virtualHosts});

	while (1) {

		if (!$self->{fdlist}) {
			$self->{logger}->out (
				LOG_REQUEST,
				"Empty file descriptor list, shutting down."
			);
			$self->shut;
			return;
		}

		my $set = $self->Epoll();
		# errno * -1 is returned on error.
		if ($set < 0) {
			if ($set * -1 == EINTR) {
				$self->err_log("Epoll() interrupted by signal.");
				next;
			}
			$self->goodbye("Epoll($set) failed: '".errorString($set * -1)."'") if ($set < 0);
		}

	# Debugging output.
		if ($self->{debug} & DBG_POLL) {
			my $msg = "Epoll(): ";
			if ($set & ACTIVE_DONE_SET) {
				$msg .= "DONE ";
				$msg .= "$_->{id} " foreach @{$self->{active}->{done}};
			}
			if ($set & ACTIVE_NEW_SET) {
				my $i = 1;
				foreach (@{$self->{active}->{new}}) {
					if ($i % 2) { $msg .= "NEW FD $_ " }
					else { $msg .= "IP $_" }
					$i++;
				}
			}
			if ($set & ACTIVE_IN_SET) {
				$msg .= "IN ";
				$msg .= "$_->{id} " foreach @{$self->{active}->{in}};
				$msg .= "\n";
			}
			if ($set & ACTIVE_OUT_SET) {
				$msg .= "OUT ";
				$msg .= "$_->{id} " foreach @{$self->{active}->{out}};
				$msg .= "\n";
			}
			$msg .= "...timed out." if (!$set);
			$self->debug_log($msg);
		}

	# Check if epoll timed out with nothing.
		if (!$set) {
			$lastcheck = $self->timeCheck($lastcheck);
			next;
		}

	# Check if there are existing connections that are now defunct.  This
	# should be done first, because new connections may be reusing their
	# fd's if they were closed before the poll.
		if ($set & ACTIVE_DONE_SET) {
			foreach (@{$self->{active}->{done}}) {
				next if $_->{fd} == $self->{fd};
				# Use of SHUT is a sort of optimization with a trade-off
				# of increased coupling; see comments about this in
				# Sloop::Connection and Sloop::Base.
				$_->cancel if !$_->{SHUT};
				$self->removeFD($_);
			}
			@{$self->{active}->{done}} = ();
		}

	# Check if there are incoming new connections.
		if ($set & ACTIVE_NEW_SET) {
			while (my $fd = shift @{$self->{active}->{new}}) {
				my $ip = shift @{$self->{active}->{new}};
				if ($fd < 0) {
					my $msg = "acceptClient() failed on: ";
					if ($fd == -1) { $msg .= "accept" }
					elsif ($fd == -2) { $msg .= "set keepalive" }
					elsif ($fd == -3) { $msg .= "set nonblock" }
					$self->err_log($msg);
				} else {
					my $client = $self->{client_constructor} (
						debug => $self->{debug},
						fd => $fd,
						ip => $ip,
						logger => $self->{logger},
						epfd => $self->{epollfd},
						events => SPFD_READ
					);
					if (!defined $client) {
						$self->err_log("client_constructor failed");
					} else { $self->addFD($client, SPFD_READ, 1) }
				}
			}
			@{$self->{active}->{new}} = ();
		}

	# Check if there are existing connections with incoming data.
		if ($set & ACTIVE_IN_SET) {
			foreach my $client (@{$self->{active}->{in}}) {
				my $cstate = $client->ready;
				if ($cstate == READY_CANCEL) { $self->removeFD($client) }
				elsif ($cstate == READY_CONTINUE) { next }
				elsif ($cstate == READY_REQUEST) { $self->processRequest($client) }
				elsif ($cstate == READY_DONE) { next }
			}
			@{$self->{active}->{in}} = ();
		}

	# Check if there are existing connections ready to receive pending data.
		if ($set & ACTIVE_OUT_SET) {
			foreach my $client (@{$self->{active}->{out}}) {
				my $check = $client->sendData;
				if ($check == -1) { $self->removeFD($client) }
			# Clients which have completed a reply may have another request
			# pipelined, in which case sendData() returns 2.
				elsif ($check == 2) {
					$client->debug_log("Pipelined request.") if $client->{debug} & DBG_CONNECTION;
					my $tr = $client->takeRequest();
					if ($tr == READY_CANCEL) { $self->removeFD($client) }
					elsif ($tr == READY_REQUEST) { $self->processRequest($client) }
				}
			}
			@{$self->{active}->{out}} = ();
		}

		# This should remain the last thing in the loop since it may
		# leave fdlist undef.
		$lastcheck = $self->timeCheck($lastcheck);
	}
}


=head2 C<$sloop-E<62>stopListening()>

This will close the server listening socket, meaning it will no longer
accept new connections; it will continue to service existing ones.
When there are no more left, it will stop stop running.

=cut
sub stopListening {
	my $self = shift;
	$self->removeFD($self);
	closeSocket($self->{fd});
	$self->{fd} = -1;
	$self->{logger}->out (
		LOG_REQUEST,
		"Closing listen socket on port $self->{port};\nno more connections will be accepted."
	);
}


#$self->addFD() adds a file descriptor with corresponding Sloop::Base derived
#object to the server list of file descriptors.
sub addFD {
	# $client is a Sloop::Base derived object, $state is from the READY Sloop::Constants.
	my ($self, $client, $state) = (shift, shift, shift);

	my $check = $self->{fdlist}->{$client->{fd}};
	if (defined $check) {
		croak "Attempt to duplicate in use file descriptor ($check->{id})!\n"
			+"This should not happen. >_<" if !$check->{SHUT};
		$self->removeFD($check);
	}

	if ($client->{debug} & DBG_CONNECTION) {
		$self->debug_log("Add connection $client->{id}.");
	}

	my $err = $self->addToEpoll($client, $state);
	if ($err) {
		$self->err_log("addToEpoll() invoked errno $err");
		$client->cancel;
		return;
	}

	$self->{fdlist}->{$client->{fd}} = $client;
}


# $self->processRequest($client) finds the appropriate content handler
# for $client->{request}, and passes $client to the handler.
sub processRequest {
	my ($self, $client) = (shift, shift);

	if (defined $client->{postRequestHandler}) {
		return if (!$client->{postRequestHandler}->($client));
	}

# Virtual hosting.
	if (defined $self->{virtualHosts}) {
		my $hostname = $client->{request}->header('host');
		if (!$hostname) {
			$self->{handlers} = $self->{virtualHosts}->{default};
		} else {
			$hostname = lc $hostname;
		# Remove 'www' and port number.
			$hostname =~ s/^www\.//;
			$hostname =~ s/:\d+$//;
			if (!exists $self->{virtualHosts}->{$hostname}) {
				$self->{handlers} = $self->{virtualHosts}->{default};
			} else {
				$self->{handlers} = $self->{virtualHosts}->{$hostname};
			}
		}
	}

	my $handler = $client->{request}->cascade($self->{handlers});
	if ($handler) {
	# Handler may be an array as a short form for using Sloop::Lib::checkURL().
		my $aorc = ref $handler;
        if ($aorc eq 'ARRAY') {
			my $ray = $handler;
			$handler = $ray->[0];
			$self->goodbye (
				"Handler array does not begin with a subroutine reference.\n"
				.(join('/', @{$client->{request}->{path}})),
				1
			) if ref $handler ne 'CODE';
			return if !Sloop::Lib::checkURL($client, @$ray[1,2]);
		} elsif ($aorc ne 'CODE') {
			$self->goodbye (
				"Handler is not a subroutine reference.\n"
				.(join('/', @{$client->{request}->{path}})),
				1
			);
		}
		$handler->($client);
		return;
	}

	my $msg = "[$client->{id}] Not Found: ".$client->{request}->requestPath;
	$client->{logger}->out(LOG_REQUEST, $msg);
	$client->generic(404);
}


# $self->removeFD($client) removes $client from the server's active file descriptor list.
sub removeFD {
	my ($self, $client) = (shift, shift);
# Leave this stuff:
#print 'fdlist: '.ref($self->{fdlist})."\n";
#print 'client: '.ref($client)."\n";
#print "fd: $client->{fd}\n";
	delete $self->{fdlist}->{$client->{fd}} if exists $self->{fdlist}->{$client->{fd}};

	if ($client->{debug} & DBG_CONNECTION) {
	# Will still report a refcount of 1, despite the -1, but tests with
	# a barking Client::DESTROY show the objects getting released (see
	# also DBG_DESTROY); possibly Devel::Peek grabs one too.
		$self->debug_log (
			"Remove $client->{id}, reference count: "
			.(Devel::Peek::SvREFCNT($client) - 1)
		);
		Test::Memory::Cycle::weakened_memory_cycle_ok($client, "Circular reference check");
	}

	$self->removeFromEpoll($client);
}


# Checks the fdlist for timeout connections.
sub timeCheck {
	my ($self, $lastcheck, $now) = (shift, shift, time);
	if ($now - $lastcheck - $self->{timeout} >= 0) {
		if ($self->{debug} & DBG_POLL) {
			$self->debug_log("Checking connection timeouts...");
		}
		foreach (values %{$self->{fdlist}}) {
			$self->removeFD($_) if $_->checktime($now)
		}
		$self->{fdlist} = undef if !scalar keys %{$self->{fdlist}};
		return $now;
	}
	return $lastcheck;
}


# Fullfills Sloop::Base.
sub checktime { return 0 };


# overrides dummy in Sloop::Base
sub shut {
	my $self = shift;
	return if $self->{SHUT};
	Sloop::Socket::TLS::end() if (exists $self->{secure});
	$self->{lastwords}() if (defined $self->{lastwords});
	closeSocket($self->{fd}) if ($self->{fd} > 0);
	$self->{SHUT} = 1;
}


1;
