#define PERL_NO_GET_CONTEXT
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#define NEED_sv_2pv_flags
#include "ppport.h"

#include <gnutls/gnutls.h>
#include <gnutls/x509.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#define LOGTAG "GnuTLS | "

#define SLOOP_TRUST_ALL 1
#define SLOOP_TRUST_CERTIFIED 2

static gnutls_certificate_credentials_t ServerCred;
static gnutls_priority_t ServerPrio;
static gnutls_certificate_credentials_t TrustFile;
static FILE *ErrLog;
static HV *SIDs;


static char *pointerToString (void *p) {
	static char str[16];
	sprintf(str, "%p", p);
	return str;
}


static HV* getHVfromSessionT (gnutls_session_t s) {
	dTHX;
	char *sid = pointerToString(s);
	SV **field = hv_fetch(SIDs, sid, strlen(sid), 0);
	if (!field) return NULL;
	return (HV*)SvRV(*field);
}


// Used by libgnutls but fine for internal use here as well.
static void TLSauditLogger (gnutls_session_t s, const char* msg) {
	dTHX;
	const char *id = "???";
	if (s) {
		HV *con = getHVfromSessionT(s);
		if (con) {
			SV **field = hv_fetch(con, "id", 2, 0);
			if (field) id = SvPV_nolen(*field);
		}
	}
	fprintf(ErrLog, LOGTAG"Error [%s] %s\n", id, msg);
}


static void setTLS_error (HV *self, const char *lead, int err, int fd) {
	dTHX;
	char msg[4096];
	// Sort of paranoid since this should be elsewhere done anyway.
	if (gnutls_error_is_fatal(err) && fd >= 0) close(fd);
	if (err) sprintf(msg, "%s: <%d> %s", lead, err, gnutls_strerror(err));
	else strcpy(msg, lead);
	hv_store(self, "tls_error", 9, newSVpv(msg, strlen(msg)), 0);
}


static int finalizeSession (HV *self, int fd, gnutls_session_t session) {
	dTHX;
	gnutls_transport_set_ptr(session, (gnutls_transport_ptr_t)fd);

	// Add linkage in both directions (from object to session,
	// and session list to object).
	char *sid = pointerToString(session);
	hv_store(SIDs, sid, strlen(sid), newRV_inc((SV*)self), 0);
	hv_store(self, "secure_session", 14, newSVuv((uint64_t)session), 0);
	return 1;
}


static void initfail (const char *lead, int err) {
	fprintf(ErrLog, LOGTAG"init() %s failed", lead);
	if (err) fprintf (
			ErrLog,
			":\n%s\n",
			gnutls_strerror(err)
	);
	else fprintf(ErrLog, ".\n");
}


// Used by Sloop::Other connections.
static int validateCert (gnutls_session_t sess) {
	dTHX;
	HV *con = getHVfromSessionT(sess);
	if (!con) {
		TLSauditLogger(sess, "Sloop::Other missing from TLS session store.");
		return 1;
	}
	// Trust level "ALL" requires no validation.
	SV **field = hv_fetch(con, "trustLevel", 10, 0);
	if (field) {
		int trustLevel = SvIV(*field);
		if (trustLevel == SLOOP_TRUST_ALL) return 0;
	}

	// Get intended hostname.
	field = hv_fetch(con, "otherHostname", 13, 0);
	if (!field) {
		TLSauditLogger(sess, "No 'otherHostname' set.");
		return GNUTLS_E_CERTIFICATE_ERROR;
	}
	const char *hostname = SvPV_nolen(*field);

	// Check certificate.
	unsigned int status;
	int check = gnutls_certificate_verify_peers3(sess, hostname, &status);
	if (check != GNUTLS_E_SUCCESS) {
		char msg[1024];
		snprintf (
			msg,
			1024,
			"Hostname '%s', verify_peers3() fail: %s",
			hostname,
			gnutls_strerror(check)
		);
		TLSauditLogger(sess, msg);
		return GNUTLS_E_CERTIFICATE_ERROR;
	}
	if (!status) return 0;

	char msg[4096];
	sprintf (msg, "Certificate for '%s':\n", hostname);
	if (status & GNUTLS_CERT_INVALID) strcat (
		msg,
		"No valid signature.\n"
	);
	if (status & GNUTLS_CERT_SIGNER_NOT_FOUND) strcat (
		msg,
		"Untrusted issuer.\n"
	);
	if (status & GNUTLS_CERT_SIGNER_NOT_CA) strcat (
		msg,
		"Not signed by authority.\n"
	);
	if (status & GNUTLS_CERT_INSECURE_ALGORITHM) strcat (
		msg,
		"Insecure algorithm.\n"
	);
	if (status & GNUTLS_CERT_NOT_ACTIVATED
		|| status & GNUTLS_CERT_EXPIRED
	) strcat (
		msg,
		"Invalid date.\n"
	);
	strcat(msg, "Rejected.");

	TLSauditLogger(sess, msg);
	return GNUTLS_E_CERTIFICATE_ERROR;
}


MODULE = Sloop::Socket::TLS		PACKAGE = Sloop::Socket::TLS

PROTOTYPES: ENABLE


int clientSession (self)
HV *self
CODE:
	SV **field = hv_fetch(self, "fd", 2, 0);
	int fd = SvIV(*field);
	gnutls_session_t session;

	// Initialize
	int check = gnutls_init(&session, GNUTLS_CLIENT);
	if (check) {
		setTLS_error(self, "gnutls_init()", check, fd);
		XSRETURN_IV(0);
	}

	// Set ciphers
	field = hv_fetch(self, "priority", 8, 0);
	const char *ep = NULL;
	if (field) check = gnutls_priority_set_direct(session, SvPV_nolen(*field), &ep);
	else check = gnutls_set_default_priority(session);
	if (check != GNUTLS_E_SUCCESS) {
		gnutls_deinit(session);
		char msg[1024];
		strcpy(msg, "set priority\n");
		if (ep) strncat(msg, ep, 1000);
		setTLS_error(self, msg, check, fd);
		XSRETURN_IV(0);
	}

	// Set trust file credentials.
	if (!TrustFile) {
		gnutls_deinit(session);
		hv_store(self, "tls_error", 9, newSVpv("No trust file.", 14), 0);
		XSRETURN_IV(0);
	}
	check = gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, TrustFile);
	if (check != GNUTLS_E_SUCCESS) {
		gnutls_deinit(session);
		setTLS_error(self, "set credentials", check, fd);
		XSRETURN_IV(0);
	}

	RETVAL = finalizeSession(self, fd, session);
OUTPUT:
	RETVAL


SV *errorStr (err)
int err
CODE:
	char msg[256];
	strcpy(msg, gnutls_strerror_name(err));
	RETVAL = newSVpv(msg, strlen(msg));
OUTPUT:
	RETVAL


void end ()
CODE:
	if (ServerCred) gnutls_certificate_free_credentials(ServerCred);
	if (TrustFile) gnutls_certificate_free_credentials(TrustFile);
	gnutls_global_deinit();


int init (config)
HV *config
CODE:
// Should only be called once per process.
// The return values are documented in TLS.pm.
	if (ErrLog) XSRETURN_IV(1);

	SV **field = hv_fetch(config, "err_log_fd", 10, 0);
	if (!field) XSRETURN_IV(2);
	ErrLog = fdopen(SvIV(*field), "w");

	// 3.1.4 is required by at least verify_peers3.
	const char *version = gnutls_check_version("3.1.4");
	if (!version) {
		initfail("check_version(>=3.1.4)", 0);
		XSRETURN_IV(3);
	}

	fprintf(ErrLog, LOGTAG"Using GnuTLS version %s\n", version);

	int check = gnutls_global_init();
	if (check != GNUTLS_E_SUCCESS)  {
		initfail("global_init", check);
		XSRETURN_IV(check);
	}

	// Initialize hash table for correlating session IDs to fds; 
	// the only purpose is so fds can be reported in TLSauditLogger().
	SIDs = newHV();

	// For logging.
	gnutls_global_set_audit_log_function(TLSauditLogger);
	field = hv_fetch(config, "logLevel", 8, 0);
	if (field) gnutls_global_set_log_level(SvIV(*field));

	// Set trust file
	field = hv_fetch(config, "trustFile", 9, 0);
	if (field) {
		const char *tfpath = SvPV_nolen(*field);
		gnutls_certificate_allocate_credentials(&TrustFile);
		check = gnutls_certificate_set_x509_trust_file (
			TrustFile,
			tfpath,
			GNUTLS_X509_FMT_PEM
		);
		if (check < 0) {
			initfail("Process trust file failed", check);
			XSRETURN_IV(check);
		}
		gnutls_certificate_set_verify_function(TrustFile, validateCert);
		fprintf(ErrLog, "%d certificates trusted from %s.\n", check, tfpath);
	} else TrustFile = NULL;

	// Get certificate data.
	field = hv_fetch(config, "certificateChain", 16, 0);
	if (!field) {
		ServerCred = NULL;
		XSRETURN_IV(0);
	}

	/* Everything after this should be server only. */

	gnutls_datum_t cchain;
	STRLEN len;
	cchain.data = (unsigned char*)SvPV(*field, len);
	cchain.size = len;

	// Import certificate chain.
	gnutls_x509_crt_t *certs;
	unsigned int num_certs;
	gnutls_x509_crt_fmt_t fmt = GNUTLS_X509_FMT_PEM;
	check = gnutls_x509_crt_list_import2 (
		&certs,
		&num_certs,
		&cchain,
		fmt,
		0
	);
	if (check < 0) {
		initfail("crt_list_import", check);
		XSRETURN_IV(check);
	}

	// Get private key.
	field = hv_fetch(config, "privKey", 7, 0);
	if (!field) {
		fprintf(ErrLog, LOGTAG"Sloop::Socket::TLS::init() Failed, "
			"No private key file specified.\n");
		exit(1);
	}

	gnutls_datum_t key_src;
	key_src.data = (unsigned char*)SvPV(*field, len);
	key_src.size = len;

	// Check for password.
	field = hv_fetch(config, "password", 8, 0);
	char *pw = NULL;
	if (field) pw = SvPV_nolen(*field);

	// Import private key.
	gnutls_x509_privkey_t key;
	gnutls_x509_privkey_init (&key);
	check = gnutls_x509_privkey_import2 (
		key,
		&key_src,
		GNUTLS_X509_FMT_PEM,
		pw,
		0
	);
	if (check != GNUTLS_E_SUCCESS) {
		initfail("privkey_import", check);
		XSRETURN_IV(check);
	}

	// Associate certificate chain and private key as server credentials.
	check = gnutls_certificate_allocate_credentials(&ServerCred);
	if (check != GNUTLS_E_SUCCESS) {
		initfail("allocate_credentials", check);
		XSRETURN_IV(check);
	}

	check = gnutls_certificate_set_x509_key (
		ServerCred,
		certs,
		num_certs,
		key
	);
	if (check != GNUTLS_E_SUCCESS) {
		initfail("set_x509", check);
		XSRETURN_IV(check);
	}

	// Set ciphers.
	field = hv_fetch(config, "priority", 8, 0);
	const char* prio = SvPV_nolen(*field);
	char *ep = NULL;
	check = gnutls_priority_init(&ServerPrio, prio, (const char**)&ep);
	if (check != GNUTLS_E_SUCCESS) {
		char msg[1024];
		strcpy(msg, "gnutls_priority_init()\n-> ");
		if (ep) strncat(msg, ep, 1000);
		initfail(msg, check);
		XSRETURN_IV(check);
	}

	RETVAL =  0;
OUTPUT:
	RETVAL



int peekInSocket (self)
HV *self
CODE:
	SV **field = hv_fetch(self, "secure_session", 14, 0);
	gnutls_session_t sess = (gnutls_session_t)SvUV(*field);
	RETVAL = gnutls_record_check_pending(sess);
OUTPUT:
	RETVAL


int shakehands (self)
HV *self
CODE:
	SV **field = hv_fetch(self, "secure_session", 14, 0);
	gnutls_session_t sess = (gnutls_session_t)SvUV(*field);

	int check = gnutls_handshake(sess);
	if (check == GNUTLS_E_SUCCESS) XSRETURN_IV(1);

	if (gnutls_error_is_fatal(check)) {
		field = hv_fetch(self, "fd", 2, 0);
		setTLS_error(self, "gnutls_handshake", check, SvIV(*field));
		XSRETURN_IV(-1);
	}

    RETVAL = 0; // try handshake again
OUTPUT:
	RETVAL


int serverSession (self)
HV *self
CODE:
	SV **field = hv_fetch(self, "fd", 2, 0);
	int fd = SvIV(*field);
	gnutls_session_t session;

	// Initialize
	int check = gnutls_init(&session, GNUTLS_SERVER);
	if (check != GNUTLS_E_SUCCESS) {
		setTLS_error(self, "gnutls_init()", check, fd);
		XSRETURN_IV(0);
	}

	// Set ciphers
	check = gnutls_priority_set(session, ServerPrio);
	if (check != GNUTLS_E_SUCCESS) {
		gnutls_deinit(session);
		setTLS_error(self, "gnutls_priority_set()", check, fd);
		XSRETURN_IV(0);
	}

	// Set certificate
	check = gnutls_credentials_set(session, GNUTLS_CRD_CERTIFICATE, ServerCred);
	if (check) {
		gnutls_deinit(session);
		setTLS_error(self, "gnutls_credentials_set()", check, fd);
		XSRETURN_IV(0);
	}

	RETVAL = finalizeSession(self, fd, session);
OUTPUT:
	RETVAL


void shut (self)
HV *self
CODE:
	SV **field = hv_fetch(self, "SHUT", 4, 0);
	int shut = SvIV(*field);

	// Already closed?
	if (shut) return;
	sv_setiv(*field, 1);

	SV *sp = hv_delete(self, "secure_session", 14, 0);
	if (sp) {
		gnutls_session_t session = (gnutls_session_t)SvUV(sp);
		// Remove from session ID table.
		char *sid = pointerToString(session);
		hv_delete(SIDs, sid, strlen(sid), 0);
		// Don't use SHUT_RDWR here or this will hang if the other side
		// does not reply.
		gnutls_bye(session, GNUTLS_SHUT_WR);
		gnutls_deinit(session);
	}

	field = hv_fetch(self, "fd", 2, 0);
	int sock = SvIV(*field);
	shutdown(sock, 2);
	close(sock);


SV* sysread (self, buf, len)
HV *self
SV *buf
int len
CODE:
// duplicates the functionality of perl's OO sysread
	SvGROW(buf, len+1);
	unsigned char *data = (unsigned char*)SvPV_nolen(buf);
	SV **field = hv_fetch(self, "secure_session", 14, 0);
	if (!field) return XSRETURN_UNDEF;
	gnutls_session_t sess = (gnutls_session_t)SvUV(*field);

	long bytes = gnutls_record_recv(sess, data, len);
	if (bytes < 0) {
		if (
			bytes == GNUTLS_E_AGAIN
			|| bytes == GNUTLS_E_INTERRUPTED
		) {
			hv_store(self, "readerr", 7, newSViv(EAGAIN), 0);
			XSRETURN_UNDEF;
		} else if (bytes == GNUTLS_E_PREMATURE_TERMINATION) XSRETURN_IV(0);
		hv_store(self, "readerr", 7, newSViv(bytes), 0);
		const char *s = gnutls_strerror(bytes);
		hv_store(self, "readErrStr", 10, newSVpv(s, strlen(s)), 0);
		if (gnutls_error_is_fatal(bytes)) XSRETURN_IV(-1);
		XSRETURN_UNDEF;
	}

	// This terminator is mostly superflous and intentionally excluded
	// from the length set on the SV and returned to the caller.  It's
	// just there in case someone uses a C string function on the data.
	data[bytes] = '\0';
	SvCUR_set(buf, bytes);
	RETVAL = newSViv(bytes);
OUTPUT:
	RETVAL


SV* syswrite (self, data, len)
HV *self
char *data
int len
CODE:
// duplicates the functionality of perl's OO syswrite
	SV **field = hv_fetch(self, "secure_session", 14, 0);
	if (!field) XSRETURN_IV(-3);
	gnutls_session_t sess = (gnutls_session_t)SvUV(*field);

	long bytes = gnutls_record_send(sess, data, len);
	if (bytes < 0) {
		if (bytes == GNUTLS_E_AGAIN) XSRETURN_IV(-1);
		setTLS_error(self, "gnutls_record_send()", bytes, -1);
		if (gnutls_error_is_fatal(bytes)) XSRETURN_IV(-2);
		XSRETURN_IV(-3);
	}

	RETVAL = newSViv(bytes);
OUTPUT:
	RETVAL
