package Sloop::Generic;
our $VERSION = '0.1.0';
use strict;
use warnings;

=encoding utf8

=head1 NAME

Sloop::Generic

=head1 DESCRIPTION

Sloop::Generic is used to generate and dispatch plain HTML pages for generic HTTP errors. It
is used internally by the server.  If you wish to customize the error pages, you can override
the page() function, or, if you just want to change the text or add a status code (not all of them
are covered), you can modify the $Pages hash.  These are subroutines keyed by code (404, 500, etc.)
that receive a Sloop::Client object to send the page to.

The $Pages hash ref is assigned to Client objects when constructed, so to use something customized
you need to pass an appropriate client_constructor via the Sloop::Server constructor.  See
the 'generic' parameter in the Sloop::Client SYNOPSIS.

Included codes in the $Pages hash are 400, 403, 404, 408, 411, 413, 500, and 503.

=cut

our $Pages = {
	400 => sub {
		my $msg = "Sorry &mdash; your request was malformed or invalid.";
		page(shift, 400, $msg);
	},
	403 => sub {
		my $msg = "Access denied.";
		page(shift, 403, $msg);
	},
	404 => sub {
		my $client = shift;
		my $msg = "The resource '".$client->{request}->requestPath."' could not be found.";
		page($client, 404, $msg);
	},
	408 => sub {
		my $msg = "Request timed out.";
		page(shift, 411, $msg);
	},
	411 => sub {
		my $msg = "A <i>Content-length</i> field is required to <b>POST</b> data.";
		page(shift, 411, $msg);
	},
	413 => sub {
		my $msg = "The Content-Length of your request exceeds an allowable maximum.";
		page(shift, 413, $msg);
	},
	500 => sub {
		my $msg = "There has been a server error.";
		page(shift, 500, $msg);
	},
	503 => sub {
		my $msg = "Service is temporarily unavailable, please try again later.";
		page(shift, 503, $msg);
	}
};

=head1 SUBROUTINES

=head2 C<Sloop::Generic::page($client, $status, $msg)>

Sends a simple HTML 5 page indicating an error.

B<$client> is a Sloop::Client derived object.

B<$status> is an integer status code.  It will be used in the HTTP header
and appear on the page.

B<$msg> is an a text string to include in the page.

=cut
sub page {
	my ($client, $status, $msg) = (shift, shift, shift);
	my $html = <<ENDOC;
<!DOCTYPE html>
<html>
<head>
	<title>$status</title>
</head>
<body>
	<h1>$status</h1>
	<p>$msg</p>
</body>
</html>
ENDOC
	$client->reply(\$html, status => $status, maxage => 0);
}


1;
