package Sloop::Static;
our $VERSION = '0.1.0';
use strict;
use warnings;

use Carp;
use Sloop::Generic;
use Sloop::Lib qw(terminateDirPath);
use Sloop::Logger;
use Sloop::Static::MimeType;
use URI::Encode qw(uri_decode);

require XSLoader;
XSLoader::load('Sloop::Static', $VERSION);

=encoding utf8

=head1 NAME

Sloop::Static

=head1 DESCRIPTION

Several independent functions for dealing with static directory content. 
Sloop::Static::directory() will read static files from a directory and
deliver them via the server to a Sloop::Client.  Sloop::Static::index() will
output an html list of links for each file in a directory.

=head1 REQUIRES

This refers only to modules not part of Sloop or the perl core.

B<URI::Encode>

=head1 Sloop::Static::directory()

This function will return the contents of a file to a client, if it is found
in a given directory tree.  Note that the path used, relative to the given
tree, is the same as the path in the array returned by C<Sloop::Client::pathFromLevel()>.
For example, if the complete request path is C</wherever/static/foo/bar/file.html>
and you have handlers like this:

 $sloop->{handlers} = {
     wherever => {
         static => sub {
             Sloop::Static::directory (
                 shift,
                 path => '/opt/sloop/static/',
             );
         }
     }
 };

The "pathFromLevel" would be C</foo/bar/file.html>, thus the file searched for
would be C</opt/sloop/static/foo/bar/file.html>.  If this path leads to a
directory and the request path did not terminate in a forward slash, a 301
redirect to the same path, with / appended, is  issued (but see "A Note About
Hostname", below).  This allows you to use, e.g., an 'index.html' which
references relative paths -- C<href='whatever.css'>, etc. -- since otherwise
browsers will assume the path is one level up.  If the path is /foo/bar,
'whatever.css' will result in a request for  /foo/whatever.css; if the the path
is /foo/bar/, it will be /foo/bar/whatever.css, which is probably what you want,
since /foo/bar/ will be expanded to /foo/bar/index.html (see the 'index' argument
below for more about this last point).

This relationship to "pathFromLevel" also affects whether or not '..' links are
shown if no index.html file or function is found for the location.  This means
if you do not provide indexes and just used the ones automatically provided by
Sloop::directory::index(), below, the user will be able to travel up and down
but not up and out of the tree.

Only C<$client-E<62>{max_send}> bytes will actually be read into memory at once; see
C<buffer_data> under Sloop::Connection ATTRIBUTES regarding how this is done.

The arguments are:

 Sloop::Static::directory(
    $client,
    index => [OPTIONAL],
    maxage => [OPTIONAL],
    mimetype => [OPTIONAL],
    path => [REQUIRED]
 );

=over

=item B<$client>

A Sloop::Client derived object.

=item B<index>

A handler to pass $client to if the path ends with a directory.
Nothing more is done in this case.
If undefined, "index.html" is appended and the lookup proceeds.
If this file does not exist, C<Sloop::Static::index()> is used.  If
you don't want that, use a value that responds to C<if (defined &&
!ref)>, e.g. C<index =E<62> 1>.

=item B<maxage>

A time in seconds to pass to C<$client-E<62>reply()> for use
in the 'Cache-Control: max-age=' HTTP header.  If non-existant,
the C<$client-E<62>{maxage}> is used.

=item B<mimetype>

The mime type to use with the 'Content-Type' header.  If undefined,
this is determined via C<Sloop::Static::Mimetype::get()>, and if that fails,
defaults to 'text/plain'.

=item B<path> (REQUIRED)

Base absolute path to use.

=back

B<A Note About Hostname:> Doing a redirect requires a complete URI for the 'Location:'
header.  Since the 'Host:' field is mandatory in the request for HTTP 1.1 clients, this
is straightforward enough.  However, HTTP 1.0 clients can accept a redirect, but may not
have included a 'Host:' field, and in that case, a value from `hostname` is used (see
C<Sloop::Client::Request::hostURI()>).  If this is incorrect, so will be the redirect. If
it turns out that the hostname command is unavailable on the system, the client will recieve
a 404.

If you are using virtual hosts, don't worry about this as it only works for clients which
include a 'Host:' anyway.

=cut
sub directory {
	my $client = shift;
	my %args = ( @_ );
	$args{maxage} = $client->{maxage} if !defined $args{maxage};
	croak "No path argument given" if !defined $args{path};
	terminateDirPath($args{path});

# Get request path.
	my @pathray = $client->{request}->pathFromLevel;
	my $distanceFromTop = $#pathray + 1;
	$args{path} .= uri_decode(join('/', @pathray)) if $distanceFromTop;

# If the request path leads to a directory, use an "index.html" or user supplied
# 'index' argument.
	if (-d $args{path} || !$distanceFromTop) {

	# If the original request-line path was not '/' terminated, redirect to fix this.
		if (!$client->{request}->{slash_end}) {
			my $URI = $client->{request}->hostURI;
			if (!$URI) {
			# HTTP 1.0 client + no `hostname` command = 404 instead.
				$client->generic(404);
				return;
			}
			$URI .= $client->{request}->requestPath.'/';
			my $page = <<END;
<!DOCTYPE html>			
<html>
<body>
<h1>301</h1>
<p>This URI should end with a '/' so your browser can find related resources correctly.
<p>You should be automatically redirected to <a href="$URI">$URI/</a>.
</body>
</html>
END
			$client->reply(\$page, status=>301, extra=>{Location=>"$URI"});
			return;
		}

		if (ref $args{index}) {
			$args{index}($client);
			return;
		}

		$args{path} .= '/index.html';
		if (!-e -r $args{path} && !defined $args{index}) {
			my $list = Sloop::Static::index (
				path => substr($args{path}, 0, -10),
				dots => $distanceFromTop,
				logger => $client->{logger}
			);
			$list = "<ul></ul>" if (!$list);
			my $where = $client->{request}->requestPath;
			my $page = <<END;
<!DOCTYPE html>			
<html>
<head>
	<title>$where</title>
<head>
<h1>$where</h1>
$list
END
			$client->reply(\$page);
			return;
		}
		push @pathray, 'index.html';
	}

# read in
	my $sz = 0;
	my $data = firstSlurp($args{path}, \$sz, $client);
	if (!defined $data) {
		my $msg = "[$client->{id}] Not found: ".$client->{request}->requestPath;
		$client->{logger}->out(LOG_REQUEST, $msg);
		$Sloop::Generic::Pages->{404}->($client);
		return;
	}
	if ($client->{'!S:S!_file_fd'}) {
	# Set in firstSlurp if there is more to read; see Sloop::Connection
	# regarding the 'buffer_data' code ref.
		$client->{buffer_data} = \&continueSlurp;
	}

# set mime type
	if (!defined $args{mime}) {
		$args{mime} = Sloop::Static::MimeType::get($pathray[$#pathray]);
		$args{mime} = 'text/plain' if (!$args{mime});
	}

	$client->reply ( $data,
		type => $args{mime},
		maxage => $args{maxage},
		extra => $args{extra},
		length => $sz
	);
}


=head1 Sloop::Static::index ()

Creates a simple HTML unordered list (<ul>) indexing the files in a directory.

 my $html = Sloop::Static::index (
	path => (required) scalar,
	url => (optional) scalar,
	dots => (optional) defined/non-zero or undef/0,
	sort => (optional) sub ref,
	logger => (optional) Sloop:logger object; undef default.
 );

=over

=item B<path>

Absolute path to the directory to index.

=item B<url>

String to append to the filename for an href, otherwise just the filename is used.
You probably don't need this.

=item B<dots>

If 0 or undef (the default), the '..' directory entry will be discarded,
otherwise it will be kept ('.' is always discarded).

=item B<sort>

If defined, this should be a subroutine reference.
The routine receives the array of files from the
directory, and should return an array.  If undefined,
perl's default lexicographical sort is used.

=item B<logger>

If defined, somewhere to explain an error -- this is used when called from
Sloop::Static::directory.

=back

B<Returns> an html string.

=cut

sub index {
	my %args = (
		sort => undef,
		dots => 0,
		url => "",
		logger=> undef,
		@_
	);

	my $dh;
	if (!opendir($dh, $args{path})) {
		if (defined $args{logger}) {
			$args{logger}->out (
				LOG_ERR, 
				"Sloop::Static::index could not open '$args{path}':\n$!"
			)
		}
		return undef;
	}

	my @files;
	foreach (readdir($dh)) {
		my $absolute = $args{path}.'/'.$_;
		next if (!-r $absolute || $_ eq '.');
		next if ($_ eq '..' && !$args{dots});
		$_ .= "/" if -d $absolute;
		push @files, $_;
	}
	if (!defined $args{sort}) { @files = sort(@files) }
	else { @files = $args{sort}(@files) };
	closedir($dh);

	my $html = "<ul>\n";
	foreach (@files) {
		$html .= "<li class=\"sloopIndexed\">"
			."<a class=\"sloopIndexed\" href=\"$args{url}$_\">$_</a></li>\n";
	}
	$html .= "</ul>";
	return $html;
}

1;
