package Sloop::Static::Cache;
our $VERSION = '0.1.1';
use strict;
use warnings;
# Version 0.1.1:
# - Corrections to documentation.
# - Catch no source file in hierarchy() and return 404.

use File::Find;
use Sloop::Base;
use Sloop::Lib qw(fastSlurp lcSuffix terminateDirPath);
use Sloop::Static::MimeType;

=encoding utf8

=head1 NAME

Sloop::Static::Cache

=head1 DESCRIPTION

This non-OO module contains a few functions for caching dynamically
generated content based on the timestamp of a source file.

=head1 SUBROUTINES

=head2 page()

Takes a client object and calls its reply() method with a cached copy of a
dynamically generated page.  If the copy does not exist, or is older than
its source (see below), it is (re)generated and saved first.

 Sloop::Static::Cache::page (
 	$client,
 	$cachedPath,
 	$srcPath,
 	$generateFunc
 );

=over

=item C<$client>

A Sloop::Client object.

=item C<$cachedPath>

Path to the cached copy; it does not have to exist, but the directory should
(and be writable -- if it or the file is not so, $client will recieve a 500
error).  The suffix determines the mime type used with
C<$client-E<62>reply()>, via Sloop::Static::MimeType::get(), defaulting to
"text/plain".

=item C<$srcPath>

Path to a file which will be submitted to the $generateFunc if there is no
cached copy, or if it has a modification time more recent than $cachedPath.
This file doesn't have to contain anything; it is not examined for anything
other than the timestamp.

If the file does not exist but C<$cachedPath> does, it will be used.
Otherwise, you will need to deal with this  in C<$generateFunc>.

=item C<$generateFunc>

A sub ref which will be called when $cachedPath needs to be (re)generated. It
will be passed one argument, $srcPath, and should return a scalar containing
whatever you want sent to the client and saved in $cachePath. If this function
returns a false value, C<$client-E<62>generic(404)> will be called; no error 
will be logged.

=back

=cut
sub page {
	my ($client, $cached, $src, $func) = (shift, shift, shift, shift);
	my $mimetype = Sloop::Static::MimeType::get($cached);
	$mimetype = 'txt' if !$mimetype;

# Check for cached.
	my @statcache = stat $cached;
	if (!@statcache) {
		my $html = generate($cached, $func, $src);
		if (!defined $html) {
			$client->err_log("Could not write $cached:\n$!.");
			$client->generic(500);
	# If generator returns a blank page, assume 404.
		} elsif (!$html) { $client->generic(404) }
		else { $client->reply(\$html, type => $mimetype) }
		return;
	}

# Stat source.
	my @statsrc = stat $src;
	if (!@statsrc) {
		$client->err_log("Could not stat $src, using cached version.");
		my $html = fastSlurp($cached);
		if (!$html) {
			$client->err_log("Could not read $cached:\n$!.");
			$client->generic(500);
		} else { $client->reply(\$html, type => $mimetype) }
		return;
	} else {
# Compare timestamps.
		if ($statsrc[9] > $statcache[9]) {
	# Replace cached.
			my $html = generate($cached, $func, $src);
			if (!defined $html) {
				$client->err_log("Could not write $cached:\n$!.");
				$client->generic(500);
			} elsif (!$html) { $client->generic(404) }
			else { $client->reply(\$html, type => $mimetype) }
			return;
		}
	# Use current cached.
		my $html = fastSlurp($cached);
		if (!$html) {
			$client->err_log("Could not read $cached.");
			$client->generic(500);
		} else { $client->reply(\$html, type => $mimetype) }
		return;
	}
}


sub generate {
	my ($path, $func, $src) = (shift, shift, shift);
# Create page.
	my $html = $func->($src);
	return "" if !$html;
# Save page.
	my $out = undef;
	return undef if (!open ($out, ">", $path));
	print $out $html;
	close $out;

	return $html;
}


=head2 hierarchy()

Much like page(), but used to deal with a potential hierarchy of URLs starting
with a base path; this hierarchy should correspond with a directory of source
files used to generate cached pages, but you have the opportunity to transform
the URL before it is used to search the tree. 

 Sloop::Static::Cache::hierarchy (
 	$client,
 	srcroot => string,
 	cache => string,
 	suffixMap => hashref,
 	transform => sub { },
 	create => sub { }
 );

All fields except suffixMap are required.

=over

=item C<$client>

A Sloop::Client object which will receive a reply().

=item C<srcroot>

The top level directory to search for the source file provided
to the 'create' function.  The relative subpath used is whatever
is returned by the 'transform' function.

=item C<cache>

The top level directory to use for caching copies of generated
pages.  It may be empty, and should be writable by the Sloop process.
It will end up with a tree paralleling that of the source files in
'srcroot'.

=item C<suffixMap>

A hash reference containing a list of suffixes from 'srcroot' with what
they should become in 'cache'.  It should also contain a "default" key
which will be used if a path returned by the 'transform' function does
not have a suffix, or one which is not in the hash.  For example:

 suffixMap => {
 	md => 'html',
 	svg => 'jpg'
 	default => 'txt'
 }

You do not have to provide this, in which case C<suffixMap =E<62> { default =E<62> 'html' }>
is used.  Note that the suffix here doesn't have to be part of the URL, but it is
used to determine the mime type for C<$client-E<62>reply()> (see the $cachedPath parameter
to page(), above).  It should correspond to paths returned by the 'transform' function.

=item C<transform>

A subrountine which will receive the remainder of the URL left in
C<$client-E<62>{request}-E<62>pathFromLevel>;
for example, if hierarchy() is being used to handle URLs beginning
with "/foo/bar" and the actual URL is "/foo/bar/whiskey/bravo",
then 'transform' will recieve "/whiskey/bravo".  This
function should then return a subpath relative to 'srcroot', above,
used to find the source file whose path is handed to the 'create'
function.  For example, C<sub { $_[0].'txt' }> would end up
returning, e.g., "/whiskey/bravo.txt", and hierarchy() would then look
for "[srcroot]/whiskey/bravo.txt".   This file should actually exist,
and the name, as returned, I<must include a suffix>.

If the file does not exist (as readable and regular), then a debug error
is logged and the client receives a 404.

=item C<create>

A sub ref which will be called when a cached file needs to be (re)generated,
either because it doesn't exist or its corresponding source file has a newer
modification time. It will be passed one argument, the path returned by
'transform',  and should return a scalar containing whatever you want sent to the
client and saved in the cache (do not save it yourself). If this function returns
undef, C<$client-E<62>generic(404)> will be called and an error logged with
C<$client-E<62>err_log()>.

=back

=cut
sub hierarchy {
	my $client = shift;
	my %args = (
		mask => 0755,
		suffixMap => { default => 'html' },
		@_
	);
# Get paths.
	terminateDirPath($args{srcroot});
	terminateDirPath($args{cache});
	my $relpath = $args{transform}->(
		join('/', $client->{request}->pathFromLevel)
	);
	my $src = $args{srcroot}.$relpath;
	if (!-f -r $src) {
		$client->debug_log("Source not found: $src");
		$client->generic(404);
		return;
	}
	my $dest = mapSuffix($args{cache}.$relpath, $args{suffixMap});
# Create cache subdirectory if necessary.
	my $cachedir = substr($dest, 0, rindex($dest, '/')).'/';
	if (!-e $cachedir && !mkdir($cachedir, $args{mask})) {
		$client->err_log("Could not create $cachedir:\n$!");
		$client->generic(500);
		return;
	}
	page($client, $dest, $src, $args{create});
}


sub mapSuffix {
	my ($path, $map) = (shift, shift);
	my $rin = rindex($path, '.') + 1;
	if ($rin) {
		my $srcsuf = substr($path, $rin);
		my $cachesuf;
		if (exists $map->{$srcsuf}) { $cachesuf = $map->{$srcsuf} }
		else { $cachesuf = $map->{default} }
		$path =~ s/\.$srcsuf/.$cachesuf/
	} else { $path .= '.'.$map->{default} }
	return $path;
}

=head2 updateTree()

This is intended to be used in conjunction with hierarchy(), above, but
not as an actual client handler.  Instead, it uses File::Find to examine
the entire source file tree and update the corresponding cache file if 
required.  This allows you to generate the cache tree at startup to make
sure everything happens properely.

 Sloop::Static::Cache::updateTree (
 	srcroot => scalar,
 	cache => scalar,
 	suffixMap => hash ref,
 	create => sub { },
 	logger => Sloop::Logger object
 );

All are required.

=over

=item C<srcroot>

=item C<cache>

=item C<create>

Are the same as for hierarchy(), above.

=item C<suffixMap>

This functions the same way as 'suffixMap' for hierarchy, B<except no 'default' key
is used>.  If a file is found in 'srcroot' with no suffix, or its suffix is not in
this hash, it is ignored.

=item C<logger>

This will be used to indicate errors.

=back

=cut
sub updateTree {
	my %args = (
		mask => 0755,
		@_
	);
	my $pathlen = length($args{srcroot});
	find (
		{
			follow => 1,
			wanted => sub {
			# Check suffix
				my $rind = rindex($_, '.') + 1;
				return if !$rind;
				my $cursuf = substr($_, $rind);
				return if !exists $args{suffixMap}->{$cursuf};
			# Get paths.
				my $relpath = '';
				$relpath = substr($File::Find::dir, $pathlen) if ($pathlen < length($File::Find::dir));
				my $cachepath = $args{cache}.'/'.$relpath.'/';
				my $cachefile = mapSuffix($cachepath.$_, $args{suffixMap});
				my @statcache = stat $cachefile;
			# Stat current version.
				if (!@statcache) {
					if (!-e $cachepath && !mkdir($cachepath, $args{mask})) {
						$args{logger}->out (0,
							__PACKAGE__."\nCould not mkdir $cachepath:\n$!."
						);
						return;
					}
					createCached (
						$cachefile,
						$args{create},
						$File::Find::name,
						$args{logger}
					);
					return;
				}
			# Compare to source file.
				my @statsrc = stat $File::Find::name;
				if (!@statsrc) {
					$args{logger}->out (0,
						__PACKAGE__."\nCould not stat $File::Find::name:\n$!"
					);
					return;
				}
				createCached (
					$cachefile,
					$args{create},
					$File::Find::name,
					$args{logger}
				) if $statsrc[9] > $statcache[9];
			}
		}, $args{srcroot}
	);
}


sub createCached {
	my $logger = pop;
	my $html = generate(@_);
	if (!defined $html) {
		$logger->out (0,
			__PACKAGE__."\nCould not write $_[0]:\n$!"
		);
	} elsif (!$html) {
		$logger->out (0,
			__PACKAGE__."\ncreate($_[0]) returned nothing."
		);
	}
}


1;
