#define PERL_NO_GET_CONTEXT
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#include "ppport.h"

#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>

#define HIGH_SHIFT 7
#define HIGH_BIT 1 << HIGH_SHIFT
#define RSV1_SHIFT 6
#define RSV1_BIT 1 << RSV1_SHIFT
#define RSV2_SHIFT 5
#define RSV2_BIT 1 << RSV2_SHIFT
#define RSV3_SHIFT 4
#define RSV3_BIT 1 << RSV3_SHIFT
#define OPCODE_MASK 15
#define PAYLOAD_LEN_MASK 127
#define TWOBYTE_MAX 65535


static void _unmaskData_ (unsigned char *data, int len) {
	unsigned char *mask = data;
	data += 4;
	len -= 4;
	for (int i = 0; i < len; i++) {
		data[i] ^= mask[i % 4];
	}
}


MODULE = Sloop::WebSocket		PACKAGE = Sloop::WebSocket

PROTOTYPES: ENABLE


void unmaskData (dataref)
SV *dataref
CODE:
	STRLEN len;
	unsigned char *data = (unsigned char*)SvPV(SvRV(dataref), len);
	_unmaskData_(data, len);
	data += 4;
	len -= 4;
	sv_setsv(dataref, newRV_inc(newSVpv((const char*)data,len)));


HV *processFrame (ws, msgref)
HV *ws
SV *msgref
CODE:
	HV *meta;

	// Get data from reference.
	STRLEN dataLen;
	unsigned char *data = (unsigned char*)SvPV(SvRV(msgref), dataLen);
	unsigned char *buf = NULL;

	// Check for partial frame.
	SV **field = hv_fetch(ws, "partialFrame", 12, 0);
	if (field) {
		meta = (HV*)(*field);
		field = hv_fetch(ws, "payloadBuffer", 13, 0);
		STRLEN sz;
		unsigned char *p = (unsigned char*)SvPV(*field, sz);
		buf = malloc(sz + dataLen);
		memcpy(buf, p, sz);
		memcpy(buf + sz, data, dataLen);
		dataLen += sz;
		data = buf;
	} else {
		if (dataLen < 2) {
			hv_store(ws, "payloadBuffer", 13, newSVpv((const char*)data, dataLen), 0);
			if (buf) free(buf);
			XSRETURN_EMPTY;
		}

		// Add fields from first 9 bits.
		meta = newHV();
		hv_store(meta, "FIN", 3, newSViv((data[0] & HIGH_BIT) >> HIGH_SHIFT), 0);
		hv_store(meta, "RSV1", 4, newSViv((data[0] & RSV1_BIT) >> RSV1_SHIFT), 0);
		hv_store(meta, "RSV2", 4, newSViv((data[0] & RSV2_BIT) >> RSV2_SHIFT), 0);
		hv_store(meta, "RSV3", 4, newSViv((data[0] & RSV3_BIT) >> RSV3_SHIFT), 0);
		hv_store(meta, "opcode", 6, newSViv(data[0] & OPCODE_MASK), 0);
		// Client -> server frames must have the mask bit set (5.2).
		int check = (data[1] & HIGH_BIT) >> HIGH_SHIFT;
		if (check != 1) {
			hv_store(meta, "BAD", 3, newSVpv("No mask/malformed.", 18), 0);
			goto end;
		}

		// Get the payload length (see 5.2).
		uint64_t len = data[1] & PAYLOAD_LEN_MASK;
		int headerLen = 2;
		if (len > 125) {
			int sz = 2;
			if (len == 127) sz = 8;
			if (dataLen < sz + 6) { // First two bytes + mask = 6
				hv_store(ws, "payloadBuffer", 13, newSVpv((const char*)data, dataLen), 0);
				if (buf) free(buf);
				XSRETURN_EMPTY;
			}
			len = 0;
			// Multibyte values are big endian.
			for (int i = 0; i < sz; i++)  len += data[1 + sz - i] << (i * 8);
			// Make sure it is not greater than maxFrameSz.
			SV **field = hv_fetch(ws, "maxFrameSz", 10, 0);
			int max = SvIV(*field);
			if (len > max) {
				hv_store(meta, "BAD", 3, newSVpv("Payload larger than maxFrameSz.", 29), 0);
				goto end;
			}
			headerLen += sz;
		}
		// Advance the 'data' pointer to the end of the length field.
		data += headerLen;
		dataLen -= headerLen;
		hv_store(meta, "length", 6, newSVuv(len), 0);

		if (dataLen - 4 < len) { // Mask = 4 bytes.
			hv_store(ws, "partialFrame", 12, newRV_inc((SV*)meta), 0);
			hv_store(ws, "payloadBuffer", 13, newSVpv((const char*)data, dataLen), 0);
			if (buf) free(buf);
			XSRETURN_EMPTY;
		}
	}
	// Frame is complete, decode payload.
	if (dataLen > 4) {
		_unmaskData_(data, dataLen);
		data += 4;
		dataLen -= 4;
	}

	// If this is a close frame with a payload, get the status code.
	field = hv_fetch(meta, "opcode", 6, 0);
	int op = SvIV(*field);
	if (op == 8 && dataLen >= 2) {
		uint16_t closeStatus;
		memcpy(&closeStatus, data, 2);
		hv_store(meta, "closeStatus", 11, newSVuv(ntohs(closeStatus)), 0);
		data += 2;
		dataLen += 2;
		hv_store(meta, "length", 6, newSVuv(dataLen), 0);
	}

	// Reassign payload to 'msgref'.
	sv_setsv(msgref, newRV_inc(newSVpv((const char*)data,dataLen)));

	end:
	if (buf) free(buf);
	sv_2mortal((SV*)meta);
	RETVAL = meta;
OUTPUT:
	RETVAL


SV *makeFrame (dataref, optref)
SV *dataref
SV *optref
CODE:
	SV *data = (SV*)SvRV(dataref);

	// Set the first 8 bits.
	unsigned char firstByte = 0;
	if (SvROK(optref)) {
		HV *opts = (HV*)SvRV(optref);
	// Check for FIN and set it to 1 if not present.
		SV **field = hv_fetch(opts, "FIN", 3, 0);
		if (field) firstByte = SvIV(*field) << HIGH_SHIFT;
		else firstByte = 1 << HIGH_SHIFT;
		field = hv_fetch(opts, "RSV1", 4, 0);
		if (field) firstByte += SvIV(*field) << RSV1_SHIFT;
		field = hv_fetch(opts, "RSV2", 4, 0);
		if (field) firstByte += SvIV(*field) << RSV2_SHIFT;
		field = hv_fetch(opts, "RSV3", 4, 0);
		if (field) firstByte += SvIV(*field) << RSV2_SHIFT;
	// Check for opcode and set it to 1 if not present.
		field = hv_fetch(opts, "opcode", 6, 0);
		if (field) firstByte += SvIV(*field);
		else firstByte += 1;
	// Default with no opts: FIN = 1, opcode = 1.
	} else { firstByte = 129; }

	// Set the length and construct the frame header.
	uint32_t dataLen = 0;
	if (SvPOK(data)) dataLen = SvCUR(data);
	int totalLen = dataLen;
	unsigned char *frame;
	// 'p' is a pointer to the end of the frame header.
	unsigned char *p;
	if (dataLen <= 125) {
		totalLen += 2;
		frame = malloc(totalLen);
		frame[1] = dataLen;
		p = frame + 2;
	} else if (dataLen <= TWOBYTE_MAX) {
		totalLen += 4;
		frame = malloc(totalLen);
		frame[1] = 126;
		p = (unsigned char*)&dataLen;
		frame[2] = p[1];
		frame[3] = p[0];
		p = frame + 4;
	} else {
		totalLen += 10;
		frame = malloc(totalLen);
		frame[1] = 127;
		uint64_t n = dataLen;
		p = (unsigned char*)&n;
		for (int i = 0; i < 8; i++) frame[2 + i] = p[7 - i];
		p = frame + 10;
	}
	frame[0] = firstByte;

	// Add the payload, if any, at 'p'.
	if (dataLen) {
		unsigned char *payload = (unsigned char*)SvPV_nolen(data);
		memcpy(p, payload, dataLen);
	}

	// Return the frame.
	RETVAL = newSVpv((const char*)frame, totalLen);
	free(frame);
OUTPUT:
	RETVAL
