package Sloop::WebSocket;
our $VERSION = '0.2.0';
# 0.1.1: Changed Digest::SHA1 to Digest::SHA.
# 0.2.0: 
#  - Fixed circular reference problem.
#  - No more OO interface.
#  - Refined close handling.
use strict;
use warnings;

require XSLoader;
XSLoader::load('Sloop::WebSocket', $VERSION);

use Carp;
use Sloop::Client;
use Sloop::Constants;
use Sloop::Lib qw(callerTag dataLength);
use Sloop::Logger;
use Digest::SHA qw(sha1);
use MIME::Base64;


use constant {
# Values from RFC 6455; see particularly sec. 5.5.
	WS_CONTINUATION => 0,
	WS_TEXT => 1,
	WS_BINARY => 2,
	WS_CONTROL => 8,
	WS_CLOSE => 8,
	WS_PING => 9,
	WS_PONG => 10
};

our $MAGIC_GUID = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11';
our $DEFAULT_MAX_FRAME = 1 << 22; # 4 MB

=encoding utf8

=head1 NAME

Sloop::WebSocket

=head1 DESCRIPTION

Implements the websocket protocol (ws://) as defined in RFC 6455.  This is
not an OO interface.  To initialize a socket, use C<start()>; this
configures a Sloop::Client and associates it with  a handler for incoming data.
To reply, use C<sendFrame()>.  To close the socket, use C<close()>.

=head1 REQUIRES

This refers to modules not part of Sloop or the perl core.

=over

=item B<MIME::Base64>

=item B<Digest::SHA>

=back

=head1 SUBROUTINES

=head2 C<Sloop::WebSocket::start($client, %options)>

Opens the websocket. This slightly changes the way the connection is handled.
It calls C<$client-E<62>setAfterReplyHandler()> and chains any existing handler
to fire after the one it sets.

Will return undef if the request is not a valid
websocket negotiation; an error will be logged via C<$client-E<62>err_log()>.
Otherwise returns a true value.

The C<%options> are:

=over

=item B<dataIn>

I<Required>. A subroutine reference that will be called for incoming messages,
except for control/close.  Will croak if undefined. See below for the
parameters passed in.

=item B<onControl>

A subroutine reference that will be called for incoming control messages,
except for close or ping; by the current protocol this just leaves pong,
which should be returned if you manually send a ping frame first.

If undefined, the event is simply noted as a request via
C<$client-E<62>{logger}>; no response is sent and any payload is discarded.  

Ping control messages are automatically handled with a pong reply.  Close
control messages are handled by C<onClose>, if defined.

=item B<onClose>

A subroutine reference that will be called for incoming close messages.
If undefined, C<$client> is sent a close frame in response and any
payload is discarded.  If defined, it is your responsibility to make
sure a close frame is sent in response as the protocol requires this
(see section 5.5.1 in RFC 6455, and C<sendFrame()> below).

=item B<maxFrameSz>

The maximum length of a frame payload in bytes.  The default is
4194304 (4 MiB).  Frames larger than this will be discarded, an
error will be logged, and C<$client-E<62>cancel()> called.

=back

All the above callbacks (dataIn, onControl, onClose) receive three arguments:

1) A C<Sloop::Client> object associated with the websocket.  This is the
   same object as passed in to C<start()>. I<Never use a capture for this
   or you will end up with a circular reference>.

2) A I<reference> to a scalar containing the payload of the message.

3) A hash reference with the following fields (generally these aren't of
   much interest):

=over

=item B<opcode>

An opcode corresponding to one of the following constants (see section
11.8 of RFC 6455):

 Sloop::WebSocket::WS_CONTINUATION
 Sloop::WebSocket::WS_TEXT
 Sloop::WebSocket::WS_BINARY
 Sloop::WebSocket::WS_CLOSE
 Sloop::WebSocket::WS_PONG

Ping frames are dealt with automatically according to the protocol and
will not fire a callback.  Fragmented messages are assembled before
any callback is fired, and 'opcode' is set to what it was in the initial
frame (so WS_CONTINUATION should not happen).

=item B<length>

This is the length of the payload in bytes.

=item B<FIN>

The value of the FIN bit (see section 5.4 of RFC 6455), either 0 or 1.
Sloop::WebSocket assembles fragmented messages and delivers them via
callback when complete, so this should always be 1.

=item B<RSV1>

=item B<RSV2>

=item B<RSV3>

Values for the RSV1, 2, and 3 bits, either 0 or 1.  Technically these must
"be 0 unless an extension is negotiated that defines meanings for non-zero
values" and if "a nonzero value is received and none of the negotiated
extensions defines the meaning of such a nonzero value, the receiving
endpoint MUST _Fail the WebSocket Connection_" (section 5.2 of RFC 6455).
However, Sloop::WebSocket itself does not interpret these one way or the
other.

=item B<closeStatus>

This will only be present for close frames that include a status code and
possibly a further payload.  It's the integer status code (see 7.4 of RFC 6455).

=back

=cut


sub start {
	my $client = shift;
	my $websock = {
		maxFrameSz => $DEFAULT_MAX_FRAME,
		onControl => sub {
			$_[0]->{logger}->out (
				LOG_ERR,
				'Sloop::Websocket onControl fired...'
			);
		},
		@_,
		continued => 0
	};
	croak("dataIn not defined") if !defined $websock->{dataIn};

	my $req = $client->{request};
	if (!defined $req) {
		$client->err_log("No request!");
		return undef;
	}
	my $hdrs = $req->{headers};
	if (!defined $hdrs->{'upgrade'}) {
		$client->err_log("No 'Upgrade:' in request.");
		return undef;
	}
	my $wskey = $hdrs->{'sec-websocket-key'};
	if (!defined $wskey) {
		$client->err_log("No 'Sec-WebSocket-Key:' in request.");
		return undef;
	}

	$client->{WebSocket} = $websock;

	# Create Sec-WebSocket-Accept value
	my $wsAccept = encode_base64(sha1($wskey.$MAGIC_GUID));
	chomp $wsAccept;
	# Use client afterReplyHandler to get out of the HTTP request-response cycle.
	my $chainHandler = $client->{afterReplyHandler};
	$client->setAfterReplyHandler (
		sub {
			my $who = shift;
			$who->setSituation(Sloop::Client::CLIENT_ALT);
			$who->{ALT_handler} = \&Sloop::WebSocket::incoming;
			if ($chainHandler) {
				my $check = $chainHandler->($who);
				if (ref $check) { return $check }
				elsif ($check) { return $chainHandler }
			}
		}
	);

	$client->{logger}->out(LOG_REQUEST, '['.$client->{id}.'] Opening websocket.');

	# Queue reply.
	$client->reply (
		undef,
		status => 101,
		extra => {
			Upgrade => 'websocket',
			Connection => 'Upgrade',
			'Sec-WebSocket-Accept' => $wsAccept
		}
	);

	return 1;
}


sub incoming {
	my $client = shift;

# Read raw data and process.
	my $msg = $client->readData;
	return Sloop::Constants::READY_CONTINUE if !$msg;

	croak("No websocket data associated with ".$client->{id})
		if !defined $client->{WebSocket};
	my $ws = $client->{WebSocket};

	my $frameInfo;
	if (defined $ws->{partialFrame}) {
		$frameInfo = $ws->{partialFrame};
		$ws->{payloadBuffer} .= $$msg;
		if (length $ws->{payloadBuffer} < $frameInfo->{length} + 4) { # 4 bytes for mask.
			return Sloop::Constants::READY_CONTINUE;
		}
		$msg = \$ws->{payloadBuffer};
		Sloop::WebSocket::unmaskData($msg);
		# Frame complete.
		$ws->{partialFrame} = undef;
	} else {
	# New frame.
		if ($ws->{payloadBuffer}) {
		# Previously incomplete header.
			$ws->{payloadBuffer} .= $$msg;
			$msg = \$ws->{payloadBuffer};
		}
		$frameInfo = Sloop::WebSocket::processFrame($ws, $msg);
		return Sloop::Constants::READY_CONTINUE if !$frameInfo;
		$ws->{payloadBuffer} = undef;
	}
	# $msg is now only the payload.

	if (defined $frameInfo->{BAD}) {
		$client->err_log($frameInfo->{BAD});
		$client->cancel;
		return Sloop::Constants::READY_CANCEL;
	}

	# Deal with fragments.
	if ($ws->{continued}) {
		$ws->{fragmentBuffer} .= $$msg;
		$ws->{fragmentLength} += $frameInfo->{length};
		return Sloop::Constants::READY_CONTINUE if !$frameInfo->{FIN};
		# Message complete.
		$msg = \$ws->{fragmentBuffer};
		$frameInfo->{length} = $ws->{fragmentLength};
		$frameInfo->{opcode} = $ws->{continued};
		$ws->{continued} = 0;
	} elsif (!$frameInfo->{FIN}) {
		# Message to be continued.
		$ws->{continued} = $frameInfo->{opcode};
		$ws->{fragmentLength} = $frameInfo->{length};
		$ws->{fragmentBuffer} = $$msg;
		return Sloop::Constants::READY_CONTINUE;
	}

	# Deal with control frames.
	if ($frameInfo->{opcode} & WS_CONTROL) {
		my $opcode = $frameInfo->{opcode};
		if ($opcode == WS_CLOSE) {
			if (!defined $ws->{CLOSED}) {
				$ws->{CLOSED} = 1;
				$client->{logger}->out (
					LOG_REQUEST,
					'['.$client->{id}.'] Websocket closed.'
				);
				if (!defined $ws->{onClose}) {
					Sloop::WebSocket::sendFrame($client, \"", { opcode => WS_CLOSE })
				} else { $ws->{onClose}->($client, $msg, $frameInfo) }
				$client->{persist} = undef;
			} else {
				$client->cancel;
				return Sloop::Constants::READY_CANCEL;
			}
		} elsif ($opcode == WS_PING) {
		# Respond with identical payload.
			Sloop::WebSocket::sendFrame($client, $msg, { opcode => WS_PONG });
		} else { $ws->{onControl}->($client, $msg, $frameInfo) }
		return Sloop::Constants::READY_CONTINUE;
	}

	# Fire user dataIn callback.
	$ws->{dataIn}->($client, $msg, $frameInfo);
	return Sloop::Constants::READY_CONTINUE;
}


=head2 C<Sloop::WebSocket::sendFrame($client, $dataref, $frameOpts)>

Method for transmitting a frame to C<$client>

Sloop::WebSocket does not automatically fragment messages, so this is
always  really a single frame.  The protocol does not stipulate a limit on
frame length beyond a value that can fit into a 64-bit number.

B<$client>

Must have previously passed through C<start()> -- i.e., it should
generally be the object provided via one of the callbacks registered there.
If not, it will be cancelled and an error logged.

B<$dataref>

A reference to a scalar containing the message payload.
This I<must> be a reference or the call is fatal.  If it is not a
scalar reference, the behaviour is undefined.  It may be a reference
to an empty string (\"") if there is no payload.

B<$frameOpts> is a optional hash ref containing one or more keys:

=over

=item opcode

A value to use in the opcode field of the websocket frame header.
In addition to the constants listed under 'opcode' in the SYNOPSIS above,
you can use Sloop::WebSocket::WS_PING.

=item RSV1

A value to use in the RSV1 field of the websocket frame header.

=item RSV2

A value to use in the RSV2 field of the websocket frame header.

=item RSV3

A value to use in the RSV3 field of the websocket frame header.

=back

If $frameOpts is not given, the frame header will use 0 for RSV fields
and Sloop::WebSocket::WS_TEXT for the opcode (see SYNOPSIS).

=cut
sub sendFrame {
	my ($client, $dataref, $frameOpts) = (shift, shift, shift);
	croak("Second arg must be a REFERENCE.") if !ref $dataref;
	if (
	# Formerly there was a check for the ALT_handler and $client->{situation},
	# however, since that's
		!$client->okay ||
		!defined $client->{WebSocket} #||
#		!defined $client->{ALT_handler} ||
#		$client->{situation} != Sloop::Client::CLIENT_ALT
	) {
		my $tag = callerTag(__PACKAGE__);
		$client->err_log("Client defunct or connection not a functioning websocket.\n$tag");
		$client->cancel();
		return;
	}

	if (
		defined $frameOpts &&
		defined $frameOpts->{opcode} &&
		$frameOpts->{opcode} == WS_CLOSE &&
		!$client->{WebSocket}->{CLOSED}
	) {
		$client->{WebSocket}->{CLOSED} = 1;
		$client->req_log('Closing websocket.');
	}

	my $frame = makeFrame($dataref, $frameOpts);
	if (my $err = $client->setEvents(Sloop::Constants::SPFD_RW)) {
		$client->err_log("setEvents() failed, errno=$err");
		$client->cancel;
	}
	${$client->{data_pending}} .= $frame;
}


=head2 C<Sloop::WebSocket::close($client)>

Convenience method for sending a close frame with no payload.

B<$client>

See C<sendFrame()>.

=cut
sub close {
	Sloop::WebSocket::sendFrame($_[0], \"", { opcode => WS_CLOSE });
}


1;
