#!/usr/bin/perl
use strict;
use warnings FATAL => qw(all);

use Cwd;
use Module::Load;

# Don't use a relative path for install path!!

my $target = $ARGV[0];
my $installPath = $ENV{SLOOP_INSTALL};
$installPath = $ARGV[1] if defined $ARGV[1];

my @MODULE_DIRS = qw(
	Lib
	Logger
	Base
	Socket/Standard
	Client
	Static
	Server
	Multipart
	Other
	WebSocket
);

my %PreReqs = (
	Multipart => [ "URI::Encode" ],
	Static => [ "URI::Encode" ],
	WebSocket => [ "Digest::SHA" ]
);

my @TEST_DIRS = qw(
	Lib
	Client
	Multipart
	Socket/Standard
	Static
	WebSocket
	Other
);

my @SECURE_DIRS = qw(
	Socket/TLS
);

my %TARGETS = (
	clean => {
		description => "Return the tree to its original state.",
		action => sub {
			clean($_) foreach (@MODULE_DIRS);
			clean($_) foreach (@SECURE_DIRS);
		}
	},
	prep => {
		description => "Run 'perl Makefile.PL' for essential modules.",
		preOp => sub {
			print "Using $installPath as INSTALL_BASE\n" if ($installPath);
			return 1;
		},
		action => sub {
			prep($_) foreach (@MODULE_DIRS);
		}
	},
	build => {
		description => "Run 'make' for essential modules, prepping first if necessary.",
		action => sub {
			build($_) foreach (@MODULE_DIRS);
		}
	},
	install => {
		description => "Build (as necessary) and install essential modules.",
		action => sub {
			install($_) foreach (@MODULE_DIRS);
			print "\nDone.\n\n";
		}
	},
	tests => {
		description => "Run tests for all essential modules.",
		preOp => sub {
			no strict;
			my @required = (
				Sloop::Base,
				Sloop::Client,
				Sloop::Client::Request,
				Sloop::Client::Regular,
				Sloop::Connection,
				Sloop::Lib,
				Test::More
			);
			my @missing;
			foreach (@required) {
				print "Checking for $_...\n";
				eval { load $_ };
				push @missing, $_ if ($@);
			}
			if ($#missing > -1) {
				print "You need the following modules available in \@INC first:\n";
				print "\t$_\n" foreach @missing;
				print "To test using a temporary path, first export \$PERL5LIB.\n"
					."For more details, see README.txt\n\n";
				return 0;
			}
			return 1;
		},
		action => sub {
			test($_) foreach (@TEST_DIRS);
		}
	},
	secure => {
		description => "Build and install modules necessary for HTTPS service.  Requires GnuTLS.",
		action => sub {
			clean($_) foreach @SECURE_DIRS;
			prep($_) foreach @SECURE_DIRS;
			install($_) foreach @SECURE_DIRS;
		}
	}
);


usage() if (!$target || !exists $TARGETS{$target});

if (defined $TARGETS{$target}->{preOp}) {
	exit 1 if (!$TARGETS{$target}->{preOp}());
}

$TARGETS{$target}->{action}();


sub usage {
	print "Argument required.  Your choices are:\n\n";
	foreach (qw(prep build install tests clean secure)) {
		print "\033[01m$_\033[0m: $TARGETS{$_}->{description}\n";
	}
	print <<END;
\n"Essential" here refers to non-TLS modules, which are built separately by
"secure".

"Prep", "build", and "install" are stages (in that order).  If you run one,
it will include the previous ones as necessary.  Running a stage that's
already been run does nothing unless you "clean" first.

If \$SLOOP_INSTALL is  defined in the environment, it will be used as the
INSTALL_BASE. You may also specify an INSTALL_BASE as the second argument;
this will override \$SLOOP_INSTALL.  This is applied during the 'prep'
stage, so if you specify it then, it will apply to the subsequent 'install'
(and can't be overriden unless you "clean"). Do NOT use a relative path. 

"Tests" requires "install".  For more information, see \033[1mREADME.txt\033[0m. If you
just want to install sloop, run '$0 install'.\n\n
END
	exit 0;
}


sub addToRecord {
	my ($path, $field) = (shift, shift);
	die if (!$path || !$field);
	unlink "$path/.mkrec" if ($field ne 'skip' && checkRecord($path, 'skip'));
	open my $out, ">>$path/.mkrec" || die $!;
	print $out "$field\n";
	close $out;
}

sub _checkRecord {
	my ($path, $field) = (shift, shift);
	die if (!$path || !$field);
	return 0 if !-e "$path/.mkrec";
	open my $in, "<$path/.mkrec" || die $!;
	while (<$in>) {
		if ($_ =~ /$field/) {
			close $in;
			return 1;
		}
	}
	close $in;
	return 0;
}

sub checkRecord {
	my ($path, $field) = (shift, shift);
	my $check = _checkRecord($path, $field);
	printBold("\nSkipping $path...\n") if $check && $field eq 'skip';
	return $check;
}


sub build {
	my $dir = shift;
	return if checkRecord($dir, 'skip');
	return if checkRecord($dir, 'build');
	prep($dir, 'prep') if (!checkRecord($dir, 'prep'));
	printBold("\nBuilding $dir...\n");
	system("make", "-C", "$dir");
	exit $? if $?;
	addToRecord($dir, 'build') 
}


sub clean {
	my $dir = shift;
	unlink "$dir/.mkrec";
	return if !-e "$dir/Makefile";
	printBold("\nMaking clean in $dir...\n");
	`make -C $dir clean`;
	unlink "$dir/Makefile.old";
}


sub install {
	my $dir = shift;
	return if (checkRecord($dir, 'skip'));
	return if (checkRecord($dir, 'install'));
	build($dir) if !checkRecord($dir, 'build');
	printBold("\nInstalling $dir...\n");
	system("make", "-C", "$dir", "install");
	exit $? if $?;
	addToRecord($dir, 'install');
}


sub prep {
	my $dir = shift;
	return if checkRecord($dir, 'prep');

	foreach (@{$PreReqs{$dir}}) {
		eval { load $_ };
		if ($@) {
			print "\nSloop::$dir requires $_. Skip and continue [Y/n]? ";
			my $confirm = <STDIN>;
			if ($confirm =~ /^n/i) {
				print "Exiting.\n";
				exit 0;
			}
			addToRecord($dir, 'skip') if (!checkRecord($dir, 'skip'));
			return;
		}
	}

	printBold("\nPrepping $dir...\n");
	my $pwd = getcwd();
	chdir $dir;
	if (defined $installPath) { system("perl", "Makefile.PL", "INSTALL_BASE=$installPath") }
	else { system("perl", "Makefile.PL") }
	exit $? if $?;
	chdir $pwd;
	addToRecord($dir, 'prep');
}


sub printBold {
	print "\033[1m".(shift)."\033[0m";
}


sub test {
	my $dir = shift;
	return if (checkRecord($dir, 'skip'));
	if (checkRecord($dir, 'test')) {
		print "Previously completed $dir...\n";
		return;
	}

	printBold("\nTesting in $dir...\n");
	my $dh;
	if (!opendir $dh, "$dir/t") {
		print STDERR "Could not open $dir/t!\n";
		return;
	}

	my $pwd = getcwd;
	chdir $dir;
	while (readdir $dh) {
		next if $_ !~ /\.t$/;
		my @check = `perl t/$_`;
		my $first = shift @check;
		print "$_ $first";
		my $n = 1;
		foreach my $t (@check) {
			print $t;
			if (substr($t, 0, 2) ne 'ok') {
				die "Failed $dir/t/$_ test #$n";
			}
		}
	}
	closedir $dh;
	chdir $pwd;
	addToRecord($dir, 'test');
}
